!*****************************************************************************
!
! Various routine for PSF photometry
!
!****************************************************************************
module ompsfphotometry_mod
 
   use param
   use errorHandling
 
   use omutility_mod      ! In mssllib various fits access routines
   use omconstants_mod    ! In mssllib defines various constants
   use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
   use OmCoinccorr_mod       ! For coincidence-loss corrections   
   use omparameters_mod 
   use omphotometry_mod   ! In mssllib for doing aperture photometry
   use leastsquares_mod
   use optimize_mod


   implicit none


  

!*******************************************************************************
!
! The following are global variables
! 
!*******************************************************************************
    real(single), private, pointer :: mypsfImage(:, :) => null()

    integer, private :: imin, imax, jmin, jmax
    real(double), private , pointer, dimension(:, :) :: xdata => null()
    real(double), private,  pointer, dimension(:) :: ydata => null(),  &
           weight => null()

    integer(int16), private :: l0=0, diFit
    

    real(single), private :: psfxCentre, psfYcentre, sumPsfImage
   

    real(double), private :: bl(50), bu(50)
    integer(int8), private, parameter :: maxShift = 1

    integer, private :: psfNx, psfny
    integer(int8), private :: numberToFit, weightingMethod = 1
    real(single), private, pointer :: myImage(:, :) => null()
    real(single), private :: myApertureRadius, maxRadius = 0.0
    integer(int16), private, parameter :: maxNdata = 5000
    integer(int8), private :: maxNumNeighbours = 9
    real(single), private :: maxRawCountRate = 50.0
    real(double), private :: computedBackground = 0.0, std = 0.0_double
    type(gridinterpolant), save :: psfInterpolant
  
 
    integer(int8) :: maxIter = 3_int8 
    real(single), private :: time, currentTime
    real(single), private, parameter :: maxTime = 5.0
    logical, private :: displayResults  = .true.
contains

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxIter( num )
    integer(int8), intent(in) :: num

    maxIter = num

end subroutine

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxNumNeighbours( num )
    integer, intent(in) :: num

    maxNumNeighbours = num

end subroutine setMaxNumNeighbours

!********************************************************
!
!
!
!*******************************************************
subroutine setMaxRawCountRate( rate )
    real(single), intent(in) :: rate

    maxRawCountRate = rate

end subroutine setMaxRawCountRate



!********************************************************
!
!
!
!*******************************************************
subroutine setweightingmethod( method )
    integer, intent(in) :: method

    weightingMethod = method

end subroutine setweightingmethod


!*****************************************************************
!
! Loads the image and creates the PSF interpolant using the CAL
! PSF image
!
!*****************************************************************
subroutine initialisePSFPhotometry(image, maxRawCountRateIn)
    real(single), pointer :: image(:, :)
    real(single), intent(in), optional :: maxRawCountRateIn 
  

    if(present(maxRawCountRateIn)) then

        call setMaxRawCountRate( maxRawCountRateIn )

    end if

  

    
    myImage => image



    myApertureRadius = 12.0


    !**************************************************
    ! Create the psf image
    !
    !*************************************************
    call CreatethePsfImage
    
end subroutine initialisePSFPhotometry






!*****************************************************
!
!
!
!
!*****************************************************
subroutine CreatethePsfImage
    real(double) :: maxPsf
   
    real(single) ::  counts
   

   
    psfnx = 16
    psfny = 16
       
     
    l0 = 14 / binx


    if(associated( myPsfImage )) then
        deallocate( myPsfImage )
        nullify( myPsfImage )
    end if



    !**********************************************************************
    ! Get the point-spread function image
    !  
    !*********************************************************************

    call CreatePSFImage21(psfnx=psfNx, psfny=psfNy,                            & 
                          psfimageOut=myPsfImage,                              &
                          xpos=1024.5_double,                                  & 
                          ypos=1024.5_double,                                  &
                          ratio=0.5_double, bin=binx, res=1,                   &
                          psfGridFitout = psfInterpolant)



    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  


    psfXcentre = real(psfnx) / 2.0 ! The x centre of the psf image
    psfYcentre = real(psfny) / 2.0 ! The y centre of the psf image


    !*******************************************************
    !
    ! Normalize psf image to a maximum height of 1.0
    !
    !*******************************************************

    maxPsf = maxval(myPsfImage) 
    mypsfImage = mypsfImage / maxPsf


    call GetCounts2( mypsfImage, psfXCentre, psfYCentre, real(12 / binx), counts)

    psfXcentre = psfxcentre + 0.5
    psfYcentre = psfYcentre + 0.5


    sumPsfImage = counts



end subroutine CreatethePsfImage

!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
! NO FIT TO THE COORDINATES
! nSources is the number of sources 
!
!************************************************************************
subroutine PSFPhotometry1(nSources,source, di)
    integer(int8), intent(in) :: nSources
    type(SourceParameters) :: source(1 : nSources)
    integer(int8), intent(in) :: di

    integer :: nx, ny
    integer :: i0, j0, i, j, n
   
  
    real(double), dimension(1 : nSources) :: gradient, gradientError
    real(double) :: constant, constantError
    real(single) :: maxRadius1
    real(double) :: newRate, newRateError, newSignificance, counts, backgroundCounts

    type(SourceParameters) :: sourceAux

!    integer, save :: count = 0


    real(double) :: x(1 : nSources * 2)

    numberToFit = nSources


    diFit = di

   
    nx = ubound(myimage, 1)
    ny = ubound(myimage, 2)

    !******************************************************************
    !
    ! Find out roughly how many pixels to be used in the fit
    !
    !******************************************************************
    n = 0


    imin = ubound(myimage, 1)
    imax = lbound(myimage, 1)

    jmin = ubound(myimage, 2)
    jmax = lbound(myimage, 2)

    maxRadius = real(di)
    maxRadius1 = real(20 / binx)
   
 
    do i = 1, nSources

        i0 = max(0, int(source(i)%xpos - 0.5))
        j0 = max(0, int(source(i)%ypos - 0.5))

        imin = min(imin, max(0, i0 - di))
        imax = max(imax, min(nx, i0 + di))

        jmin = min(jmin, max(0, j0 - di))
        jmax = max(jmax, min(ny, j0 + di))

        j = (i - 1) * 2 + 1
   

        x(j) = source(i)%xpos 
        x(j + 1) =  source(i)%ypos 
    

  
    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata")
    ydata => allocateDouble(1, n, "ydata")
    weight => allocateDouble(1, n, "weight")

    n = 0


!*****************************************************
!
! FOR TESTING
!
!*****************************************************
!myImage = 10.0
!write(*, *) "Line 435"
!constant = PSFFunction12(x) 
!write(*, *) "Line 437"




!write(*, *) "Line 540, psfnx=", psfnx, " psfny=", psfny
!write(*, *) "The sources"
!do i = 1, nSources
!    write(*, *) i, ") x=", source(i)%xpos, " y=", source(i)%ypos
!end do
!write(*, *) "maxradius=", maxradius   


    !***************************************************************
    !
    ! Do the least-squares fit
    !
    !**************************************************************
    call doLeastSquaresFit(nSources, x,                    &
                             constant, constantError,      &
                             gradient, gradientError) 

  
    !*******************************************************************
    !
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !
    !*******************************************************************

    computedBackground = constant
    

    counts =  gradient(1) * sumPsfImage
      
    newRate =  counts / source(1)%exposure
   




    backgroundCounts = constant * PI * real(144  / &
                  (binX * binX)) 
    

    newRateError = ComputeSourceRateError1(counts, backGroundCounts, source(1)%exposure)
    




    if(counts > 0.0) then
        newSignificance = counts / sqrt(counts + backgroundCounts)        
    else
        newSignificance = 0.0
    end if 
!write(*, *) "new rate=", newRate, "+/-", newRateError
!  write(*, *) "gradient error=", gradientError(1)
    if( displayResults) then
 
        write(errstr, '(a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2, a, f0.2)' )  &
                "Old raw count-rate=", source(1)%rawCountRate, "+/-", source(1)%rawCountRateErr,  &
                ", new=", newRate, "+/-", newRateError,           &
                ", old significance=", source(1)%significance, ", new=", newSignificance  
        call message(AppMsg, Verbose)
    end if

    source(1)%apertureRadius = myApertureRadius


    if( newRate > 0.0) then
        source(1)%rawCountRate = newRate
        source(1)%rawCountRateErr = newRateError
        source(1)%significance = newSignificance
        source(1)%rawBackgroundCountRate = backgroundCounts / source(1)%exposure
        source(1)%rawBackgroundCountRateErr = sqrt(max(0.0_double, backgroundCounts)) / source(1)%exposure

        !*************************************************2010-03-03 vny
        ! Implementing the coincidence-loss correction for the psf-photometry 
        ! count rates 
        !---------------------------------------------------------------------
        sourceAux=source(1)
        call PointSourceCoincLossCorr(sourceAux, 12.0, frameTime) 
        
        source(1)%corrCountRate=sourceAux%corrCountRate
        source(1)%corrCountRateErr=sourceAux%corrCountRateErr
        source(1)%corrBackgroundCountRate=sourceAux%corrBackgroundCountRate
        source(1)%corrBackgroundCountRateErr=sourceAux%corrBackgroundCountRateErr
        source(1)%coincidenceLossCorr=sourceAux%coincidenceLossCorr
        !************************************************* 2010-03-03 vny

    else

        call warning("Negative count-rate- ignoring source")
    end if


  
   
    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


   

end subroutine PSFPhotometry1





!************************************************************************************
!
!
! Least-squares fit to PSF, including background, and including x and y
! for 1 or more sources
!
!
!***********************************************************************************
subroutine PSFPhotometry2(nSources, di, source, x)
  

    integer(int8), intent(in) :: nSources
    integer :: nx, ny
  
    integer(int8), intent(in) :: di  

    type(SourceParameters) :: source(1 : nSources)

    real(double) :: x(:)
 
    integer(int32) :: i0, j0
    integer(int32) :: n
    integer :: n0, ngen, ngenout, fitNumber(1 : nSources * 2)
    integer(int8) :: dx
   
 
    real(double) :: funcValue   
    real(double) :: tolerance
    real(single) :: dc
    integer :: i, j, nop

   

    diFit = di
    numberToFit = nSources

    nx = ubound(myimage, 1)
    ny = ubound(myimage, 2)


    !******************************************************************
    !
    ! Find out roughly how many pixels to be used in the fit
    !
    !******************************************************************
    n = 0

    imin = ubound(myimage, 1)
    imax = lbound(myimage, 1)

    jmin = ubound(myimage, 2)
    jmax = lbound(myimage, 2)

    
    dc = real(4 / binx)
    dx = di + dc
    do i = 1, nSources
        j = (i - 1) * 2 + 1
        fitNumber(j) = j
        fitNumber(j + 1) = j + 1


        x(j) = source(i)%xpos 
        bl(j) = source(i)%xpos - dc 
        bl(j) = max( -1.0_double / dble(binX), bl(j) )
    
    


        bu(j) = source(i)%xpos + dc
        bu(j) = min( dble(nx) + 0.5_double  + dble(1.0_double / dble(binX)), bu(j) )
   
    

        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - dc
        bl(j + 1) = max( -1.0_double / dble(binX), bl(j + 1) )
    
        bu(j + 1) = source(i)%ypos + dc
        bu(j + 1) = min( dble(ny) + 0.5_double  + dble(1.0_double / dble(binX)), bu(j + 1) )
  
   


        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))
        
        imin = min(imin, max(0, i0 - dx))
        imax = max(imax, min(int(nx, int32), i0 + dx))

        jmin = min(jmin, max(0, j0 - dx))
        jmax = max(jmax, min(int(ny, int32), j0 + dx))

    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dDouble(1, n, 1, int(nSources, int32), "xdata")
    ydata => allocateDouble(1, n, "ydata")
    weight => allocateDouble(1, n, "weight")

    nop = nSources * 2
 
    tolerance = 1e-6


!write(*, *) "Line 534 before calling optimize x(1)=", x(1), " x(2)=", x(2)
!write(*, *) "bl(1)=", bl(1), " bu(1)=", bu(1)
!write(*, *) "bl(2)=", bl(2), " bu(2)=", bu(2)


    do i = 1, 1

        n0 = 100
        ngen = 200  
        call optimize(nop, fitNumber, x, bl, bu, &
                   n0, ngen, tolerance, psfFunction100, funcValue, ngenout)
!        write(*, *) "Iteration ", i, ") funcvalue=", funcValue, " ngens=", ngenout
    end do  

  
 
!write(*, *) "Line 548 after calling optimize x(1)=", x(1), " x(2)=", x(2)

 

    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


   

end subroutine PSFPhotometry2


!***********************************************************
! 
! Fitting subroutine used by optimize
! 
!**********************************************************
function PSFFunction100(x) result(functionValue)

    real(double) :: x(:)
    real(double) :: functionValue
  
    real(double), dimension(1 : numberToFit) :: gradient, gradientError
    real(double) :: constant, constantError
    integer :: i, k, ndata
    real(double) :: fit, residual

    gradientError = 0.0

    constantError = 0.0
  
    maxRadius = real(diFit)

  
    !***************************************************************
    !
    ! Do the least-squares fit
    !
    !**************************************************************
    call doLeastSquaresFit(numberToFit, x,                 &
                             constant, constantError,      &
                             gradient, gradientError, ndata) 

  




    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = dble(constant)

        do k = 1, numberToFit ! Sum for each source
            fit = fit + gradient(k) * xdata(i, k)
        end do
  
        residual = fit - ydata(i)

        functionValue = functionValue + residual * residual * weight(i)

!         functionValue = functionValue + residual * dble(weight(i))



    end do 

!    functionValue = -abs(functionvalue)


    functionValue = -functionvalue / dble(ndata) 

   

!write(*, *) "PSFFUNCTION100 x=", x, " functionvalue=", functionvalue

   
end function PSFFunction100

!*************************************************************
!
!
!
!************************************************************
real(single) function PsfInterpolation(xpos, ypos)
    real(single), intent(in) :: xPos, yPos

    psfInterpolation = max(0.0, gridInterpolate(xpos, ypos, psfInterpolant))

end function psfInterpolation


!*********************************************************
!
!
!
!*********************************************************
real(single) function weightingFunction( data )
    real(double), intent(in) :: data

!    real(single), intent(in) :: radius

    if ( data .gt. 0.0_double .and. weightingMethod .eq. 1) then
        

        weightingFunction = 1.0_double / max(1.0_double, data)
        
!        weightingFunction = (1.0_double - radius / maxRadius)  / max(1.0_double, data)
!        weightingFunction = max(0.0, weightingFunction)



    else
        weightIngFunction  = 1.0_double
    end if



end function weightingFunction


!*************************************************************************
!
! The main routine for PSF photometry 
!
!
!************************************************************************
subroutine PSFPhotometry(nSources, source, di, type, minNumNeighbours)
    integer(int32), intent(in) :: nSources
    type(SourceParameters), pointer :: source(:)
    integer(int8), intent(in) :: di, type, minNumNeighbours
    integer :: i
    integer(int8) :: j, k, l
    real(single) :: tolerance, xOld, yOld
    real(double) :: x(1 : nSources * 2)
 
  
  
    type(SourceParameters), pointer :: tempSource(:) => null()
    
    
    tolerance = real(26 / binx)

    allocate(tempSource(1 : nSources))

!write(*, *) "Line 1584, nsources=", nSources, " summyimage=", sum(MyImage)

      
    !************************************************************
    !
    ! Loop through all the sources and perform aperture photometry
    !
    !************************************************************
    time = -10000.0
    displayResults = .true.

    do i = 1, nSources

!        write(*, *) i, ") line 1587"
        if(source(i)%extensionFlag == 1) then
            cycle
        end if

        if(source(i)%rawCountRate > maxRawCountRate) then
            cycle
        end if


!     write(*, *) i, ") line 1596"
   

        k = 1
   
        tempSource(1) = source(i) 
      



        !**********************************************************
        ! Find the sources with distance <= tolerance from source i
        !
        !*********************************************************
        call  findNeighbours(i, source, tolerance, tempSource, k)
        if( k .le. minNumNeighbours) then
            cycle
        end if      

        call cpu_time( currentTime)

        if( (currentTime - time) >= maxTime) then
            displayResults = .true.

            ! Write out the source list to the oswDataSet
            write(errstr,'(a, i0, a, i0)') &
                "PSF fitting to source ", i, "- number of sources included in the fit=", k
            call message(AppMsg, Verbose)
        end if



!        write(*, *) i, ") psfphotometry line 1466- fitting with ", k - 1, " neighbours"      
!        write(*, *) "The sources in te fit"
!        do j = 1, k
!            write(*, *) j, ") x=", tempSource(j)%xpos, " y=", tempSource(j)%ypos
!        end do

        select case(type)

        case (1)

            !************************************************
            ! Do not fit the coordinates
            !
            !************************************************           
            call PSFPhotometry1(k, tempSource, di)
           
        case (2)

            !************************************************
            !
            ! Fit the coordinates
            !
            !************************************************
            xOld = tempSource(1)%xpos
            yOld = tempSource(1)%ypos

           
            call PSFPhotometry2(k, di, tempSource, x)

         
        
            do j = 1, k
                l = (j - 1) * 2 + 1     
                tempSource(j)%xpos = x(l)
                tempSource(j)%ypos = x(l + 1)

            end do
            numberToFit = k 
            call PSFPhotometry1(numberToFit, tempSource, di)

!write(*, *) "final k=", k


            if( displayResults ) then
    
                write(errstr,'(a, f0.2, a, f0.2, a, a, f0.2, a)')                  &
                    "(X old - X new)=", (xOld - tempSource(1)%xpos),                &
                    " pixels, (Y old - Y new)=", (yOld - tempSource(1)%ypos), " pixels", & 
                    ", computed background=", computedBackground, " counts per pixel"

                call message(AppMsg, Verbose)
   
            end if
       


        end select

        source(i) = tempSource(1)
        time = currentTime
        displayResults = .false.
 

    end do


    deallocate(tempSource)
    nullify(tempSource)



    nx = size(myImage, 1)
    ny = size(myImage, 2)

  

    if(associated(myImage)) then
        nullify(myImage)
    end if







!write(*, *) "Line 1656, nsources=", nSources, " summyimage=", sum(MyImage)


!call saveImage(myImage, int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
!    "checkimage.fits", "checkImage.fits")
!call fatal("stopped at line 1596")


end subroutine PSFPhotometry

!************************************************
!
! Locates all sources within a distance tolerance
! of source i and adds them to the list to be
! included in the PSF fitting process
!
!***********************************************
recursive subroutine findNeighbours( i, source, tolerance, tempSource, k)
    integer, intent(in) :: i
    type(SourceParameters), pointer :: source(:)
    real(single), intent(in) :: tolerance

    integer(int8), intent(inout) :: k
    type(SourceParameters), pointer :: tempSource(:)
    integer :: j, l, minIndex    
    real(single) :: dx, dy, radius, minRadius = 1e6
    logical, pointer :: found(:) => null()



    if( maxNumNeighbours == 0) then
        return
    end if

    allocate(found(1 : size(source)))

    found = .false.
    found(i) = .true.
 
   

    do l = 1, maxNumNeighbours
        minIndex = -1
        minRadius = 1e6
        do j = 1, size(source)
    
            if( found(j)) cycle

            if(source(j)%extensionFlag == 1) then
                cycle
            end if
            if(source(j)%rawCountRate > maxRawCountRate) then
                cycle
            end if

            dx = source(i)%xpos - source(j)%xpos
            if(dx > tolerance) cycle

            dy = source(i)%ypos - source(j)%ypos
            if(dy > tolerance) cycle
            radius = sqrt(dx * dx + dy * dy)

            if(radius <= tolerance) then
                if( radius .le. minRadius) then
                    minRadius = radius
                    minIndex = j
                
                end if
            end if    
        end do

        if(minIndex > 0) then
            k = k + 1
            found(minIndex) = .true.
            tempSource(k) = source(minIndex)         
            if(k > maxNumNeighbours) then
                deallocate(found) 
                return
            end if     
        else
            exit
        end if

    end do

    deallocate(found)
    nullify(found)
    
end subroutine findNeighbours

!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
! NO FIT TO THE COORDINATES
! nSources is the number of sources 
!
!************************************************************************
subroutine doLeastSquaresFit(nSources, x,               &
                             constant, constantError,   &
                             gradient, gradientError, ndata) 

    integer(int8), intent(in) :: nSources
    real(double)  :: x(:)
    real(double), intent(out), dimension(1 : nSources) :: gradient, gradientError
    real(double), intent(out) :: constant, constantError
    integer, optional, intent(out) :: nData

    integer :: i, j, k, l, n
    real(single) :: x1, y1, dx, dy
    real(single) :: radius
  
    n = 0

    !******************************************************************
    !
    ! Loop through all the pixels that are within the source area
    !
    !******************************************************************
    do i = imin, imax
        x1 = real(i + 1)
        do j = jmin, jmax
            y1 = real(j + 1)

            !**********************************************************************
            !
            ! Get the PSF counts for this pixel
            !
            !**********************************************************************
            n = n + 1
          
            do k = 1, nSources ! nSources is the number of stars 
                xData(n, k) = 0.0       
                      
                l = (k  - 1) * 2 + 1
                dx = x1 - x(l)
         
                if(abs(dx) .gt. maxRadius) then
                    cycle
                end if
                dy = y1 - x(l + 1)
             
                if(abs(dy) .gt. maxRadius) then
                    cycle
                end if
  

                radius = sqrt(dx * dx + dy * dy)
                if(radius .gt. maxRadius) then
                    cycle
                end if
                          
                xData(n, k) = PsfInterpolation(dx, dy)
         
            end do
            
            !****************************************************
            ! Data point okay
            !*************************************************** 
            yData(n) = dble(myimage(i, j))
                
            !**************************************************************************
            ! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
            ! this is the best weighting procedure
            !**************************************************************************
            weight(n) = weightingFunction( ydata(n) )  
                 
        end do !   do i = imin, imax
    end do  ! do j = jmin, jmax
  

    !**********************************************************************************
    !
    ! Least-squares fit with constant (background) term
    !
    !**********************************************************************************  
    call LeastSquaresFit21(n, xdata=xdata, ydata=ydata,              &
            weight=weight, nterms=int(nSources, int8),               &
            maxiter=maxIter, nstd1=3.0,                              &
            gradient=gradient, gradienterror=gradientError,          &
            constant=constant, constanterror=constantError, std1=std)


    if(present(nData)) then
        nData = n
    end if

end subroutine doLeastSquaresFit


end module ompsfphotometry_mod










