!*************************************************************
! Module sourcedetection_mod
! Created 15th April 2002- C.Brindler1157
! Contains a subroutine for detecting "point" sources and one 
! for detecting any other kind of source
!
!****************************************************************


module omdetectdetection_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for array
    use OmSkyBackground_mod  ! For computing sky background mode
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use parameters_mod       ! For computing source moments
    use OmSmoothImage_mod    ! For smoothing image
    use OmConstants_mod      ! For binx, biny, etc
    use omnumerics_mod 
    use omphotometry_mod
    use leastsquares_mod
    use ominterpolate_mod
    use nonlinls_mod
    use brent_mod
    use cubicshepard_mod
    use splinefit_mod
    use omsourcemap_mod
    use psffunctions_mod  
    use omutility_mod
    use gridinterpolation_mod
    use centroid_mod
    use zeroin_mod



public DetectPointSources, DetectExtendedSources
public IsPointSource, IsPointSource2



!************************************************************************
!
! Various interfaces
!
!***********************************************************************
interface DetectExtendedSources 
    module procedure DetectExtendedSources1, DetectExtendedSources2,  &
                     DetectExtendedSources3,  DetectExtendedSources4, &
                     DetectExtendedSources5
end interface DetectExtendedSources 


interface NumberOfStars
    module procedure NumberOfStars1, numberofstars2, &
                     numberOfStars3, numberofstars4
end interface NumberOfStars

interface NumberOfNeighbours
    module procedure NumberOfNeighbours1, numberofNeighbours2, &
                     NumberOfNeighbours3
end interface NumberOfNeighbours

interface BackgroundGradientOkay
    module procedure  BackgroundGradientOkay1,  BackgroundGradientOkay2, &
                      BackgroundGradientOkay3
end interface BackgroundGradientOkay

interface CheckVicinity
    module procedure  CheckVicinity1, checkVicinity2, checkvicinity3
end interface CheckVicinity

interface SourceSeparation
    module procedure  SourceSeparation1, sourceSeparation2
end interface SourceSeparation




!****************************************
! A structure to store detection parameters, 
! etc!
!****************************************
type DetectionParameters
    real(single) :: nStd(3) = (/5.0, 2.0, 1.0/)   ! The standard-deviation constant
    integer      :: nCircles = 4 ! The number of circles to compute mean counts for
    integer      :: nSize = 16   ! The number of circle intervals
    integer      :: kmin = 8     ! The minimum number of data points round circle
    real(single) :: minSigNoise = 5.0 ! 
    real(single) :: c2 = 1.0
    real(single) :: minSemib = 0.5
    real(single) :: semiAConstant = 1.0 
    real(single) :: meanSemia = 1.0_double
    real(double) :: semiaSum = 0.0_double
    integer :: n = 0
    real(single) :: exposure   
    integer(int16) :: minNoPixels
end type DetectionParameters


!***************************************************
!
! Global variables
!
!***************************************************
real(single), private, pointer :: FWHMImagePtr(:, :) => null()
real(single), private :: psfXcentre, psfYCentre
real(double), private :: fCentre
real(single), private, pointer, dimension(:) :: mean => null(), sigma => null()
real(single), private, pointer :: rCosTheta(:, :) => null(), rSinTheta(:, :) => null()
real(single), private, pointer :: imagePtr(:, :) => null() 
real(single), private, pointer :: smoothedImagePtr(:, :) => null() 
real(single), private, pointer :: backgroundImagePtr(:, :) => null() 
real(single), private, pointer :: backgroundSigmaImagePtr(:, :) => null() 
real(single), private :: xLower, yLower, xUpper, yUpper
real(single), private :: minimumCounts ! For isValidSplitSource

real(single), private, pointer :: fCircle(:) => null()
integer(int8), private :: minCount = -1
integer(int16), private :: nx1, ny1, lbx, lby, ratio1, ratio2
integer(int16), private, pointer :: pLevelImagePtr(:, :) => null() ! level image map for point-sources
integer(int16), private, pointer :: eLevelImagePtr(:, :) => null() ! level image map for extended-sources
integer(int16), private :: number = 0     ! The number assigned to any elevelImagePtr pixel above
                                          ! the specified height above the pixel 

type(DetectionParameters), private, save :: detectParams 
logical(boolean), pointer, private :: maskImage(:, :) => null()
logical(boolean), pointer, private :: emaskImage(:, :) => null()
integer(int8), private :: extensionFlag = 0_int8
real(single), private :: minHeight = 3.0  ! Was 5  
real(single), private, parameter :: xOffset(5) = (/0.0, -0.25, 0.25, -0.5, 0.5/)
real(single), private, parameter :: yOffset(5) = (/0.0, -0.25, 0.25, -0.5, 0.5/)

character(100), private :: comment
real(single), private :: psfSemia1 = -1.0, psfSemiB1 = -1.0
real(single), private :: psfSemia = -1.0, psfSemiB = -1.0


!**********************************************************************************

integer, private :: nSourceRadii
real(single), pointer, dimension(:) :: sourceCircleradii => null(), sourceCirclef =>  null()
real(double) :: fBackground


real(single), private, dimension(10) :: xcoeff, dxcoeff
integer(int8), private :: order




character(100), private :: title

!*********************************************************
! The following is for test purposes to find out why a
! particular source hasn't been detected
! ipixel, jpixel- pixel coordinates of brightest pixel
! in undetected source
!*********************************************************

!integer, public, parameter :: ipixel=11, jPixel=4
!logical, private :: checkPixel = .false.

contains

!*********************************************************
!
!
!********************************************************
real(single) function getPSFSemia()

    getPSFSemia = psfsemia


end function getPSFSemia


!*********************************************************
!
!
!********************************************************
real(single) function getPSFSemib()

    getPSFSemib = psfsemib


end function getPSFSemib


!*******************************************************
!
!
!*******************************************************
subroutine setTitle( text )
    character(*), intent(in) :: text

    title = text

end subroutine setTitle

!*********************************************************
! Access functions
! for retrieving/setting private data from elsewhere
!*******************************************************
function GetMaskImage() result(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage1 => maskImage

end function GetMaskImage
!*********************************************************
!
!********************************************************
function GetEMaskImage() result(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage1 => emaskImage

end function GetEMaskImage
!*********************************************************
!
!********************************************************
subroutine setNumber(no)
    integer(int16), intent(in) :: no
    number = no
end subroutine setNumber

!*********************************************************
!
!
!********************************************************
subroutine setMaskImage(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    maskImage => maskImage1
     
end subroutine setMaskImage

!*********************************************************
!
!
!********************************************************
subroutine setImagePtr(image)
    real(single), pointer :: image(:, :)

    imagePtr => image
     
end subroutine setImagePtr

!*********************************************************
!
!
!********************************************************
subroutine setSmoothedImagePtr(image)
    real(single), pointer :: image(:, :)

    smoothedImagePtr => image
     
end subroutine setSmoothedImagePtr

!*********************************************************
!
!
!********************************************************
subroutine setBackgroundImagePtr(image)
    real(single), pointer :: image(:, :)

    backgroundImagePtr => image
     
end subroutine setBackgroundImagePtr


!*********************************************************
!
!
!********************************************************
subroutine setELevelImagePtr(image)
    integer(int16), pointer :: image(:, :)

    elevelImagePtr => image
     
end subroutine setElevelImagePtr


!*********************************************************
!
!
!********************************************************
subroutine nullifyELevelImagePtr()
  

    nullify( elevelImagePtr )
     
end subroutine nullifyElevelImagePtr

!*********************************************************
!
!
!********************************************************
subroutine setPLevelImagePtr(image)
    integer(int16), pointer :: image(:, :)

    plevelImagePtr => image
     
end subroutine setPlevelImagePtr

!*********************************************************
!
!
!********************************************************
subroutine setEMaskImage(maskImage1)
    logical(boolean), pointer :: maskImage1(:, :)

    emaskImage => maskImage1
     
end subroutine setEMaskImage

!********************************************************
!
!*******************************************************
subroutine setMinimumCounts(minCounts)
    real(single), intent(in) :: minCounts

    minimumCounts = minCounts

end subroutine setMinimumCounts


!********************************************************
!
! minCount is the minimum number of pixels above the 
! background to be considered a point source
!
!*******************************************************
subroutine setMinCount( minCountIn )
    integer(int8), intent(in) :: minCountIn

    minCount = minCountIn

end subroutine setMinCount









!********************************************************
!
!
!*******************************************************
subroutine setMinimumNoPixels(minNoPixels)
    integer(int16), intent(in) :: minNoPixels

    detectParams%minNoPixels = minNoPixels - 1
 
end subroutine setMinimumNoPixels
!********************************************************
!
!
!*******************************************************
subroutine setMinimumHeight(height)
    real(single), intent(in) :: height

    minheight = height
 
end subroutine setMinimumHeight

!***********************************************************
! 
!
!***********************************************************
subroutine setBackgroundSigmaImagePtr(sigmaImagePtr)
    real(single), pointer :: sigmaImagePtr(:, :)
    backgroundsigmaImagePtr => sigmaImagePtr    

end subroutine setBackgroundSigmaImagePtr



!********************************************************
! Subroutine DetectPointSources
! This routine is intended to be the first
! search for point-sources- hence maskimage
! is initialised to true so that all pixels
! will be checked.
!
!
!********************************************************
subroutine DetectPointSources(nx,                   &  ! x image dimension
                              ny,                   &  ! y - image dimesnion
                              image,                &  ! The image
                              smoothedImage,        &  ! The smoothed image
                              backGroundImage,      &  ! The background image
                              backgroundSigmaImage, &  !
                              levelImage,           &  ! The level-image map
                              qualityImage,         &
                              nSigma,               &
                              exposure,             &
                              sourceList,           &  ! The linked-list source-list
                              noSources,            &   ! The number of point-sources detected
                              fastMode)                ! the fast-mode flag: if true the task is
                                                       ! working with the small fast window

    integer(int16), intent(in) :: nx, ny
    real(single), pointer, dimension(:, :) :: image, smoothedImage
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)


    integer(int16), pointer :: levelImage(:, :)
    logical(boolean), pointer   :: qualityImage(:, :)

    real(single), intent(in) :: nSigma, exposure
    type(SourceLinkedList), intent(inout):: sourceList  ! The linked-list source-list
    integer(int16), intent(out) :: noSources  
    integer(int16) :: i, j
    real(single) :: height
    logical, intent(in):: fastMode

    noSources = 0
    write(errstr,'(a)') "Looking for point sources"
    call message(AppMsg, Verbose)

  
    detectParams%exposure = exposure

    detectParams%minNoPixels = 1

 
    ! Initialise some global variables
    ! Need nstd=0.1 to detect extreme strong mod-8 sources

 
    call GetPSFWidths(psfsemiA1, psfsemiB1)

    psfParams%psfSemiA = psfSemiA1   
    psfParams%psfSemiB = psfSemiB1 

    lbx = lbound(image, 1)
    lby = lbound(image, 2)
    
    nx1 = ubound(image, 1)
    ny1 = ubound(image, 2)
 
 
    !******************************************************************
    !
    ! Find very bright sources (badly affected by mod-8 sources)  
    ! NB This bit of code has to be here so that maskimage is allocated
    !
    !******************************************************************
   

    !write(*,*) "L453................ fastMode=", fastMode
    call InitialiseDetection(nx, ny, 1.0, 2, 0.1, 0.1, 1.5) ! was 1.5

    where(image .lt. 0.0 .or. image .lt. (backgroundImage - 5.0 * backgroundSigmaImage)) 
        maskImage = .true.
    elsewhere 
        emaskImage = .false.
        maskImage = .false. ! Test all pixels for a possible source
    end where 

    
    where(qualityImage)
        emaskImage = .true.
 !       maskImage = .true. ! Don't test these pixels for possible sources
    end where


    !***************************************************************
    !
    ! Get rid of nans which can cause serious trouble
    !
    !************************************************************** 
    do j = lby, ny1
        do i = lbx, nx1
            if( qualityImage(i, j)) then
                if ( isNullValue(image(i, j))) then 
                    image(i, j) = 0.0
                    emaskImage(i, j) = .true.
                    maskImage(i, j) = .true.
                end if
            end if
        end do
    end do  

    detectParams%minNoPixels = 1
    !write(*,*) "L488................"
    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               2000.0,               &  ! was 2000  
                               IsPointSourceMod8,    & 
                               HeightFunction2,      &
                               fastMode)             ! if true the subroutine is working 
                                                     ! with the small Fast window


    detectParams%minNoPixels = 1

    !write(*,*) "L503................"
    ! Find other very bright sources (near mod-8 sources)
    call InitialiseDetection(nx, ny, 1.0, 3, 2.0, 2.0, 1.5) 




    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &                          
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               2000.0,               & 
                               IsPointSource,        &
                               HeightFunction2,      &
                               fastMode)


    detectParams%minNoPixels = 1
    !write(*,*) "L523................"
    ! Find other very bright sources (near mod-8 sources)
    call InitialiseDetection(nx, ny, 1.0, 3, 2.0, 2.0, 1.5) 




    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &                          
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               50.0,                 & 
                               IsPointSource,        &
                               HeightFunction2,      &
                               fastMode)



    !****************************************************************************
    ! For removing extended light round bright sources (mod-8 patterns, etc)
    !****************************************************************************
    call BrightSourceMask(noSources, 1_int16, sourceList, smoothedImage, backgroundimage, .true.)



    detectParams%minNoPixels = 4 / binx
 

    !write(*,*) "L553................"
    call InitialiseDetection(nx, ny, 0.8, 3, 2.0, 2.0, 1.5) 
    call SearchForPointSources(image,                &
                               smoothedImage,        &
                               backgroundImage,      &
                               backgroundSigmaImage, &
                               levelImage,           &
                               sourceList,           &
                               noSources,            &
                               50.0,                 & ! The height above background
                               IsPointSource2,       &
                               HeightFunction2,      &
                               fastMode) 


    !*************************************************************
    ! 
    !*************************************************************                         
    call InitialiseDetection(nx, ny, 1.0, 2, 3.0, 1.0, 1.1) 

    

    height = max(5.0, 5.0 + nSigma - 2.0)

    !write(*,*) "L576................"
    call SearchForPointSources(image,                  &
                               smoothedImage,          & 
                               backgroundImage,        &  
                               backgroundsigmaImage,   &
                               levelImage,             &
                               sourceList,             & 
                               noSources,              &
                               height,                 & ! The height above the background
                               IsPointSource3,         &  
                               HeightFunction1,        &
                               fastMode)           
         


 

    detectParams%minNoPixels = 4 / binx - 1
 


    call InitialiseDetection(nx, ny, 1.0, 2, 4.0, 1.0, 1.1)  



    height = max(3.0, 3.0 + nSigma - 2.0) ! was 3


    !write(*,*) "L603................"
    call SearchForPointSources(image,                  &
                               smoothedImage,          & 
                               backgroundImage,        &  
                               backgroundsigmaImage,   &
                               levelImage,             &
                               sourceList,             & 
                               noSources,              &
                               height,                 & ! The height above the background
                               IsPointSource33,        & 
                               HeightFunction1,        &
                               fastMode) 

    !****************************************************************************
    !
    ! For removing extended light round bright sources (mod-8 patterns, etc)
    !
    !****************************************************************************
    call BrightSourceMask(noSources, 1_int16, sourceList, smoothedImage, backgroundimage, .false.)

    !write(*,*) "L622................"
!call saveimage( levelimage, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 641 displaying levelimage.fits")


    where(levelImage < 0) levelImage = 0

end subroutine DetectPointSources

!*********************************************
! Deallocates various arrays used in the
! detection routines
!
!********************************************
subroutine Delete
    if (associated(maskImage)) then
        deallocate(maskImage, emaskimage)
        nullify(maskImage, emaskimage)


        deallocate(rCosTheta, rSinTheta, fCircle, mean, sigma)
        nullify(rCosTheta, rSinTheta, fCircle, mean, sigma)
    end if

    if (associated(psfParams%psfImage)) then
        deallocate(psfParams%psfImage)
        nullify(psfParams%psfImage)
    end if

    if (associated(imagePtr)) then
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)
    end if

    if (associated(backgroundsigmaimagePtr)) then
        nullify(backgroundsigmaimagePtr)
    end if


end subroutine Delete   


!***********************************************************
! 
!
!
!***********************************************************
subroutine SearchForPointSources(&
           image,                &  ! The image
           smoothedImage,        &  ! The smoothed image
           backgroundImage,      &  ! The background image
           backgroundsigmaImage, &  ! The background sigma image
           levelImage,           &  ! The level-image image
           sourceList,           &  ! The linked-list source-list
           noSources,            &  ! The total number of detected point-sources
           heightAboveBck,       &  ! The minimum pixel-height above the background to be checked
           IsValidPointSource,   &   ! The function to be used
           HeightFunction,       &
           fastMode)       ! logical flag for Fast-mode detection

    real(single), pointer, dimension(:, :) :: image, smoothedImage, backgroundImage, backgroundsigmaimage
    type(SourceLinkedList), intent(inout):: sourceList
    integer(int16), intent(inout) :: noSources
    real(single), intent(in) :: heightAboveBck
    integer(int32) :: k, npixels, count
    logical, intent(in) :: fastMode

    interface
        Function IsValidPointSource(i0, j0, sourceP) result(d)
        use dal
        use OmSourceParam_mod    ! Defines structures for storing source parameters
            integer(int16), intent(inout) :: i0, j0    
            type(SourceParameters) :: sourceP
            logical :: d 
        end function IsValidPointSource
    end interface

  
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    type(SourceParameters) :: data
       
    integer(int16), pointer :: levelImage(:, :)     
    integer(int16) :: i0, j0
    integer(int8) :: dx, dx2
    integer(int16) :: count1, count2, nl, minCount1     

    integer(int16), pointer, dimension(:) :: iIndex => null(), jIndex => null()
    integer(int16) :: i, j, l, m
    logical(boolean), pointer :: tempMaskImage(:, :) => null()
    integer(int16) :: nx, ny, imin, imax, jmin, jmax
    real(single) :: fMax, eConstant   

    fMax = detectParams%exposure * 50.0
    eConstant = detectParams%exposure * 5.0  ! 5000 if exposure is 1000


    nx = nx1 + 1
    ny = ny1 + 1

    imagePtr => image    
    smoothedImagePtr => smoothedImage    

    backgroundImagePtr => backgroundImage    
    backgroundsigmaImagePtr => backgroundsigmaImage    
    plevelImagePtr => levelImage    
 

    !write(*, *) "L735 Search for point sources, fastMode=", fastMode


    allocate(tempMaskImage(0 : nx1, 0 : ny1))


 
    dx = 6 / binx  
    dx2 = 4 / binx
  
    if(minCount < 0_int8) then
        minCount = 4 / binx   
    end if
    minCount1 = 6 / binx   

    !write(*,*) "L750 minCount=", minCount, " minCount1=", minCount1
    !write(*,*) "L751 nx1=", nx1, " ny1=", ny1
    !write(*,*) "L752 lbx=", lbx, " lby=", lby
!mincount1= 4 / binx
!mincount1 = 2
!mincount = 2 

    do j = lby, ny1
        do i = lbx, nx1
            if (maskImage(i, j) .or. levelImage(i, j) < 0 .or. isNullValue(image(i, j)) ) then
                tempMaskImage(i, j) = .true.
                !write(*,*) "l761 i=", i, " j=", j
                cycle
            end if

            count = 0 ! Count how many pixels above the background
            count2 = 0
            do l = max(0_int16, j - 1_int16), min(ny1, j + 1_int16) 
                do m = max(0_int16, i - 1_int16), min(nx1, i + 1_int16) 

                    if ( image(m, l) .ge. heightfunction(m, l, heightAboveBck, backgroundimage)) then
                        count = count + 1

                        if(image(m, l) > fMax) then
                            count2 = count2 + 1 
                        end if 
                    end if
                 end do
            end do 
            !write(*,*) "L779 count=", count, " count2=", count2, " minCount=", minCount

            if (count .lt. minCount) then
                tempMaskImage(i, j) = .true.
            else         
                if(count2 .eq. 0) then
                    tempMaskImage(i, j) = .false.
                else
                    if(count2 < minCount1) then

                        tempMaskImage(i, j) = .true.
                    else
                        tempMaskImage(i, j) = .false.
                 

                    end if
                end if  
            end if
         end do
    end do

    !write(*, *) "L801 Search for point sources"


!write(*, *) "Line 797 mask image ipixel, jpixel=", maskimage(ipixel, jpixel)
!write(*, *) "Line 798 temp mask image ipixel, jpixel=", tempmaskimage(ipixel, jpixel)
!write(*, *) "Line 675 No of none true tempmaskimage=", count(.not. tempMaskImage) 
!write(*, *) "Line 675 No of none true maskimage=", count(.not. maskImage) 


    ! Create the 1-d arrarys arrayIndex, iIndex and jIndex
    ! From the 2-d arrays image and maskImage
    call SetUpIndices(image, tempMaskImage, iIndex, jIndex, nPixels)


!write(*,*) "L814 iIndex=", iIndex
!write(*, *) "Search for point sources Line 789 npixels=", npixels



!write(*, *) "Line 684 nPixels=", nPixels
!write(*, *) "Line 687 No of none true tempmaskimage=", count(.not. tempMaskImage) 
!write(*, *) "Line 687 No of none true maskimage=", count(.not. maskImage) 
!write(*, *) "Line 713 tempmaskimage ", ipixel, ", ", jpixel, " =", tempmaskimage(ipixel, jpixel)



    deallocate(tempMaskImage)
    nullify(tempMaskImage)

    nl = max( 1_int16, noSources )
  
    sourceNumber = maxval(levelImage)

    !write(*,*) "L834 sourceNumber=", sourceNumber
    !*********************************************************    
    ! Set up the mask so that where any image pixel
    ! where the mask image pixel is false
    ! the image pixel is skipped
    !*********************************************************





!*****************************************************************
! NB FOR TESTING TO SEE WHY A PARTICULAR SOURCE IS REJECTED
! SET MASK IMAGE TO PIXEL COORDINATES OF MAXIMUM SOURCE PIXEL
!****************************************************************

    count1 = 10




  
    !******************************************************************
    ! Loop through all the pixels in the list
    !****************************************************************

    do k = 1, nPixels

        ! Detect sources in order of decreasing maximum pixel value    
        ! Get the location (i, j) of the maximum pixel- excluding
        ! those pixels where maskImage = true

        i0 = iIndex(k)
        j0 = jIndex(k)
        

        !if (.not. maskImage(i0,j0) .and. levelimage(i0, j0) .eq. 0) then
        !   write(*, *) "L883 i0=", i0, " j0=", j0, " pixel=", image(i0, j0)
        !   write(*,*) "L884 maskImage=",maskImage(i0, j0), "levelImage=", &
        !        levelimage(i0, j0) 
        !endif


        if (maskImage(i0, j0) .or. levelimage(i0, j0) .ne. 0) then
!            if (checkPixel) then
!                write(*, *) "Line 829 pixel ", ipixel, ", ", jpixel, " is true"
!                write(*, *) "levelimage ipixel, jpixel =", levelimage(iPixel, jPixel)  
!             end if  
            cycle        
        end if
      
 

        !************************************************** 
        ! See if this pixel is centred on a source      
        !*************************************************
        !write(*,*) "L903 IsValidPointSource(i0,j0,data)=", IsValidPointSource(i0, j0, data)

        !if (IsValidPointSource(i0, j0, data)) then ! Possible new source
        ! 2010-04-02: In the case of the fast mode accept the first (brightest) pixel 
        ! as the source disregarding the returned value of IsValidPointSource 
        if (IsValidPointSource(i0, j0, data) .or. fastMode .and. sourceNumber .eq. 0) then ! Possible new source
           !write(*,*) "L904 possible new source"
           sourceNumber = sourceNumber + 1 ! The new source number

           data%rawCountRate = -image(i0, j0)

           if (mod(sourceNumber, count1) == 0) then
              !write(*,*) "L911"
              write(errstr,'(a, i0, a, f0.1)') &
                   "Number detected so far=", sourceNumber, &
                   " pixel level=", image(i0, j0)
              call message(AppMsg, Verbose)
              if (sourceNumber > 99) count1 = 100
           end if

           !**************************************************
           !
           ! Initialise some source parameters
           !
           !*************************************************
           data%extensionFlag = 0
           data%qualityFlag = 0
           data%confusionFlag = 0
           data%number = sourceNumber
           data%noSourcePixels = 0


           !*****************************************************************************
           ! In order to be able to discriminate between a "point" and "extended" source
           ! we need to "refine" the source pixel map
           !*****************************************************************************
           !write(*, *) "L937 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)

           call RefineSourceMap5(data, image, backgroundImage, backgroundsigmaImage, &
                levelImage, 2.0, detectparams%exposure, maskImage) 
           !write(*, *) "L941 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)




           !**************************************************************
           ! Check for hot pixels
           !*************************************************************            
           if (.not. checkforbadPixels(data, i0, j0)) then

              !*****************************************************
              ! Check for a strong mod-8 pattern
              !****************************************************

              if (IsMod8Source( data, image, imin, imax, jmin, jmax ) ) then

                 data%extensionFlag = 0
                 call sourceCircle( data, image, backgroundimage )
 
              end if
           end if

           if (image(i0, j0) .gt. eConstant .or. btest(data%qualityFlag, 4) ) then  ! Was 1000

              call ComputePointSourceParameters(sourceNumber, nx1, ny1, &
                   image, backgroundImage, i0, j0, data, dx2, &
                   computemonteerrors=.true., fastMode=fastMode)


              !write(*, *) "L972 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)

              call RefineSourceMap3(data, image, levelImage, maskImage) 
   
              !write(*, *) "L976 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)

              data%extensionFlag = 0
            

              if (btest(data%qualityFlag, 4) ) then

                 data%imin = imin
                 data%imax = imax
                 data%jmin = jmin
                 data%jmax = jmax
                 data%noSourcePixels = 0 

              end if

           else
 

              call ComputeExtendedSrcParameters2(int(data%number, int16), nx1, ny1, image, backgroundImage, &
                   levelImage, data, computemonteerrors = .true.)  ! True for Monte-carlo errors
 

              !**********************************************************************
              !
              ! Classify the source - point (0) or extended (1)
              !
              !********************************************************************* 
              !write(*, *) "L1001 Calling classify source "
              call ClassifySource(data, image, backgroundImage)                   




              if(data%extensionFlag == 0) then ! Point-source

                 call ComputePointSourceParameters(sourceNumber, nx1, ny1, &
                      image, backgroundImage, i0, j0, data, dx2, &
                      computemonteerrors=.true., fastMode=fastMode)

                 !write(*, *) "L1015 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)


                 call RefineSourceMap3(data, image, levelImage, maskImage) 
         

                 !write(*, *) "L1021 i0=", i0, ", j0=", j0, " maskImage=", maskImage(i0, j0)
                 !write(*,*) "L1022 nx1=", nx1, " ny1=", ny1
                 ! In the case of fast detection mask nearby pixels
                 if (fastMode) then
                    do i=i0-2,i0+2
                       if (i .ge. 0 .and. i .lt. nx1) then
                          do j=j0-2,j0+2
                             if (j .ge. 0 .and. j .lt. ny1) then
                                maskImage(i,j)=.true.
                             endif
                          enddo
                       endif
                    enddo
                 endif

                 data%noSourcePixels = 0 

!                else  ! Extended source
!                    if( abs(data%xpos - real(i0) ) > 6 / binx .or. &
!                        abs(data%ypos - real(j0) ) > 6 / binx) then
!                        where(levelImage(data%imin : data%imax, data%jmin : data%jmax) .eq. &
!                            sourceNumber)
!                            levelImage(data%imin : data%imax, data%jmin : data%jmax) = 0
!                        end where
!                        sourceNumber = sourceNumber - 1 ! The new source number
!                        cycle
!                    end if


              end if

           end if

           !**************************************
           ! Add it to the source list            
           !*************************************    
           call AddSourceToList(sourceList, data)


!    if(checkPixel) then
!      write(*, *) "************************************************"
!      write(*, *) "source found- x=", data%xpos, " y=", data%ypos, " eflag=", data%extensionFlag
!      write(*, *) "source number=", data%number, " imin=", data%imin, " imax=", data%imax, &
!                 " jmin=", data%jmin, " jmax=", data%jmax
!      write(*, *) "i0=", i0, " j0=", j0          
!    end if    
        
        end if ! Isvalidpointsource


     end do    ! End of looping through the pixels
 

!write(*, *) "Line 1058 mask image ", ipixel, ", ", jpixel, " =", maskimage(ipixel, jpixel)


!write(*, *) "Search for point sources line 1054"


     noSources = maxval(levelImage)
     write(errstr,'(a, i0)') "Number of likely point-sources detected=", noSources
     call message(AppMsg, Verbose)
     
     deallocate(iIndex, jIndex)  
     nullify(iIndex, jIndex)  
    

     nullify(imagePtr, smoothedImagePtr, backgroundImagePtr, plevelImagePtr)    

!write(*, *) "Search for point sources line 1067"


     
   end subroutine SearchForPointSources
!*****************************************************************
!
!****************************************************************
function getKmin(i0, j0) result(kmin)
    integer(int16), intent(in) :: i0, j0
    integer(int8) :: kMin  

    kmin = detectParams%kMin
    if (i0 < 6 / binx .or. (nx1 - i0) < 6 / binx ) then
        kmin = kmin - 4
    end if

    if (j0 < 6 / binx .or. (ny1 - j0) < 6 / binx ) then
        kmin = kmin - 4
    end if
    kmin = max( 4_int8,  kmin)

    
 end function getKmin



!***********************************************************
! Function IsPointSource
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source
 

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k, kmask
    real(single) :: xc, yc
    integer(int8) :: kmin


!if (checkPixel) then
!write(*, *) "Line 1117 ispointsource Checking  pixel ", i0, ", ", j0
!end if


    kmin = getKmin(i0, j0)
  
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5
    
    !write(*,*) "L1133 ************** xc=", xc, " yc=", yc, " nCircles=", detectParams%nCircles
    !write(*,*) "L1135 xOffset(1)=", xOffset(1)
    do l = 1, 5

        xCentre = xc + xOffset(l)
         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1
            
                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) then
                        cycle

                    end if 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) then                   
                        cycle
                    end if 
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        


                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                            kmask = kmask + 1
                            cycle
                        end if   
                    end if  

                    k = k + 1 
                    fCircle(k) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do
 
                if (k .lt. kMin) then 


!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "Is point source Line 1215 k=", k, " rejecting source, kmin=", kmin, &
!   " nsize=", detectParams%nsize
!end if

                    cycle mLoop 
           
                end if   
 
!if(checkPixel) then
!write(*, *) "k=", k, " fcircle=", fCircle(1 : k)
!end if



                call computeStatistics3(k, fCircle, mean(i), sigma(i))
                !write(*,*) "L1198 i=", i, " mean=", mean(i), " sigma=", sigma(i)

                if (i == detectParams%nCircles) cycle

                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              
!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1581 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                    
                    cycle mLoop          
                end if

            end do

            !**********************************************************
            ! Found a possible point-source
            !*********************************************************
            source%xPos = xCentre + 0.5  ! Approximate x coordinate
            source%yPos = yCentre + 0.5  ! Approximate y coordinate
 
            IsPointSource = .true.

!if (checkPixel) then
!    write(*, *) "Source ", i0, ", ", j0, " has been found!"
!    write(*, *) "Image value=", imagePtr(i0, j0), " background =", backgroundImagePtr(i0, j0)
!    write(*, *) "mean semi a=", detectParams%meanSemiA
!    write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                     " y=", source%yPos, "+/-", source%PosErr
!    write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!    write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if

            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource = .false.
end Function IsPointSource

!***********************************************************
! Function IsPointSourceMod8
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSourceMod8(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source
   

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    integer(int8) :: kmin



    kmin = detectParams%kMin - 2
    if (i0 < 6 / binx .or. (nx1 - i0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if

    if (j0 < 6 / binx .or. (ny1 - j0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if








 

!if (checkPixel) then
!write(*, *) "Line 1304 ispointsourcemod8 Checking  pixel ", i0, ", ", j0
!end if
 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)
         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************

            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) cycle   
                    end if  

                    k = k + 1 

                    fCircle(k) = Interpolate(smoothedImagePtr, xPos, yPos)

    
                end do
                if (k .lt. kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre     
!write(*, *) "Line 1356 k=", k, " rejecting source, n=", detectParams%nsize, &
! " kmin=", detectParams%kmin
!end if

                    cycle mLoop 
           
                end if   
 
                ! Compute the mean pixel value and variance around the circle 



                call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))
              


                if (i == detectParams%nCircles) cycle

                if ((mean(i) - mean(i + 1)) <= &
                                max( 0.1 * mean(i + 1) ,             &
                                detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1)))) then  

              
!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1383 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                    
                    cycle mLoop                
                end if
            end do

            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
           
            IsPointSourceMod8 = .true.

!if (checkPixel) then
!write(*, *) "Source ", i0, ", ", j0, " has been found!"
!write(*, *) "Image value=", imagePtr(i0, j0), " background =", backgroundImagePtr(i0, j0)
!write(*, *) "mean semi a=", detectParams%meanSemiA
!write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if


            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSourceMod8 = .false.
end Function IsPointSourceMod8



!***********************************************************
! Function IsPointSource2
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource2(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
 
    real(single) :: fCentre  ! The centre pixel counts
    integer(int8) :: kmin



    kmin = detectParams%kMin
    if (i0 < 6 / binx .or. (nx1 - i0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if

    if (j0 < 6 / binx .or. (ny1 - j0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if



    if (.not. BackgroundGradientOkay(i0, j0)) then
        IsPointSource2 = .false.
        return
    end if


 
!if (checkPixel) then
!write(*, *) "Line 1640 ispointsource2  Checking  pixel ", i0, ", ", j0
!end if
 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5




    do l = 1, 5

        xCentre = xc + xOffset(l)
         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre


            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 
                    if ( imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (i .gt. 1) then
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) cycle   
                    end if  

                    k = k + 1 
                    fCircle(k) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do
                
                if ( any(fcircle(1 : k) .ge. fcentre)) then
                    cycle mloop
              
!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre      
!write(*, *) "Line 1696  rejecting source count(anycircle > ge fcentre=",  count(fcircle(1 : k) .ge. fcentre)
!end if
                end if

                if (k .lt. kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre      
!write(*, *) "Line 1704 k=", k, " rejecting source", " kmin=", kmin
!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean pixel value and variance around the circle 
                call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))



                if (i == detectParams%nCircles) cycle

 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource2 xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1341 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                 
                    cycle mLoop

                
           
                end if


            end do


            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
            !write(*,*) "L1552 xPos=", source%xPos, " yPos=", source%yPos

            IsPointSource2 = .true.

!if (checkPixel) then
!write(*, *) "Ispointsource2 Source ", i0, ", ", j0, " has been found!"
!write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if



            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource2 = .false.
end Function IsPointSource2


!***********************************************************
! Function IsPointSource3
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource3(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k, k1
    real(single) :: xc, yc
    real(single) :: fCentre  ! The centre pixel counts
    real(single) :: fMin, fMax, minPixRatio
    logical :: hasNeighbour
    integer(int8) :: kmin
    

!if (checkPixel) then 
!    write(*, *) "ispointsource3 line 1741 checking pixel ", ipixel, " , ", jpixel
!    write(*, *) "i0=", i0, " j0=", j0, " maskimage=", maskImage(i0, j0)
!end if
    if (.not. BackgroundGradientOkay(i0, j0)) then ! Was 0.1



!if(checkPixel) then
!    write(*, *) "Ispointsource3 REjecting source at line 1751"
!end if
        IsPointSource3 = .false.
        return
    end if





    kmin = getKmin(i0, j0)
 
    


!if(checkpixel) then
!write(*, *) "ispointsource3 Line 1774, kmin=", kmin, " detectParams%kmin=", detectParams%kmin
!end if

    minPixRatio = 0.5  
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)
            hasNeighbour = .false.
            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre

            fcentre = fcentre - 2.0 * sqrt(fcentre)

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if (imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        
                    if (i .gt. 1) then                          
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                            hasNeighbour = .true.
                            
                            cycle

                        end if   
                    end if

                    k = k + 1 
                    fCircle(k) = Interpolate(smoothedImagePtr, xPos, yPos)
            
                end do

                if (k .lt. kmin) then
  
!if (checkPixel) then
!write(*, *) "line 1837 i=", i, "k=", k, " kmin=", kmin
!end if 
                  cycle mLoop
                end if 
                if (hasNeighbour) then
                    minPixRatio = 0.6   ! WAS 0.6
                    ! Compute the mean pixel value and variance around the circle 

                    call heapsort(fCircle(1 : detectParams%nSize), k)


                    call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))


                    fMax = maxval(fcircle(1 : k - 2) )                 
                    if ( fmax .gt. 10.0 * fcentre) then
    

!if (checkPixel) then
!write(*, *) "line 1850 k=", k, " fcircle=", fcircle(1 : k)
!write(*, *) "ispointsource3 line 1694 rejecting source fmax>fcentre, fmax=", fmax, " fcentre=", fcentre
!write(*, *) "xcentre=", xcentre, " yc=", ycentre, "  fmax=", fmax, " fcentre=", fcentre
!end if
                        cycle mloop

                    end if


                    if (i .ne. 1) then
                       
                        k1 = k - 8 / binx
                        if(k1 < 8 / binx) cycle mLoop
!if (checkpixel) then
!write(*, *) "Line 1859 k=", k, " k1=", k1, " calling minimumlist"
!end if              
                        call MinimumList(k, fCircle(1 : detectParams%nSize), k1 ) ! was 4




                     

                    end if
                end if

!if (checkPixel) then
!write(*, *) "Line 1874 ispointsource3 hasneighbour=", hasneighbour, " minpixratio=", minpixratio, " k=", k
!end if




                ! Compute the mean and variance

                call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))


                fMax = maxval(fcircle(1 : k))                 
                if ( fmax .ge. fcentre) then
!if (checkPixel) then
!write(*, *) "line 1889 k=", k, " fcircle=", fcircle(1 : k)
!write(*, *) "ispointsource3 line 1694 rejecting source fmax>fcentre"
!write(*, *) "xcentre=", xcentre, " yc=", ycentre, "  fmax=", fmax, " fcentre=", fcentre
!end if
                    cycle mloop
                end if


                fmin = minval(fcircle(1 : k))                 
             
                if ( fMin / fMax .lt. minpixratio) then

!if (checkPixel) then
!write(*, *) "ispointsource3 line 1901 rejecting source"
!write(*, *) "pixel ratio=",  (fmin / fmax)
!write(*, *) "fcentre=", fcentre, " mean1=", mean(1), " std=", sqrt(sigma(1))
!end if

                    cycle mloop
                end if


                if (i == detectParams%nCircles) cycle
 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

             
!if (checkPixel) then
!write(*, *) "IsPointsource3 line 1917 xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1714 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                 
                    cycle mLoop

                           
                end if


            end do

      
            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5

    
            IsPointSource3 = .true.

!if (checkPixel) then
!write(*, *) "line 1941 ispointsource3 Source ", ipixel, ", ", jpixel, " has been found!"
!write(*, *) "i0=", i0, " j0=", j0
!write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if


            return
        end do mLoop
    end do


!if(checkpixel) then
!write(*, *) "ispointsource3 Line 1957, kmin=", kmin, " detectParams%kmin=", detectParams%kmin
!end if




    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource3 = .false.

end Function IsPointSource3


!***********************************************************
! Function IsPointSource33
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource33(i0, j0, source)
    integer(int16), intent(inout) :: i0, j0    
    type(SourceParameters) :: source


    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    real(single) :: minPixRatio, fcentre, fpixel
    integer(int8) :: kmin   




!if (checkPixel) then 
!write(*, *) "Line 1776 ispointsource33  Checking  pixel ", i0, ", ", j0
!write(*, *) "image=", imagePtr(i0, j0), " background=", backgroundImagePtr(i0, j0)
!write(*, *) "maskimage=", maskimage(i0, j0)
!end if

    if ( .not. BackgroundGradientOkay(i0, j0) ) then
        IsPointSource33 = .false.

!if (checkPixel) then 
!      write(*, *) "Rejecting source at line 1936"
!end if 
        return
    end if


    minPixRatio = 0.7


 
   
   kmin = getKmin(i0, j0)
 


    xc = real(i0) + 0.5
    yc = real(j0) + 0.5


    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)
          
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre

            fcentre = fcentre - 2.0 * sqrt(fcentre)
!if(checkPixel) then
!write(*, *) "l=", l, " m=", m, " xcentre=", xcentre, " ycentre=", ycentre, " fcentre=", fcentre
!end if

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1


                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                 
                    fpixel = Interpolate(smoothedImagePtr, xPos, yPos)

      
                    if( fpixel > 100.0 * fcentre) cycle mloop

                    if(i > detectParams%ncircles) cycle 



                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
                    if ( imagePtr(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16))) .lt. 0.0) cycle   
        

                    if (i .gt. 1) then                          
                        if (maskImage(min(nx1, nint(xPos, int16)), min(ny1, nint(yPos, int16)))) then
                          
                            
                            cycle

                        end if   
                    end if

                    k = k + 1 
                    fCircle(k) = fpixel
            
                end do

                if (k .lt. kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "Line ispointsource33 line 1870 k=", k, " rejecting source"
!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean and variance
                call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))

  
                if (i == detectParams%nCircles) cycle

 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource33 line 2108 xCentre=", xCentre, " yCentre=", yCentre       
!write(*, *) "line 2107 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=", detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                 
                    cycle mLoop           
                end if
 


!if (checkPixel) then
!write(*, *) "IsPointsource33 line 2108 xCentre=", xCentre, " yCentre=", yCentre       
!write(*, *) "line 1966 not rejecting point -mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=", detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
   



           end do

      
            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5

   
            IsPointSource33 = .true.

!if (checkPixel) then
!write(*, *) "line 2156 ispointsource33 Source ", ipixel, ", ", jpixel, " has been found!"
!write(*, *) "i0=", i0, " j0=", j0, " maskimage=", maskimage(i0, j0), " kmin=", kmin
!write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if



            return
        end do mLoop
    end do




    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource33 = .false.

end Function IsPointSource33

!*******************************************************
!
!
!
!*****************************************************
logical function BackgroundGradientOkay1(i0, j0)
    integer(int16), intent(in) :: i0, j0
    
   
    integer :: il, iu, jl, ju
 
    integer(int16) :: dx = 2
 
    real :: fmin, fmax, fmax1, mean, fraction, sourceCounts, sigNoise
    if (imagePtr(i0, j0) .ge. backgroundImagePtr(i0, j0) + 50.0) then
        BackgroundGradientOkay1 = .true.
        return
    end if

    fmax = imagePtr(i0, j0)
    dx = 14 / binx


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    fmax1 = maxval( imagePtr(il : iu, jl : ju))
    if (fmax1 .gt. fmax * 1000.0) then


      

        BackgroundGradientOkay1 = .false.
        return
    end if




    dx = 4 / binx

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    fmax = maxval(backgroundImagePtr(il : iu, jl : ju))
    if (fmax .lt. 0.1) then
        BackgroundGradientOkay1 = .true.
        return
    end if
 
    fmin = minval(backgroundImagePtr(il : iu, jl : ju), backgroundImagePtr(il : iu, jl : ju) .ge. 0.0)

    mean = (fmax + fmin) / 2.0
    sourceCounts = imagePtr(i0, j0) - mean
    sigNoise = sourceCounts / sqrt(imagePtr(i0, j0))

    fraction =  (fmax - fmin) / fmax
   

    select case (nint(signoise))
        case (:5)  ! All values below 5
            if ( fraction .gt. 0.25) then
                BackgroundGradientOkay1 = .false.

                return
            end if
        case (6 : 10) ! Values between 6 and 7
            if ( fraction .gt. 0.25) then
                BackgroundGradientOkay1 = .false.

                return
            end if
        
        case default
            BackgroundGradientOkay1 = .true.
            return
    end select
   
end function BackgroundGradientOkay1   
!*******************************************************
!
!*****************************************************
logical function BackgroundGradientOkay2(i0, j0, maxFrac)
    integer(int16), intent(in) :: i0, j0
    real(single), intent(in) :: maxFrac
   
    integer(int16) :: il, iu, jl, ju
 
    integer(int16), parameter :: dx = 2
    real :: fMin, fMax, mean, fraction, sourceCounts, sigNoise
! BackgroundGradientOkay2 = .true.
!return


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    fmax = maxval(backgroundImagePtr(il : iu, jl : ju))
    if (fmax .lt. 0.1) then
        BackgroundGradientOkay2 = .true.
        return
    end if
 
    fmin = minval(backgroundImagePtr(il : iu, jl : ju), backgroundImagePtr(il : iu, jl : ju) .ge. 0.0)

    mean = (fmax + fmin) / 2.0
    sourceCounts = imagePtr(i0, j0) - mean
    sigNoise = sourceCounts / sqrt(imagePtr(i0, j0))
    if (sigNoise .ge. 10.0) then

        BackgroundGradientOkay2 = .true.
        return

    end if


    fraction =  (fmax - fmin) / fmax
    if (fraction .gt. maxFrac) then

        BackgroundGradientOkay2 = .false.
    else
        BackgroundGradientOkay2 = .true.
    

    end if


   
end function BackgroundGradientOkay2   

!*******************************************************
!
! i0, j0, coordinates of maximum source pixel counts
!
!*****************************************************
logical function BackgroundGradientOkay3(i0, j0, pixel)
    integer(int16), intent(in) :: i0, j0
    type(pixelStruct) :: pixel(:)
    real(single) :: maxRadius
    real(single), parameter :: maxFrac = 0.4   
    integer(int16) :: n, i, j, k, il, iu, jl, ju
    integer(int16) :: dx = 2
    real(single) :: fSourceMax
    real(single) :: radius, fMin, fMax, mean, fraction

! BackgroundGradientOkay3 = .true.
!return

    do k = 1, 2  
        dx = 6 * k / binx ! was 6
        il = max(lbx, minval(pixel%x) - dx)
        iu = min(nx1, maxval(pixel%x) + dx)
        jl = max(lby, minval(pixel%y) - dx)
        ju = min(ny1, maxval(pixel%y) + dx)

        fSourceMax = imagePtr(i0, j0) * real(5 * k) ! was 1.5
    

        do j = jl, ju
            do i = il, iu                             
                if (elevelImageptr(i, j) .ne. number) then
                    if (imagePtr(i, j) .gt. fSourceMax) then
                        BackgroundGradientOkay3 = .false.

!if (abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jPixel) .lt. 4) then
!write(*, *) "Backgroundgradientokay1 line 1857 rejecting source at ", ipixel, ", ",  jpixel
!write(*, *) "fsourcemax=", fsourcemax, " image=", imagePtr(i, j), " i0=", &
!i0, " j0=", j0, " i=", i, " j=", j
!end if

                        return
                    end if
                end if
            end do
        end do
    end do


    n = size(pixel)
   

    maxRadius = real(16 / binx)

    dx = ceiling(maxradius) + 1
    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)

   
    fmax = 0.0
    fmin = 1e6
    do j = jl, ju
        do i = il, iu
            radius = sqrt(real(i - i0) ** 2 + (j - j0) ** 2)
            if (radius .le. maxRadius) then
                 
                fmax = max(fmax, backgroundImagePtr(i, j))
                fmin = min(fmin, backgroundImagePtr(i, j))
                 
                if (elevelImageptr(i, j) .eq. number .and. &
                       imagePtr(i, j) .gt. fSourceMax) then
                    BackgroundGradientOkay3 = .false.
!if (abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jPixel) .lt. 4) &
!write(*, *) "Backgroundgradientokay1 line 1897 rejecting source at ", ipixel, ", ",  jpixel


                    return
                end if


            end if
        end do
    end do

    if (fmax .lt. 0.1) then
        BackgroundGradientOkay3 = .true.
        return
    end if

    fmin = max(0.2, fmin)  

    mean = (fmax + fmin) / 2.0



    fraction =  (fmax - fmin) / fmax
    if (fraction .gt. maxFrac) then

!if ( abs(i0 - ipixel) .lt. 4 .and. abs(j0 - jpixel) .lt. 4) then
!write(*, *) "Line 2056 backgroundgradientokay fraction=", fraction, " maxfrac=", maxfrac
!write(*, *) "fmin =", fmin, " fma=", fmax
!end if


        BackgroundGradientOkay3 = .false.
    else
        BackgroundGradientOkay3 = .true.
    

    end if


   
end function BackgroundGradientOkay3   

!*******************************************************
! WidthOkay
! Checks to see that the width of the source
! in the middle is greater than that at the ends
!*****************************************************
logical function WidthOkay(pixel)
    type(pixelStruct) :: pixel(:)

    integer(int16) :: xc, yc, xmin, xmax, ymin, ymax, dx, dy, dx1, dy1
!    integer :: i
   

    xmin = minval(pixel%x)
    xmax = maxval(pixel%x)

!    checkPixel = .false.
!widthOkay = .true.
!return


!if (any(pixel%x .eq. ipixel .and. pixel%y .eq. jpixel)) then 
!checkPixel = .true.
! endif

    if ( (xmax - xmin) .gt. 8) then

!if (checkPixel) write(*, *) "Line 1973"
        xc = (xmin + xmax) / 2              ! Centre x


        ymin = minval(pixel%y, pixel%x .eq. xc)
        ymax = maxval(pixel%y, pixel%x .eq. xc)
        dy = ymax - ymin                    ! Vertical width at centre
        
    
        ymin = minval(pixel%y, pixel%x .lt. xmin + 1 .or. pixel%x .gt. xmax - 1)  
      
        ymax = maxval(pixel%y, pixel%x .lt. xmin + 1 .or. pixel%x .gt. xmax - 1)  
      
   

  
        dy1 = ymax - ymin                   ! Vertical width at left-hand edge

        if (dy1 .gt. dy * 3) then ! was 2
!if (checkPixel) write(*, *) "Line 1892"

            widthOkay = .false.
            return
        end if
    

    end if

    ymin = minval(pixel%y)
    ymax = maxval(pixel%y)
!    if (checkPixel) write(*, *) "Line 1904 ymin=", ymin, " ymax=", ymax

    if ( (ymax - ymin) .gt. 8) then
        yc = (ymin + ymax) / 2              ! Centre y
        xmin = minval(pixel%x, pixel%y .eq. yc)
        xmax = maxval(pixel%x, pixel%y .eq. yc)
        dx = xmax - xmin                    ! Horizontal width at centre


        xmin = minval(pixel%x, pixel%y .lt. ymin + 1 .or. pixel%y .gt. ymax - 1)  
        xmax = maxval(pixel%x, pixel%y .lt. ymin + 1 .or. pixel%y .gt. ymax - 1)  
      
   
        dx1 = xmax - xmin                   ! Horizontal width at left-hand edge


        if (dx1 .gt. dx * 3) then ! was 2
            widthOkay = .false.
            widthOkay = .false.
            return
        end if   
        

    end if

    widthOkay = .true.

end function WidthOkay
!**********************************************
!
!**********************************************
subroutine SetMaskPixels(pixel)
    type(pixelStruct) :: pixel(:)

    integer(int16) :: i

    if (associated(eMaskImage)) then
    do i = 1, size(pixel)

        eMaskImage(pixel(i)%x, pixel(i)%y) = .true.
    end do


    end if

    if (associated(MaskImage)) then
    do i = 1, size(pixel)

        MaskImage(pixel(i)%x, pixel(i)%y) = .true.
    end do


    end if


end subroutine SetMaskPixels

!***********************************************************
! Function IsPointSource22
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!***********************************************************
logical Function IsPointSource22(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
    
    type(SourceParameters) :: source

    !************************************
    ! Local variables
    !************************************  
    real(single) :: xCentre, yCentre, xPos, yPos
    integer(int16) :: i, j, l, m
    integer(int32) :: k
    real(single) :: xc, yc
    real(single) :: fCentre
    integer(int8) :: kmin


    kmin = detectParams%kMin
    if (i0 < 6 / binx .or. (nx1 - i0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if

    if (j0 < 6 / binx .or. (ny1 - j0) < 6 / binx ) then
        kmin = kmin - 4 / binx
    end if



!if (checkPixel) then
!write(*, *) "Line ispointsource22 line 1946  Checking  pixel ", i0, ", ", j0
!end if


    if (.not. BackgroundGradientOkay(i0, j0)) then
        IsPointSource22 = .false.
        return
    end if

 
    xc = real(i0) + 0.5
    yc = real(j0) + 0.5




    do l = 1, 5

        xCentre = xc + xOffset(l)

         
mLoop : do m = 1, 5 
            yCentre = yc + yOffset(m)

            !*************************************************************************
            ! Counts at the centre
            !************************************************************************
            fcentre = Interpolate(smoothedImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre
          
          

            !****************************************************
            ! Determine mean pixel value around source centre
            ! at 3 different radii
            !****************************************************
            do i = detectParams%nCircles, 1, -1

                fCircle = 0.0
                k = 0  

                do j = 1, detectParams%nSize
                    xPos = xCentre + rCosTheta(j, i)
                    if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
                    yPos = yCentre + rSinTheta(j, i)
                
                    if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle 

                    ! Check to see if a star has been found nearby- exclude pixels near to it                       
  

                    k = k + 1 
                    fCircle(k) = Interpolate(smoothedImagePtr, xPos, yPos)
                 end do

            
                if (k .lt. kMin) then 

!if (checkPixel) then
!write(*, *) "xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "Line 496 k=", k, " rejecting source"
!end if

                    cycle mLoop 
           
                end if   
 

                ! Compute the mean pixel value and variance around the circle 


                call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))
!                if (i .lt. 3) then
!                    fBackgroundCentre = Interpolate(backgroundImagePtr, xCentre, yCentre)  ! Pixel counts at the very centre
!                    if (mean(i) .lt. ( fBackgroundCentre + max(1.0, 0.5 * sqrt(fBackgroundCentre)))) cycle mloop
!                end if

                if (i == detectParams%nCircles) cycle


 
                if ((mean(i) - mean(i + 1)) <= &
                    detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))) then  

              

!if (checkPixel) then
!write(*, *) "IsPointsource22 xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "line 1341 rejecting point-mean(", (i+1), ")=", mean(i+1),  " mean(", i, ")=", mean(i)
!write(*, *) "sigma(", (i+1), ")=", sigma(i+1),  " sigma(", i, ")=", sigma(i)
!write(*, *) " sqrt(sigma(i) + sigma(i + 1)))=",  sqrt(sigma(i) + sigma(i + 1)) 
!write(*, *) "(mean(i) - mean(i + 1)) =", (mean(i) - mean(i + 1))
!write(*, *) "  nstd * sqrt(sigma(i) + sigma(i + 1)))=",   detectParams%nstd(i) * sqrt(sigma(i) + sigma(i + 1))                    
!end if  
                 
                    cycle mLoop

                
           
                end if


            end do

         
   
           if (fcentre - mean(1) .lt. detectparams%nstd(1) * sqrt(sigma(1))) then

!if (checkPixel) then
!write(*, *) "IsPointsource22 line 2070 xCentre=", xCentre, " yCentre=", yCentre
!write(*, *) "Rejecting source - fcentre=", fcentre, " mean(1)=", mean(1), " sigma(1)=", sqrt(sigma(1))
!end if



               cycle mLoop
           
            end if

            source%xPos = xCentre + 0.5
            source%yPos = yCentre + 0.5
            IsPointSource22 = .true.



!if (checkPixel) then
!write(*, *) "Source ispointsource22 ", i0, ", ", j0, " has been found!"
!write(*, *) "New source x=", source%xPos, "+/-", source%PosErr, &
!                       " y=", source%yPos, "+/-", source%PosErr
!write(*, *) "a=", source%semiA, "+/-", source%semiAErr, &
!                        " b=", source%semiB, "+/-", source%semiBErr
!write(*, *) "theta=", source%theta, "+/-", source%thetaErr
!end if





            return
        end do mLoop
    end do

    !**************************************************
    ! Reach here then not a point-source
    !************************************************* 
    isPointSource22 = .false.
end Function IsPointSource22

!***********************************************************
! Function IsPointSource4
! For finding possible sources with just 1 or 2 bright pixels
! Determines if the underlying pixels on the image
! constitute a point-source (for bright ones)
! return TRUE is so, FALSE if not
!
!***********************************************************
logical Function IsPointSource4(i0, j0, source)
  
    integer(int16), intent(inout) :: i0, j0
  
    
    type(SourceParameters) :: source
    
    integer(int16) :: il, iu, jl, ju, i, j
    integer(int8) :: dx
    real(double) :: imageCounts, backgroundCounts, sigNoise
    real(single) :: fMax, minImage, fBackground


    IsPointSource4 = .false.
     

!if(checkpixel) then
!write(*, *) "ispointsource4 checking pixel ", i0, ", ", j0
!if(associated(elevelImagePtr)) then
!write(*, *) "elevelimageptr=", elevelimagePtr(i0, j0)
!end if
!end if
 

    fMax = imagePtr(i0, j0) ! was times 2
    dx = 4 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

 
    if (any(imagePtr(il : iu, jl : ju) .gt. fMax)) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 2591- i0=", i0, " j0=", j0, &
!         " fmax=", fmax, " image=", maxval(imagePtr(il : iu, jl : ju))
!write(*, *) "Ratio =",   maxval(imagePtr(il : iu, jl : ju)) / fMax
!end if
        IsPointSource4 = .false.
        return


    end if  

    fMax = imagePtr(i0, j0) * 5.0

    dx = 6 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

  

    if (any(imagePtr(il : iu, jl : ju) .gt. fMax)) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 2591- i0=", i0, " j0=", j0, &
!         " fmax=", fmax, " image=", maxval(imagePtr(il : iu, jl : ju))
!write(*, *) "Ratio =",   maxval(imagePtr(il : iu, jl : ju)) / fMax
!end if
        IsPointSource4 = .false.
        return


    end if  
 
    fMax = imagePtr(i0, j0) * 10.0

    dx = 8 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

    if (any(imagePtr(il : iu, jl : ju) .gt. fMax)) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 2618- i0=", i0, " j0=", j0, &
!         " fmax=", fmax, " image=", maxval(imagePtr(il : iu, jl : ju))
!write(*, *) "Ratio =",   maxval(imagePtr(il : iu, jl : ju)) / fMax
!end if
        IsPointSource4 = .false.
        return


    end if  

    fMax = imagePtr(i0, j0) * 25.0

    dx = 12 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

    if (any(imagePtr(il : iu, jl : ju) .gt. fMax)) then


!if (checkPixel) then
!write(*, *) "rejecting source at line 2647- i0=", i0, " j0=", j0, &
! " fmax=", fmax, " image=", maxval(imagePtr(il : iu, jl : ju)) 

!do i = il, iu
!    do j = jl, ju
!        write(*, *) "image(", i, ", ", j, ")=", imagePtr(i, j)
!    end do
!end do
!end if

        IsPointSource4 = .false.

        return

    end if 

 
    fMax = imagePtr(i0, j0) * 50.0

    dx = 20 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

    if (any(imagePtr(il : iu, jl : ju) .gt. fMax)) then


!if (checkPixel) then
!write(*, *) "rejecting source at line 2673- i0=", i0, " j0=", j0, " fmax=", fmax 
!end if

        IsPointSource4 = .false.

        return

    end if  

    fMax = max(2000.0, imagePtr(i0, j0) * 100.0)
    dx = 40 / binx    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)

    if ( any(imagePtr(il : iu, jl : ju) .gt. fMax)) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 2697- i0=", i0, " j0=", j0, " fmax=", fmax 
!end if

        IsPointSource4 = .false.
        return

    end if  


    imageCounts = imagePtr(i0, j0)
    backgroundCounts = backgroundImagePtr(i0, j0)

    fBackground = backgroundImagePtr(i0, j0) ! The background counts at i0, j0


    fBackground = fBackground + max(2.0, sqrt(fBackground) )


    minImage = imagePtr(i0, j0)
  
    dx = 1    
 
    il = max(0_int16, i0 - dx)
    jl = max(0_int16, j0 - dx)

    iu = min(nx1, i0 + dx)
    ju = min(ny1, j0 + dx)



    do j = jl, ju
        do i = il, iu
            if(i .eq. i0 .and. j .eq. j0) cycle

!if(checkPixel) then
!write(*, *) "i=", i, " j=", j
!write(*, *) "image=", imagePtr(i, j), " background=", backgroundImagePtr(i, j)
!end if

            if ( imagePtr(i, j) .gt. imagePtr(i0, j0)  ) then
                IsPointSource4 = .false.

!if (checkPixel) then
!write(*, *) "rejecting source at line 2758"
!write(*, *) "i0=", i0, " j0=", j0, " image=", imagePtr(i0, j0)
!write(*, *) "i=", i, " j=", j, " image=", imagePtr(i, j)
!end if

                return
            end if  

            minImage = min(minImage, imagePtr(i, j)) 

            imageCounts = imageCounts + dble(imagePtr(i, j))
            backgroundCounts = backgroundCounts + dble(backgroundimagePtr(i, j))    
        end do
    end do



    fBackground = backgroundImagePtr(i0, j0) ! The background counts at i0, j0

    if ( minImage .lt. (fBackground  + max(2.0, 0.5 * sqrt(fBackground) )) ) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 2803- minimage=", minimage, " i0=", i0, " j0=", j0, &
!      " fbackground=", fBackground
!end if


        IsPointSource4 = .false.
        return
    

    end if




    if ( minImage .gt. (imagePtr(i0, j0) - max(3.0, sqrt(imagePtr(i0, j0)))) ) then ! was 2.0 * 

!if (checkPixel) then
!write(*, *) "rejecting source at line 2849 minimage=", minimage, " image(i0,j0)=", imagePtr(i0, j0)
!end if

        IsPointSource4 = .false.
        return
    

    end if

           
    if(imageCounts .gt. 0.0) then
        sigNoise = (imageCounts - backgroundCounts) / sqrt(imageCounts)
    else
        sigNoise = 0.0
    end if

    if (sigNoise < detectParams%minSigNoise) then


!if (checkPixel) then
!write(*, *) "rejecting source at line 2845- signoise=", signoise 
!end if



        IsPointSource4 = .false.
        return
    

    end if


    !**********************************************************
    !
    ! Found a possible point-source
    !
    !*********************************************************
    source%xPos = real(i0 + 1)  ! Approximate x coordinate
    source%yPos = real(j0 + 1)  ! Approximate y coordinate
 

    IsPointSource4 = .true.


!if( checkpixel ) then
!write(*, *) "****************************************************************"
!write(*, *) "ispointsource4 found source at ",i0, " ,", j0
!write(*, *) "i0=", i0, " j0=", j0, " x=", source%xpos, " y=", source%ypos
!write(*, *) "****************************************************************"
!end if




!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage(plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 5353 displaying e & p levelimage.fits")


    

end Function IsPointSource4




!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction1(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: image

! HeightFunction1 = max(minHeight, image(i, j) + sqrt(image(i, j))) ! Friday was 7.0
!write(*, *) "heightfunction1 line 2903- minheight=", minheight, " image=", image(i, j), &
!          " sigmaimage=", backgroundsigmaimagePtr(i, j), " nsigma=", nsigma 

    HeightFunction1 = max(minHeight, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j)) ! Friday was 7.0

end Function  HeightFunction1

!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function ExtendedHeightFunction(i, j, nSigma, backgroundImage)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage


! NB 1 was 3
    ExtendedHeightFunction = max(backgroundImage(i, j) + 1.0, backgroundImage(i, j) + &
             nSigma * backgroundsigmaimagePtr(i, j))




end Function  ExtendedHeightFunction


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction11(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer :: image(:, :)

    HeightFunction11 = max(minHeight, image(i, j) + nSigma * sqrt(image(i, j))) ! Friday was 7.0


end Function  HeightFunction11


!*****************************************************
! Returns the minimum pixel value above the backround
! to be used in the source-detection
!
!
!****************************************************
real(single) Function HeightFunction2(i, j, height, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: height
    real(single), pointer :: image(:, :)

    HeightFunction2 = image(i, j) + height  ! Was 4.0

end Function  HeightFunction2


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction3(i, j, nSigma, image)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer :: image(:, :)


    HeightFunction3 = max(2.0, image(i, j) + nSigma * backgroundsigmaimagePtr(i, j)) ! Friday was 7.0


end Function  HeightFunction3


!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction4(i, j, nSigma, backgroundimage, backgroundsigmaimage)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage, backgroundsigmaImage


    HeightFunction4 = max(2.0, backgroundimage(i, j) + nSigma * backgroundsigmaimage(i, j)) ! Friday was 7.0

  
end Function  HeightFunction4

!*****************************************************
! Returns the minimum pixel value above the background
! to be used in the source-detection
!
!****************************************************
real(single) Function HeightFunction5(i, j, nSigma, backgroundimage)
    integer(int16), intent(in) :: i, j
    real(single), intent(in) :: nSigma
    real(single), pointer, dimension(:, :) :: backgroundImage


    HeightFunction5 = max(2.0, backgroundimage(i, j) + nSigma * sqrt(backgroundimage(i, j))) ! Friday was 7.0

 

end Function  HeightFunction5





!********************************************************************
! subroutine SetUpIndices
! 
!
!********************************************************************
subroutine SetUpIndices(image, tempmaskImage, iIndex, jIndex, k)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: tempmaskImage(:, :)
    integer(int16), pointer, dimension(:) :: iIndex, jIndex
    integer(int32), intent(out) :: k
    integer(int32), pointer, dimension(:) :: arrayIndex => null(), tempIndex => null()

    integer(int16) :: nx, ny, i, j
    integer(int32) ::  i1, j1

  
    real(single), pointer :: pixelValue(:) => null()

    nx = ubound(tempmaskImage, 1)
    ny = ubound(tempmaskImage, 2)


    k = count(.not. tempmaskImage) 
    arrayIndex => allocateInt32(1, k, "arrayIndex")
    iIndex => allocateInt16(1, k, "iIndex")
    jIndex => allocateInt16(1, k, "jIndex")
    pixelValue => allocateSingle(1, k, "pixelValue")

  

!if (MaskImage(ipixel, jpixel)) then

!write(*, *) "omdetectdetection setupindices line 2466 tempmaskimage ipixel, jpixel is true"
!end if


    if (binx .eq. 2) then
        detectParams%minNoPixels = 1
    else
        detectParams%minNoPixels = 2
    end if
    !**********************************************************
    !
    !
    !**********************************************************
    minHeight = 3.0  ! was 5.0
    k = 0 
    do j = lby, ny
        do i = lbx, nx
            if ( .not. tempmaskImage(i, j)) then

         
                k = k + 1
                pixelValue(k) = image(i, j)  ! Record the pixel value
                iIndex(k) = i                 
                jIndex(k) = j

            end if
        end do
    end do


!    minHeight = 7.0

    ! Index sort the pixel values 
    call IndexSort(pixelValue(1 : k), arrayIndex)


    deallocate(pixelValue)
    nullify(pixelValue)


    ! Reorder the arrays into order of decreasing pixel value
    tempIndex => allocateInt32(1, k, "tempIndex")
    tempIndex(1 : k) = iIndex(1 : k)
   
    do i1 = 1, k
        j1 = arrayIndex(i1)
         
        iIndex(k + 1 - i1) = tempIndex(j1)
    end do

    tempIndex(1 : k) = jIndex(1 : k)

    do i1 = 1, k
        j1 = arrayIndex(i1)    
        jIndex(k + 1 - i1) = tempIndex(j1)
    end do


    deallocate(arrayIndex, tempIndex)
    nullify(arrayIndex, tempIndex)


end subroutine SetUpIndices

!******************************************
!
! Initialises various things that are
! used in source detection routine
!*****************************************
subroutine InitialiseDetection(nx, ny,           &
                               radiusConstant,   &
                               nCircle,          &
                               c1,               &
                               c2,               &
                               psfSemiaConstant, &
                               nPoints,          &
                               startAngle,       &
                               finishAngle)
    integer(int16), intent(in) :: nx, ny
    real(single), intent(in) :: radiusConstant
    integer, intent(in) :: nCircle
    real(single), intent(in) :: c1, c2
    real(single), intent(in), optional :: psfSemiaConstant
    integer, optional :: nPoints
    real(single), intent(in), optional :: startAngle, finishAngle
    integer(int8) :: i, j
    real(single) :: radius, angle, deltaAngle, theta1, theta2
    real(single) :: rCon


    detectParams%minSigNoise = max(0.0_double, realParameter("minsignificance") - 1.0)

    if (present(startAngle)) then
        theta1 = startAngle
        theta2 = finishAngle
    else
        theta1 = 0.0
        theta2 = 360.0
    end if
    
    if (binx .eq. 1) then
        detectParams%c2 = 1.5
  
    else
        detectParams%c2 = 1.0
    end if
   


    nx1 = nx - 1
    ny1 = ny - 1
    
    xLower = 0.1
    yLower = 0.1
    xUpper = real(nx) - 0.1 ! was 1.1
    yUpper = real(ny) - 0.1 ! was 1.1



    if (present(nPoints)) then
        detectParams%nSize = nPoints
    else 
        detectParams%nSize = 16
    end if 

    detectParams%kMin = detectParams%nSize / 2 - 1

    detectParams%nCircles = nCircle + 3
    detectParams%nstd(1) = c1
    detectParams%nstd(2) = c2


    if (.not. associated(maskImage)) then
        allocate(maskImage(0 : nx1, 0 : ny1))
        maskImage = .false.
    end if    
  
    if (.not. associated(emaskImage)) then
        allocate(emaskImage(0 : nx1, 0 : ny1))
        emaskImage = .false.
    end if    
  

    if (associated(rCosTheta)) deallocate(rCosTheta)
    if (associated(rsinTheta)) deallocate(rsinTheta)
    if (associated(fCircle)) deallocate(fCircle)
    if (associated(mean)) deallocate(mean)
    if (associated(sigma)) deallocate(sigma)

    rCosTheta => Allocate2DSingle(1, detectParams%nSize, 1, detectParams%nCircles, "rCosTheta")
    rSinTheta => Allocate2DSingle(1, detectParams%nSize, 1, detectParams%nCircles, "rCosTheta")
    fCircle => AllocateSingle(1, detectParams%nSize, "fCircle")
    mean => AllocateSingle(1, detectParams%nCircles, "mean")
    sigma => AllocateSingle(1, detectParams%nCircles, "sigma")


    deltaAngle = (theta2 - theta1) / real(detectParams%nSize)

    if (binx .eq. 1) then
        rcon = 1.5
    else 
        rcon = 1.0    
    end if

  




    do i = 1, detectParams%nCircles
   
        radius = radiusConstant * real(i) * rcon
   

        do j = 1, detectParams%nSize
            angle = (theta1 + real(j - 1) * deltaAngle) * degreesToRadians 
            rCosTheta(j, i) = radius * cos(angle)
            rSinTheta(j, i) = radius * sin(angle)
        end do
    end do



    detectParams%nCircles = nCircle

 

    detectParams%semiaSum = 0.0_double
    detectParams%n = 0
 
    detectParams%semiAConstant = psfSemiaConstant


    detectParams%semiAConstant = 1.25 ! WAS 1.25


end subroutine InitialiseDetection



!*********************************************************
!
! The following functions are for extended source detection
!
!*********************************************************





!****************************************************************
! Extended-source detection- case where no point-sources
! have been detected
!***************************************************************
subroutine DetectExtendedSources1(nx, ny, image,  & 
                        smoothedImage, backgroundIMage,           &
                        maxImage, elevelImage,            &
                        sourceList, nPSources, noSources, &
                        minNoPixels, maxNoPixels,         &
                        minCount, height, heightFunction, &
                        IsValidSource, extensionFlag1,    &
                        filterLevelImage)

    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   


    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
    end interface

    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
            logical :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
 
    integer(int16), pointer :: elevelImage(:, :)
    integer(int16), pointer :: plevelImage(:, :) => null()



    plevelImage => allocate2dint16(lbx, nx1, lby, ny1, "pLevelImage")
    plevelImage = 0

    call DetectExtendedSources2(nx, ny, smoothedImage, & 
                        image, backgroundIMage, &
                        maxImage, plevelImage, elevelImage, &
                        sourceList, nPSources, noSources, &
                        minNoPixels, maxNoPixels, &
                        minCount, height, heightFunction, &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
   
    deallocate(pLevelImage)
    nullify(plevelImage)

end subroutine DetectExtendedSources1

!********************************************************
! Subroutine DetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources2(nx, ny, image,            & 
                        smoothedImage, backgroundIMage,     &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)



    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   
    logical, pointer :: usePixel(:, :) => null()  
       
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
 




    extensionFlag = extensionFlag1   
       

!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)
    
    comment = "Number of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
 
    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500
    ratio2 = 500 / binX  ! was 500
      

 
    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources
    

    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source


    do j = lby, ny1
        do i = lbx, nx1


            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .le. 0.0  .or. isNullValue( image(i, j) ) ) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number  
             
        end do
    end do 

!write(*, *) "Detectextendedsources2 line 3501"



    if (filterLevelImage) then

        call ErodeImage(elevelImagePtr, number)
        call DilateImage(elevelImagePtr, number)
    end if
  
  




    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = -nosources    
    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true.
 

    do j = lby, ny1
 

        do i = lbx, nx1
            
            if (eLevelImagePtr(i, j) .ne. number) cycle
            
            usePixel(i, j) = PixelOkay(i, j, minCount, number) 
            if (usePixel(i, j)) then 
                count1 = count1 + 1
            end if
        end do
    end do





    where( .not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)
 

                      
    if (count1 .lt. minNoPixels) then

        nullify(pLevelImagePtr, elevelImagePtr)    
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)         
        return
        
    end if


    ! Set the levelImage pixels in the extended regions to the source
    ! number






!write(*, *) "Detectextendedsources2 line 3395, initial sourceno=", sourceNumber
!write(*, *) "minval(elevelimageptr)=", minval(elevelImagePtr)

                        
    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 



    noSources = abs(minval(elevelImagePtr))




!write(*, *) "Detectextendedsources2 line 3540, initial sourceno=", sourceNumber
!write(*, *) "minval(elevelimageptr)=", minval(elevelImagePtr)
!write(*, *) "nosources=", noSources


    nullify(pLevelImagePtr, elevelImagePtr)    
    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)



end subroutine DetectExtendedSources2

!********************************************************
! Subroutine DetectExtendedSources3
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources3(nx, ny, il, iu, jl, ju, image,    & 
                        smoothedImage, backgroundIMage,             &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
    integer(int16), intent(in) :: nx, ny
  
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    real(single), intent(in) :: height   
    logical, pointer :: usePixel(:, :) => null()
           
    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value 
        end Function HeightFunction
  
    end interface


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
   

    extensionFlag = extensionFlag1   
       

!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)
    
    comment = "Number of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
    ratio1 = 12 / binx   ! For test
         


   ratio2 = 500 / binX  ! was 500



  

    dx = 12 / binX
    dx2 = 2 * dx 


    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources
    


    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source



    do j = jl, ju
        do i = il, iu

    

            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number  
             


        end do
    end do 




    if (filterLevelImage) then
       call ErodeImage(elevelImagePtr, number)
       call DilateImage(elevelImagePtr(il : iu, jl : ju), number)
    end if
  
      
 

    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = minval(elevelimagePtr)    

    allocate(usePixel(il : iu, jl : ju))
    usePixel(il : iu, jl : ju) = .false.  


    do j = jl, ju
 

        do i = il, iu
            
            if (eLevelImagePtr(i, j) .ne. number) cycle

            usePixel(i, j) = PixelOkay(i, j, minCount, number)
            if (usePixel(i, j)) then 
         
                count1 = count1 + 1
            else                    
                elevelImagePtr(i, j) = npSources
            end if
          
        end do
    end do
 
    where( .not. usePixel(il : iu, jl : ju)                      &
             .and. elevelImagePtr(il : iu, jl : ju) .eq. number) &
           elevelImagePtr(il : iu, jl : ju) = 0

    deallocate(usePixel)
    nullify(usePixel)   


!write(*, *) "Detectextendedsources3 il=", il, " iu=", iu, &
!" jl=", jl, " ju=", ju, " count1=", count1, " minnopixels=", minnopixels

    if (count1 .lt. minNoPixels) then

        noSources = abs(minval(elevelImagePtr))
        nullify(pLevelImagePtr, elevelImagePtr)    
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

        return
         
    end if



    ! Set the levelImage pixels in the extended regions to the source
    ! number

    call SetPixelNumbers(il, iu, jl, ju, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 

    noSources = abs(minval(elevelImagePtr))
    nullify(pLevelImagePtr, elevelImagePtr)    
    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine DetectExtendedSources3

!********************************************************
! Subroutine DetectExtendedSources4
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources4(image,    & 
                        smoothedImage, backgroundIMage,             &
                        levelImage1,&
                        sourceList, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        number1,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)
     real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

     
    type(SourceLinkedList) :: sourceList
   
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int16), intent(in) :: number1 
   

    integer(int16) :: nx, ny
 
    
    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int8) :: dx, dx2      
    integer(int32) :: count1


    nx = size(image, 1)
    ny = size(image, 2)


    extensionFlag = extensionFlag1   
       
    
    comment = "Number of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1

   
    ratio1 = 20 / binX  ! NB was 20
    ratio1 = 10 / binx   ! For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
    ratio2 = 500 / binX  ! was 500


    dx = 12 / binX
    dx2 = 2 * dx 

    eLevelImagePtr => levelImage1  ! stores extended sources



    if (.not. associated(pLevelImagePtr)) then
        pLevelImagePtr => allocate2dInt16(lbx, nx1, lby, ny1, "pLevelImagePtr")

    end if 
    



    imagePtr => image
    smoothedimagePtr => smoothedimage
    backgroundImagePtr => backgroundImage


      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
   


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source

    number = number1

    if (filterLevelImage) then
        call ErodeImage(elevelImagePtr, number)

        call DilateImage(elevelImagePtr, number)
    end if
       
    sourceNumber = minval(elevelimagePtr)    

    count1 = count(elevelImagePtr .eq. number)
   
 
    if (count1 .lt. minNoPixels) then

        nullify(elevelImagePtr)    
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)
        deallocate(pLevelImagePtr)
        nullify(pLevelImagePtr)

         
        return
         
    end if


    ! Set the levelImage pixels in the extended regions to the source
    ! number

  

    call SetPixelNumbers1(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 



    noSources = abs(minval(elevelImagePtr))
   
    nullify(elevelImagePtr, imagePtr, smoothedImagePtr, backgroundImagePtr)

    deallocate(pLevelImagePtr)
    nullify(pLevelImagePtr)

end subroutine DetectExtendedSources4

!********************************************************
! Subroutine DetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine DetectExtendedSources5(image,                    &
                        smoothedImage, backgroundIMage,     &
                        maxImage, levelImage1, levelImage2, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, height, heightFunction,   &
                        IsValidSource, extensionFlag1, &
                        filterLevelImage)

    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage

    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources

    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels
    integer(int8), intent(in) :: minCount
    real(single), intent(in) :: height
    logical, pointer :: usePixel(:, :) => null()   

    interface
        Function HeightFunction(i, j, height, image) result(value)
            use dal
            integer(int16), intent(in) :: i, j
            real(single), intent(in) :: height
            real(single), pointer :: image(:, :)
            real(single) :: value
        end Function HeightFunction

    end interface

    interface

        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:)

           logical :: status

        end function

    end interface

    integer(int8), intent(in) :: extensionFlag1
    logical, intent(in) :: filterLevelImage
    integer(int16), pointer :: levelImage1(:, :)
    integer(int16), pointer :: levelImage2(:, :)

    integer(int16) :: i, j
    integer(int8) :: dx, dx2
    integer(int32) :: count1

!write(*, *) "DetectExtendedSources5 line 4007"
    extensionFlag = extensionFlag1


!    write(errstr,'(a, a)') "Looking for ", title
!    call message(AppMsg, Verbose)

    comment = "Number of likely " // trim(title) // " detected so far="

    nx1 = ubound(image, 1)
    ny1 = ubound(image, 2)
    lbx = lbound(image, 1)
    lby = lbound(image, 2)


    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500

    ratio2 = int(PI * 144.0) / binX  ! was 500
   ratio2 = 500 / binX  ! was 500


    dx = 12 / binX
    dx2 = 2 * dx

    eLevelImagePtr => levelImage2  ! stores extended sources
    pLevelImagePtr => levelImage1  ! stores point sources


    imagePtr => image
    smoothedimagePtr => smoothedimage

    backgroundImagePtr => backgroundImage


    ! Set all the levelImage pixels above the background in the level image
    ! to some arbitrary value
    number = nPSources + 1

    ! Loop through all the image pixels to find ones possibly
    ! linked to a source

    do j = lby, ny1
        do i = lbx, nx1


            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle

            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  HeightFunction(i, j, height, backgroundImage)) &
                                  elevelImagePtr(i, j) = number

        end do
    end do



!write(*, *) "DetectExtendedSources5 line 4066"




    if (filterLevelImage) then
        call ErodeImage(elevelImagePtr, number)

        call DilateImage(elevelImagePtr, number)
    end if



    ! Get rid of isolated source pixels
    count1 = 0
    sourceNumber = -nosources
    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true.
  



!write(*, *) "DetectExtendedSources5 line 4088"



    do j = lby, ny1


        do i = lbx, nx1

            if (eLevelImagePtr(i, j) .ne. number) cycle

            usePixel(i, j) = PixelOkay(i, j, minCount, number)
            if (usePixel(i, j)) then
                count1 = count1 + 1
          
            end if
        end do
    end do


    where( .not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)  

!write(*, *) "DetectExtendedSources5 line 4112"


    if (count1 .lt. minNoPixels) then

  
        nullify(pLevelImagePtr, elevelImagePtr)
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

        return

    end if




    backgroundImagePtr => backgroundImage
    ! Set the levelImage pixels in the extended regions to the source
    ! number

    imagePtr => image
    smoothedimagePtr => smoothedimage

!write(*, *) "DetectExtendedSources5 line 4135"


    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource)


!write(*, *) "DetectExtendedSources5 line 4144"


    noSources = abs(minval(elevelImagePtr))
    nullify(pLevelImagePtr, elevelImagePtr)

    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine DetectExtendedSources5


!********************************************************
! Subroutine GrismDetectExtendedSources
! Case where point-sources have been detecetd
! Sets the pixels on the levelImage to the source number
! (0=background +for point-sources, - for extended sources)
!********************************************************
subroutine GrismDetectExtendedSources(nx, ny, image,    & 
                        smoothedImage, backgroundIMage,             &
                        maxImage, elevelImage, &
                        sourceList, nPSources, noSources,   &
                        minNoPixels, maxNoPixels,           &
                        minCount, isValidSource, extensionFlag1)
    integer(int16), intent(in) :: nx, ny
    real(single), pointer :: image(:, :), smoothedImage(:, :), backgroundImage(:, :)

    real(single), intent(in) :: maxImage
     
    type(SourceLinkedList) :: sourceList
    integer(int16), intent(in) :: nPSources
  
    integer(int16), intent(inout) :: noSources
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    integer(int8), intent(in) :: minCount 
    logical, pointer :: usePixel(:, :) => null()   
  
       

    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical :: status

        end function 

    end interface

    integer(int8), intent(in) :: extensionFlag1
    integer(int16), pointer :: elevelImage(:, :)
    

    integer(int16) :: i, j    
    integer(int8) :: dx, dx2      
    integer(int32) :: count1
 
    extensionFlag = extensionFlag1   
       
    write(errstr,'(a, a)') "Looking for ", title
    call message(AppMsg, Verbose)
    
    comment = "Number of likely " // trim(title) // " detected so far="


    nx1 = nx - 1
    ny1 = ny - 1
    lbx = 0
    lby = 0
   
    ratio1 = 12 / binx   ! was 10 For test
    ratio2 = 300 / binX  ! was 500
          
    ratio2 = int(PI * 144.0) / binX  ! was 500
 
    ratio2 = 500 / binX  ! was 500
 
 
    dx = 12 / binX
    dx2 = 2 * dx 

    eLevelImagePtr => elevelImage  ! stores extended sources
   
   
    imagePtr => image
    smoothedimagePtr => smoothedimage

    backgroundImagePtr => backgroundImage

      
    ! Set all the levelImage pixels above the background in the level image 
    ! to some arbitrary value
    number = nPSources + 1     


    ! Loop through all the image pixels to find ones possibly
    ! linked to a source


    do j = lby, ny1
        do i = lbx, nx1


            if (elevelImagePtr(i, j) .ne. 0) cycle
            if (smoothedImage(i, j) .gt. maxImage) cycle
         
            if (image(i, j) .lt. 0.0) cycle
            if (backgroundImage(i, j) .lt. 0.0) cycle  ! Ignore negative-valued pixels
            if (smoothedImage(i, j) .ge.  backgroundImage(i, j)) elevelImagePtr(i, j) = number  
             
        end do
    end do 

  
 
    ! Get rid of isolated source pixels    
    count1 = 0
    sourceNumber = -nosources    
    allocate(usePixel(0 : nx1, 0 : ny1))
    usePixel = .true.
   
    do j = lby, ny1
 

        do i = lbx, nx1
            
            if (eLevelImagePtr(i, j) .ne. number) cycle
            
            usePixel(i, j) = PixelOkay(i, j, minCount, number) 
            if (usePixel(i, j)) then
                count1 = count1 + 1
            end if
        end do
    end do

    where (.not. usePixel) elevelImagePtr = 0
    deallocate(usePixel)
    nullify(usePixel)
                     
    if (count1 .lt. minNoPixels) then

        nullify(elevelImagePtr)
        nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)


         
        return
        
    end if


 

   
    backgroundImagePtr => backgroundImage
    ! Set the levelImage pixels in the extended regions to the source
    ! number


    imagePtr => image
    smoothedimagePtr => smoothedimage


    call SetPixelNumbers(lbx, nx1, lby, ny1, count1, minNoPixels,  &
                      maxNoPixels, sourceList, IsValidSource) 


    noSources = abs(minval(elevelImagePtr))
    nullify(elevelImagePtr)    

    nullify(imagePtr, smoothedImagePtr, backgroundImagePtr)

end subroutine GrismDetectExtendedSources

!******************************************************************
! Set the levelImage pixels in the extended regions to the source
! number
!******************************************************************
subroutine SetPixelNumbers(istart, ifinish, jstart, jfinish, count1, minNoPixels, &
              maxNoPixels, sourceList, IsValidSource, noSourcesOut) 
     
    integer(int16), intent(in) :: istart, ifinish, jstart, jfinish
    integer(int32), intent(inout) :: count1
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    type(SourceLinkedList) :: sourceList

    integer(int16), optional, intent(out) :: noSourcesOut
 
    interface
      
            function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical :: status

        end function 

    end interface
    type(SourceParameters) :: srcParams
    integer(int16) :: i1, j1, noSources   
    integer(int32) :: i, j, il, il1
    Type(PixelStruct), pointer :: pixel(:) => null()

 

    allocate(pixel(1 : count1))

!    write(*, *) "sepixelnumbers line 4337"


    count1 = 0
    ! Fill the array of pixel strctures with information on the pixels
    do j1 = jstart, jfinish
        do i1 = istart, ifinish   
            if (elevelImagePtr(i1, j1) .eq. number) then
                count1 = count1 + 1
                pixel(count1)%x = i1
                pixel(count1)%y = j1
                pixel(count1)%counts = smoothedimagePtr(i1, j1)
                pixel(count1)%backgroundCounts = backgroundImagePtr(i1, j1)
                pixel(count1)%number = 0  ! Source number                         
                pixel(count1)%noNeighbours = NumberOfNeighbours(i1, j1, number)
           

            end if
        end do
    end do


!   write(*, *) "sepixelnumbers line 4362"


    sourceNumber = minval(elevelImagePtr)
    noSources = -1




    il1 = 1
    il = 1 ! First source pixel

iLoop : do i = 1, count1 - 1


  
        do j = i + 1, count1
                    
            if (Neighbours(pixel(i), pixel(j))) then
               

                ! The first new extended source                                                                                                   
                noSources = 0                             
                sourceNumber = sourceNumber - 1                                      
                                
                pixel(i)%number = sourceNumber ! The first source - first pixel
                pixel(j)%number = sourceNumber ! First source - second pixel
                call MovePixel(i, j, pixel)    ! Move pixel j to immediately after pixel i
                il = i                         ! The first pixel for this source       
                il1 = i + 1
 
                exit iLoop     
                                   
            end if
        end do

    end do iLoop     



!   write(*, *) "sepixelnumbers line 4405, count1=", count1


    !*************************************************************
    ! If il1 equals il then no sources
    !*************************************************************
    if (il1 .eq. il) then 
        deallocate(pixel)
        nullify(pixel)
        if (noSources == -1) noSources = 0    
        return
    end if


!write(*, *) "sepixelnumbers line 4457, count1=", count1, " nx1=", nx1, " ny1=", ny1


    ! Repeat the above process for remaining pixels
    do i = il1, count1 - 1


        do j = i + 1, count1
   


            if (pixel(i)%number .lt. 0 .and. pixel(j)%number .lt. 0) cycle



            if (Neighbours(pixel(i), pixel(j))) then
        
                if (pixel(i)%number .lt. 0) then
                    ! Pixel j is connected to pixel i
                    pixel(j)%number = pixel(i)%number

                    il1 = il1 + 1                         
                else


                    ! Both pixels i and j belong to an unnumbered source                    
                    !
                    ! Examine list of source pixels
                    call SetExtendedSource(sourceList, &
                        noSources, il, il1, minNoPixels, maxNoPixels, pixel, & 
                        srcParams, IsValidSource)

                    pixel(i)%number = sourceNumber
                    pixel(j)%number = sourceNumber
                    il = i       ! First pixel of new source
                    il1 = i + 1  ! Second pixel of new source
                  
   
                end if    

                ! Reposition pixel j in pixel array
                ! to immediately after position i
                call MovePixel(i, j, pixel)
       
                   
            end if
        end do

    end do     



!   write(*, *) "sepixelnumbers line 4470"

    ! Examine any remaining pixels

    if(il1 .gt. il) then 

        call SetExtendedSource(sourceList, &
           noSources, il, il1, minNoPixels, maxNoPixels, pixel, &
           srcParams, IsValidSource)
                  
    end if


!   write(*, *) "sepixelnumbers line 4487"


    deallocate(pixel)
    nullify(pixel)

   
 

    if (noSources == -1) noSources = 0    

    if (present(noSourcesOut)) noSourcesOut = noSources
 



!    write(errstr,'(a, i0)') "Total number found=", noSources
!    call message(AppMsg, Verbose)
                                 
                  
end subroutine SetPixelNumbers

!******************************************************************
! Set the levelImage pixels in the extended regions to the source
! number
!******************************************************************
subroutine SetPixelNumbers1(istart, ifinish, jstart, jfinish, count1, minNoPixels, &
              maxNoPixels, sourceList, IsValidSource, noSourcesOut) 
     
    integer(int16), intent(in) :: istart, ifinish, jstart, jfinish
    integer(int32), intent(inout) :: count1
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
    type(SourceLinkedList) :: sourceList
    integer(int16), optional, intent(out) :: noSourcesOut
    interface
      
            function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
     
           logical :: status

        end function 

    end interface
    type(SourceParameters) :: srcParams
       
    integer(int32) :: i, j, il, il1, il2
    integer(int16) :: noSources
    Type(PixelStruct), pointer :: pixel(:) => null()


    allocate(pixel(1 : count1))



    count1 = 0


    ! Fill the array of pixel strctures with information on the pixels
    do j = jstart, jfinish
        do i = istart, ifinish   
            if (elevelImagePtr(i, j) .eq. number) then
                count1 = count1 + 1
                pixel(count1)%x = i
                pixel(count1)%y = j
               

            end if
        end do
    end do



    sourceNumber = minval(elevelImagePtr)
    noSources = -1
    il1 = 1
    il = 1 ! First source pixel

iLoop : do i = 1, count1 - 1

        do j = i + 1, count1
                    
            if (Neighbours(pixel(i), pixel(j))) then
               

                ! The first new extended source                                                                                                   
                noSources = 0                             
                sourceNumber = sourceNumber - 1                                      
                                      
                pixel(i)%number = sourceNumber ! The first source - first pixel
                pixel(j)%number = sourceNumber ! First source - second pixel
                call MovePixel(i, j, pixel)    ! Added today- Pixel j moved to immediately after pixel i
                il = i                         ! The first pixel for this source       
                il1 = i + 1 
                exit iLoop     
                                   
            end if
        end do

    end do iLoop     



    il2 = il1
    ! Repeat the above process for remaining pixels
    do i = il1, count1 - 1

        do j = i + 1, count1


            if (pixel(i)%number .lt. 0 .and. pixel(j)%number .lt. 0) cycle
            if (Neighbours(pixel(i), pixel(j))) then
        
                if (pixel(i)%number .lt. 0) then
                    ! Pixel j is connected to pixel i
                    pixel(j)%number = pixel(i)%number

                    il1 = il1 + 1                         
                else
                    ! Both pixels i and j belong to an unnumbered source                    
                    !
                    ! Examine list of source pixels
                    call SetExtendedSource1(sourceList, &
                        noSources, il, il1, minNoPixels, maxNoPixels, pixel, & 
                        srcParams, IsValidSource)
           
                    pixel(i)%number = sourceNumber
                    pixel(j)%number = sourceNumber
                    il = i       ! First pixel of new source
                    il1 = i + 1  ! Second pixel of new source
                    il2 = i + 1

   
                end if    

                ! Reposition pixel in pixel array
                ! to immediately after position i
                call MovePixel(i, j, pixel)
       
                   
            end if
        end do

    end do     


    ! Examine any remaining pixels
    if (il2 .lt. il1 .or. il .eq. 1) then


        call SetExtendedSource1(sourceList, &
           noSources, il, il1, minNoPixels, maxNoPixels, pixel, &
           srcParams, IsValidSource)
                  
    end if

  
    deallocate(pixel)
    nullify(pixel)

    if (noSources == -1) noSources = 0    
    write(errstr,'(a, i0)') "Total number found=", noSources
    call message(AppMsg, Verbose)
    if (present(noSourcesOut)) noSourcesOut = noSources
                                 
                  
end subroutine SetPixelNumbers1


!*****************************************
! Subroutine SetExtendedSource
!*****************************************
subroutine SetExtendedSource(sourceList,               &
                             noSources, il, iu,        &
                             minNoPixels, maxNoPixels, &
                             pixel, srcParams, IsValidSource)
    type(SourceLinkedList), intent(inout) :: sourceList

    integer(int16), intent(inout) :: noSources    
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
        
    Type(PixelStruct), pointer :: pixel(:)
 
    type(SourceParameters), intent(inout) :: srcParams


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use Parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical :: status

        end function 

    end interface
   
  




!if (any(pixel(il : iu)%x .eq. ipixel .and. pixel(il : iu)%y .eq. jPixel)) then
!write(*, *) "Line 3970 checking validity of source ", ipixel, ", ", jpixel, &
!" nopixels=", srcParams%noSourcePixels, " il=", il, " iu=", iu, &
!" min=", minNoPixels, " max=", maxNoPixels
!end if

    
    if (AcceptExtendedSource(il, iu,  &
                 minNoPixels,         &
                 maxNoPixels,         &
                 srcParams)) then



        srcParams%qualityFlag = 0
        srcParams%confusionFlag = 0
        srcParams%extensionFlag = extensionFlag
      

!if (any(pixel(il : iu)%x .eq. ipixel .and. pixel(il : iu)%y .eq. jPixel)) then
!write(*, *) "Line 4004 checking validity of source ", ipixel, ", ", jpixel
!end if


 


        ! Check if valid source  
        if ( IsValidSource(srcParams, pixel(il : iu)) ) then  ! Valid source
          
       
     
            ! Set the pixels on the level-image map
            call SetLevelImagePixels(il, iu, pixel)


            ! This is to reduce problems with centroiding
            srcParams%rawCountRate = -maxval(pixel(il : iu)%counts)


                                          
            ! Add it to the source list
            srcParams%number = sourceNumber      
          
            

        
            srcParams%axialRatio = srcParams%semiA / srcParams%semiB
            srcParams%alpha = srcParams%theta

            ! Add the source to the linked-list
            call AddSourceToList(sourceList, srcParams)            

            noSources = noSources + 1          

!write(*, *) "Line 4545 new number of sources=", noSources, " sourcenumber=", sourceNumber
!write(*, *) "minval(elevlimagePtr)=", minval(elevelimagePtr)




!            if(len(trim(title)) .gt. 3) then
!                write(errstr,'(a, i0)') trim(comment), noSources
!                call message(AppMsg, Verbose)
!            end if
            sourceNumber = sourceNumber - 1  ! The next extended-source number

!         else

!if (any(pixel(il : iu)%x .eq. ipixel .and. pixel(il : iu)%y .eq. jPixel)) then
!write(*, *) "Line 4048 invalid source ", ipixel, ", ", jpixel
!end if





         end if   


    end if             
end subroutine SetExtendedSource

!*****************************************
! Subroutine SetExtendedSource
!*****************************************
subroutine SetExtendedSource1(sourceList,               &
                             noSources, il, iu,        &
                             minNoPixels, maxNoPixels, &
                             pixel, srcParams, IsValidSource)
    type(SourceLinkedList), intent(inout) :: sourceList

    integer(int16), intent(inout) :: noSources    
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
        
    Type(PixelStruct), pointer :: pixel(:)
 
    type(SourceParameters), intent(inout) :: srcParams


    interface
      
        function IsValidSource(srcParams, pixel) result(status)
            use dal
            use OmSourceParam_mod    ! Defines structures for storing source parameters
            use Parameters_mod

            type(SourceParameters), intent(inout) :: srcParams
            type(PixelStruct) :: pixel(:) 
    
            logical :: status

        end function 

    end interface
   

                         
    ! Cgeck to see if the number of pixels is within the allowed range 
    if ( AcceptExtendedSource(il, iu,  &
                 minNoPixels,          &
                 maxNoPixels,          &
                 srcParams)) then
   





        srcParams%qualityFlag = 0
        srcParams%confusionFlag = 0
        srcParams%extensionFlag = extensionFlag
    


        ! Check if valid source  
        if ( IsValidSource(srcParams, pixel(il : iu)) ) then  ! Valid source
            
 
            ! Set the pixels on the level-image map
            call SetLevelImagePixels(il, iu, pixel)
          
                                          
            ! Add it to the source list
            srcParams%number = sourceNumber      
         

            ! This is to reduce problems with centroiding
            srcParams%rawCountRate = -maxval(pixel(il : iu)%counts)




            ! Add the source to the linked-list
            call AddSourceToList(sourceList, srcParams)            

            noSources = noSources + 1        ! The new number of sources        
           
            sourceNumber = sourceNumber - 1  ! The next extended-source number

         end if   


    end if             
end subroutine SetExtendedSource1

!*****************************************************
! Function AcceptExtendedSource
!****************************************************
logical function AcceptExtendedSource(il, iu, &
                 minNoPixels,         &
                 maxNoPixels,         &
                 srcParams)
    integer(int32), intent(in) :: il, iu
    integer(int32), intent(in) :: minNoPixels, maxNoPixels 
  
    type(SourceParameters), intent(inout) :: srcParams
   
   
    srcParams%noSourcePixels = iu - il + 1 ! The number of pixels associated with the source
 
    if (srcParams%noSourcePixels .lt. minNoPixels .or. &
             srcParams%noSourcePixels .gt. maxNoPixels) then

      
        AcceptExtendedSource = .false.
         
   
    else 
        AcceptExtendedSource = .true.
        
    end if

end function AcceptExtendedSource


!****************************************************
! Subroutine SetLevelImagePixels
!***************************************************
subroutine SetLevelImagePixels(il, iu, pixel)

    integer(int32), intent(in) :: il, iu
    Type(PixelStruct), pointer :: pixel(:)
    integer(int32) :: i
    

    do i = il, iu        
        
        elevelImagePtr(pixel(i)%x, pixel(i)%y) = pixel(i)%number    
      
        
    end do
end subroutine setLevelImagePixels

!******************************************************
! Returns the number of stars within an extended region
!
!*****************************************************
integer(int16) Function NumberOfStars1(pixel)
    type(PixelStruct) :: pixel(:)


    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
   
    numberOfStars1 = 0
 






  
    ! Find the number of star pixels amongst the extended source pixels

    starNumber = -1
    starNumberList = 0

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            if (plevelImagePtr(x, y) .ne. starNumber) then
                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars1 = numberOfStars1 + 1
                starNumberList(numberOfStars1) = starNumber
                if (numberOfStars1 .eq. 20) exit
            end if     
        end if
    end do 
           
 

end function NumberOfStars1

!******************************************************
!
! Returns the number of stars within an extended region
! and the location and maximum pixel image of that star
!
!*****************************************************
integer(int16) Function NumberOfStars2(pixel, maxStarPixel)
    type(PixelStruct) :: pixel(:)
    type(PixelStruct), intent(out) :: maxStarPixel
    

    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
    numberOfStars2 = 0
 
  
    ! Find the number of star pixels amongst the extended source pixels

    starNumber = -1
    starNumberList = 0
    

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            if (imagePtr(x, y) .gt. maxStarPixel%counts) maxStarPixel = pixel(i)
            if (plevelImagePtr(x, y) .ne. starNumber) then
                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars2 = numberOfStars2 + 1
                starNumberList(numberOfStars2) = starNumber
                if (numberOfStars2 .eq. 20) exit
            end if     
        end if
    end do 
           
 
end function NumberOfStars2
!******************************************************
!
! Returns the number of stars within an extended region
! and the location and maximum pixel image of that star
!
!*****************************************************
integer(int16) Function NumberOfStars3(pixel, starCounts)
    type(PixelStruct) :: pixel(:)
    real(double), intent(out) :: starCounts
    

    integer(int16) :: i
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
    integer :: x, y
    numberOfStars3 = 0
 
  
    ! Find the number of star pixels amongst the extended source pixels

    starCounts = 0.0_double
    starNumberList = 0
    

    do i = 1, size(pixel)
        x= pixel(i)%x
        y= pixel(i)%y

        if (plevelImagePtr(x, y) .gt. 0) then
            starCounts = starCounts + dble(imagePtr(x, y))
            if (plevelImagePtr(x, y) .ne. starNumber) then

                starNumber = plevelImagePtr(x, y)
                if (any(starNumberList .eq. starNumber)) cycle  
                numberOfStars3 = numberOfStars3 + 1
                starNumberList(numberOfStars3) = starNumber
                if (numberOfStars3 .eq. 20) exit
            end if     
        end if
    end do 
           
 
end function NumberOfStars3


!******************************************************
!
! Returns the number of stars within an extended region
! and the location and maximum pixel image of that star
!
!*****************************************************
integer(int16) Function NumberOfStars4(source, maxstarPixel)
  
    type(SourceParameters), intent(in) :: source
    
    type(PixelStruct), intent(out) :: maxStarPixel
 
   
 
    integer(int16) :: i, j
    integer(int16) :: starNumber
    integer(int16) :: starNumberList(20)
  
    numberOfStars4 = 0
 
  

!if(checkpixel) then
!write(*, *) "numberofstars4 line 4845 imin=", source%imin, " imax=", source%imax, &
!" jmin=", source%jmin, " jmax=", source%jmax
!end if

    ! Find the number of star pixels amongst the extended source pixels

    starNumber = -1
    starNumberList = 0
    

    do j = source%jmin, source%jmax
        do i = source%imin, source%imax
  

            if (plevelImagePtr(i, j) .gt. 0) then
         
   
                if (plevelImagePtr(i, j) .ne. starNumber) then
                    if (imagePtr(i, j) .gt. maxStarPixel%counts) maxStarPixel = &
                        pixelStruct( i, j, starnumber, imagePtr(i, j), 0.0, 0 )








     
                    starNumber = plevelImagePtr(i, j)
                    if (any(starNumberList .eq. starNumber)) cycle  
                    numberOfStars4 = numberOfStars4 + 1
                    starNumberList(numberOfStars4) = starNumber
                    if (numberOfStars4 .eq. 20) return
                end if     
            end if
        end do
    end do 
           
 
end function NumberOfStars4





!******************************************************
!
!
!*****************************************************
integer(int16) Function NumberOfHoles(pixel)
    type(PixelStruct) :: pixel(:)


    integer(int16) :: i, j
  
    numberOfHoles = 0
 

    do j = minval(pixel%y), maxval(pixel%y)
        

        do i = minval(pixel%x, pixel%y .eq. j), maxval(pixel%x, pixel%y .eq. j)

            if (eLevelImagePtr(i, j) .eq. 0) NumberOfHoles = NumberOfHoles + 1
          

        end do
    end do


end function NumberOfHoles

!***************************************************
! Function PixelOkay
! 
!**************************************************
logical function PixelOkay(i, j, minCount, number)

    integer(int16), intent(in) :: i, j 

    integer(int8), intent(in) :: minCount
    integer(int16), intent(in) :: number
    integer(int8)  :: count

    ! Local variables
    
    integer(int16) :: k, l, l1, l2
 

    count = 0


    count = NumberOfNeighbours1(i, j, number)

    if (count .ge. minCount) then
        pixelOkay = .true.
    else
        pixelOkay = .false.
    end if
    return


    l1 =  max(lby, j - 1_int16)
    l2 =  min(ny1, j + 1_int16)

    do k = max(lbx, i - 1_int16), min(nx1, i + 1_int16)
        do l = l1, l2
            if (k == i .and. l == j) cycle

            if (elevelImagePtr(k, l) .eq. number) then 
                count = count + 1
                
                
                if (count .ge. minCount) then
                    PixelOkay = .true.
                    return
                end if
          
            end if

        end do
    end do      

             
    PixelOkay = .false.

end function PixelOkay
!***************************************************
! Function NumberOfNeighbours1
! Determines the nubmer of pixels that are connected
! to this pixel
!        x
! ie   xxoxxx if o is the pixel i,j and the x's are
!       x
! the pixels in this group of pixels above the background,
! NumberofNeighbours = 3
!**************************************************
function NumberOfNeighbours1(i, j, number) result(count)
    integer(int16), intent(in) :: i, j 
    integer(int16), intent(in) :: number
    integer(int8)  :: count
    ! Local variables
   
 
    count = 0

    if (i .gt. 0) then
        if (elevelImagePtr(i - 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (i .lt. nx1) then
        if (elevelImagePtr(i + 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .gt. 0) then
        if (elevelImagePtr(i, j - 1) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .lt. ny1) then
        if (elevelImagePtr(i, j + 1) .eq. number) then 
            count = count + 1
             
        end if
    end if


end function NumberOfNeighbours1
!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours2(levelimagePtr, i, j, number) result(count)
    integer(int16), pointer :: levelimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
    integer(int16), intent(in) :: number 
  
    integer(int16) :: nx, ny

  
    integer(int8)  :: count
    ! Local variables
   
 
    count = 0
    nx = ubound(levelimagePtr, 1)
    ny = ubound(levelimagePtr, 2)

    if (i .gt. 0) then
        if (levelImagePtr(i - 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (i .lt. nx) then
        if (levelImagePtr(i + 1, j) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .gt. 0) then
        if (levelImagePtr(i, j - 1) .eq. number) then 
            count = count + 1
             
        end if
    end if

    if (j .lt. ny) then
        if (levelImagePtr(i, j + 1) .eq. number) then 
            count = count + 1
             
        end if
    end if


end function NumberOfNeighbours2

!***************************************************
! Function NumberOfNeighbours2
! 
!**************************************************
function NumberOfNeighbours3(maskImagePtr, i, j) result(count1)
    logical,  pointer :: maskimagePtr(:, :)
    integer(int16), intent(in) :: i, j 
  
  
    integer(int16) :: nx, ny
    integer(int16) :: il, iu, jl, ju, i1, j1
    
    integer(int8)  :: count1
    ! Local variables
   
 
    nx = ubound(maskimagePtr, 1)
    ny = ubound(maskimagePtr, 2)

    il = max(0_int16, i - 1_int16)
    iu = min(nx, i + 1_int16)

    jl = max(0_int16, j - 1_int16)
    ju = min(ny, j + 1_int16)
 



!write(*, *) "Numberofneighbours line 4309 i=", i, " j=", j

!write(*, *) "Line 4312 No of none true maskimagePtr=", count(.not. MaskImagePtr) 


    count1 = 0
    do j1 = jl, jl
        do i1 = il, iu

            if (i1 .eq. i .and. j1  .eq. j) cycle

      


            if (maskImagePtr(i1, j1)) cycle
!            write(*, *) "Pixel ", i1, ", ", j1, " okay", " value=", imagePtr(i1, j1) 
!            if ( imagePtr(i, j) .ge. HeightFunction1(i1, j1, 2.0, backgroundimagePtr)) then 
                  count1 = count1 + 1
!            end if

        end do
    end do


end function NumberOfNeighbours3

!***********************************************
! Subroutine MovePixel - Inserts pixel(j), between
! pixels i and pixels i + 1
!**********************************************
subroutine MovePixel(i, j, pixel)
    integer(int32), intent(in) :: i, j
    Type(PixelStruct), pointer :: pixel(:)
    Type(PixelStruct) :: pixelJ
    integer(int32) :: k

    if (j .eq. i + 1) return  
    
    pixelJ = pixel(j)

    do k = j, i + 2, -1
        pixel(k) = pixel(k - 1)
    end do

    pixel(i + 1) = pixelJ 
      
end subroutine MovePixel
!***********************************************************
! Function Neighbours
!***********************************************************
logical function Neighbours(pixel1, pixel2)
    Type(PixelStruct), intent(in) :: pixel1, pixel2

    if (abs(pixel1%x - pixel2%x) .le. 1) then
        if(abs(pixel1%y - pixel2%y) .le. 1) then
            Neighbours = .true.
        else
            Neighbours = .false.

        end if
    else
        Neighbours = .false.
    end if
end function Neighbours


!*****************************************************
! IsExtendedSourceNoChecks1
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical Function IsExtendedSourceNoChecks1(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    integer(int32) :: noHoles

    noHoles = NumberOfHoles(pixel)


    if ( (noHoles * 7) .lt. size(pixel)) then
        IsExtendedSourceNoChecks1 = .false.
        return
    end if
 
    IsExtendedSourceNoChecks1 = .true.
    return 
  
    !******************************************************
    ! Check that centre is near 992 - x0
    !******************************************************
    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)

    
    IsExtendedSourceNoChecks1 = .true.

end function IsExtendedSourceNoChecks1

!*****************************************************
! IsExtendedSourceNoChecks2
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical Function IsExtendedSourceNoChecks2(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    integer :: noStars
   
   
    type(PixelStruct) :: maxStarPixel



    !**********************************************
    !
    !*********************************************
    if(maxval(pixel%counts) .lt. 1000.0) then
       IsExtendedSourceNoChecks2 = .false.
       return
     

    end if


    noStars = NumberOfStars(pixel, maxStarPixel)

    if (noStars .eq. 0 .or. noStars .gt. 2) then
        IsExtendedSourceNoChecks2 = .false.
        return
     

    end if

    !********************************************************
    ! Ignore star if peak counts < 20 sigma
    !*******************************************************
    if ( (maxStarPixel%counts - maxStarPixel%backgroundCounts) / &
             sqrt(maxStarPixel%counts) .lt. 20.0) then
        IsExtendedSourceNoChecks2 = .false.
        return
    end if
    

    IsExtendedSourceNoChecks2 = .true. ! One star
  
    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)
    if ( abs(source%imax - source%imin - source%jmax + source%jmin) .gt. 10) then
        IsExtendedSourceNoChecks2 = .false.
    else
        IsExtendedSourceNoChecks2 = .true.
    end if

   
end function IsExtendedSourceNoChecks2


!*****************************************************
!
! IsValidSmokeRing
! Checks to see if the parameters of an extended
! source lies within certain limits
!
!*****************************************************
logical Function IsValidSmokeRing(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    real(single), parameter :: maxRatio = 1.25 ! okay with 1.4
    integer(int16) :: i, j, i0, j0
    real(single) :: radius, radius1, radius2, radius3, radius4, dx, dy
    integer(int32) :: n, n1, n2, n3, n4, n5  
    real(double) :: sum1, sum2    

    !*******************************************
    !
    ! Compute the source parameters
    !
    !*******************************************

    call ComputeExtendedSourceParameters(sourceNumber, &
                                         pixel,        &
                                         source)
   

    !*****************************************************
    ! Check the axes lengths
    !
    !****************************************************

    if (source%semia < real(8 / binx)) then
        isValidSmokeRing = .false.
        return
    end if

    if (source%semia > real(24 / binx)) then   
        isValidSmokeRing = .false.
        return
    end if


    if (source%semia / source%semib .gt. maxRatio) then
        isValidSmokeRing = .false.
        return
    end if


    !*******************************************
    !
    ! Check that there is a hole at the centre
    !
    !*******************************************
    
    radius1 = real(6 / binx)
    radius2 = real(14 / binx)
    radius3 = real(36 / binx) 

    radius3 = real(40 / binx) 
    radius4 = real(60 / binx) 


  

    i0 = nint(source%xpos - 0.5)
    j0 = nint(source%ypos - 0.5)
 
    i0 = max(0_int16, i0)
    j0 = max(0_int16, j0)
    
   
    !*************************************************************
    !
    ! Check that there is a hole in the source
    !
    !************************************************************
    n = size(pixel) ! The number of pixels assigned to the source
    n1 = 0
    n2 = 0
    n3 = 0
    n4 = 0
    n5 = 0
    sum1 = 0.0_double
    sum2 = 0.0_double


    do j = minval(pixel%y), maxval(pixel%y)

        dy = real(j0 - j)      
        do i = minval(pixel%x), maxval(pixel%x)
            dx = real(i0 - i)
            radius = sqrt(dx * dx + dy * dy)

            !***************************************************************
            !
            ! Check for central hole
            !
            !************************************************************** 
            if (radius .le. radius1) then ! The inner radius of the annulus 
                if( .not. isNullValue(imagePtr(i, j)) ) then
                    n1 = n1 + 1 
                    sum1 = sum1 + dble(imagePtr(i, j))
                end if             
                if ( elevelimagePtr(i, j) .ne. 0) n2 = n2 + 1
                    
                cycle                
            end if               

            if (radius .lt. radius2) cycle
         
            !***************************************************************
            !
            ! Check the annulus region
            !
            !************************************************************** 
            if (radius .le. radius3) then
                if( .not. isNullValue(imagePtr(i, j)) ) then
          
                    n3 = n3 + 1               
                    sum2 = sum2 + dble(imagePtr(i, j)) 
                end if
                if ( elevelimagePtr(i, j) .eq. 2) n4 = n4 + 1
   
                cycle
            end if

            if (radius > radius4) cycle
            !***************************************************************
            !
            ! Check outside annulus
            !
            !***************************************************************               
            if ( elevelimagePtr(i, j) .ne. 0 .and. plevelimagePtr(i, j) == 0) n5 = n5 + 1
            
        end do

    end do 
 


    if ( (real(n2) / real(n1)) > 0.33  .or. (real(n4) / real(n3)) < 0.4 &
               .or. (real(n5) / real(n)) > 0.25  ) then

        IsValidSmokeRing = .false.
        return
    end if

    if(n1 > 0 .and. n3 > 0) then
        sum1 = max(0.0_double, sum1 / dble(n1))
        sum2 = sum2 / dble(n3)
        if(sum2 < ( sum1 + max(1.0_double, sqrt(sum1)))  ) then
            IsValidSmokeRing = .false.
            return
        end if


    end if
    !**********************************************
    ! Remember that it is probably a smoke ring    
    !********************************************* 
    source%qualityFlag = ibset(source%qualityFlag, 2) 
    IsValidSmokeRing = .true.
   
end function IsValidSmokeRing
!*****************************************************
!
! IsValidRubbishSource
! Checks to see if the parameters of an extended
! source liw within certain limits
!
!*****************************************************
logical Function IsValidRubbishSource(srcParams, pixel)     
    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
  
    srcParams%iMin = minval(pixel%x)
    srcParams%iMax = maxval(pixel%x)
    srcParams%jMin = minval(pixel%y)
    srcParams%jMax = maxval(pixel%y)


    if (areaOkay(srcParams, Pixel, 3.0)) then 
        IsValidRubbishSource  = .false.
    else
        IsValidRubbishSource  = .true.
 

    end if

    
end function IsValidRubbishSource

!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended
! source liw within certain limits
!*****************************************************
logical Function IsValidExtendedSource(source, pixel)
     
 
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    
    integer(int16)  :: noHoles, noNeighbours, noStars
    integer(int16) :: width, height 

    integer(int16), parameter :: ratio = 7
  
    integer(int16) :: i0, j0, n, il, iu, jl, ju

!    logical :: checkPixel
    real(double) :: starCounts


!integer i


!checkPixel = .false.
!if (any(pixel%x .eq. ipixel .and. pixel%y .eq. jpixel)) checkPixel = .true.

!if (checkPixel) then
!write(*, *) "Line 5596 source%nossource pixels =", source%noSourcepixels
!end if



    if (maxval(pixel%noNeighbours) .le. 6 / binx) then

!if (checkPixel) then
!write(*, *) "Rejecting source at line 4597"
!endif



        IsValidextendedsource  = .false.
        return
    end if




 !   if ( CheckForMod8Pattern1( pixel, 2000.0, int(60 / binx, int8) ) ) then
!if(check) then
!write(*, *) "Rejecting source at line 6494"
!end if
!        IsValidextendedsource  = .false.
!        return
!    end if         
    
  





    !*******************************************
    ! Compute the source parameters
    !*******************************************

    source%rawCountRate = -maxval(pixel%counts)

!write(*, *) "Line 5876 calling computeextendedsourceparams rawcuntrate=", srcParams%rawCountRate

    call ComputeExtendedSourceParameters(sourceNumber, &
                                         pixel,        &
                                         source)


    !********************************************************************
    !
    !
    !*******************************************************************
    if (.not. sourceAreaOkay(source, size(pixel), 2.0) ) then

        IsValidextendedsource  = .false.
        return
    end if







    if ( surroundingBrightSource( source ) ) then
        
        IsValidextendedsource  = .false.
        return
    end if



!if (checkPixel) then
!write(*, *) "Line 4716 imin=", source%imin, " imax=", source%imax
!write(*, *) "Line 4717 jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "number=", source%number, " no of pixels=", size(pixel)
!write(*, *) "source%nossource pixels =", source%noSourcepixels
!end if


    if (.not. areaOkay(source, Pixel, 8.0)) then ! was 5 


!if (checkPixel) then
!write(*, *) "Rejecting source at line 4628"
!endif

        IsValidextendedsource  = .false.
         return
     end if

    
    i0 = max(lbx, int(source%xpos - 0.5, int16))
    j0 = max(lby, int(source%ypos - 0.5, int16))
    il = max(lbx, i0 - 5_int16)
    iu = min(nx1, i0 + 5_int16)
    jl = max(lby, j0 - 5_int16)
    ju = min(ny1, j0 + 5_int16)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then


!if (checkPixel) then
!write(*, *) "Rejecting source at line 4645"
!endif


        IsValidExtendedSource = .false.

        return
         
    end if
    



    n = size(pixel)
    if (n .lt. 2 * count(pLevelImagePtr(source%imin : source%imax, source%jmin : source%jMax) .gt. 0)) then 

!if (checkPixel) then
!write(*, *) "Rejecting source at line 4655"
!endif

        IsValidExtendedSource = .false.

        return
         
    end if
    
    
!if (checkPixel) then
!write(*, *) "4777 source%noSourcePixels =", source%noSourcepixels
!write(*, *) "Line 4777 source number = ", source%number
!write(*, *) "x=", source%xpos, " y=", source%yPos
!write(*, *) "Pixels-"

!do i = 1, size(pixel)
!write(*, *) i, pixel(i)%x, " ",  pixel(i)%y
!end do
!end if
    ! Check here to see if might be just mod-8 pattern
   
    width = source%iMax -  source%iMin
    height = source%jMax - source%jMin


    if (width .gt. ratio * height .or. height .gt. ratio * width) then 



!if (checkPixel) then
!write(*, *) "Rejecting source at line 4710"
!endif


        
        IsValidExtendedSource = .false.

        return
         
    end if


    
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel, starCounts)

!if (checkpixel) then
!write(*, *) "Line 4804 source number = ", srcParams%number
!write(*, *) "x=", srcParams%xpos, " y=", srcParams%yPos
!write(*, *) "No of pixels =", srcParams%noSourcepixels
!write(*, *) "total counts=", sum(pixel%counts), " starcounts=", starcounts
!end if




    noNeighbours = sum(pixel%noNeighbours)

    if (noNeighbours  .lt. source%noSourcePixels * 2) then 



!if (checkPixel) then
!write(*, *) "Rejecting source at line 4744"
!endif



        IsValidExtendedSource  = .false.
        return
    end if



!if (checkpixel) then
!write(*, *) "Line 4831 source number = ", source%number
!write(*, *) "x=", source%xpos, " y=", source%yPos
!write(*, *) "No of pixels =", source%noSourcePixels
!write(*, *) "noHoles=", noHoles, " ratio1=", ratio1
!end if

    if (maxval(imagePtr(source%imin : source%imax, source%jmin : source%jmax)) .gt. 2000.0) then
      


        if (source%noSourcePixels .lt. noHoles * ratio1) then    ! was 200             

!if (checkpixel) then
!write(*, *) "Rejecting source at line 4846 nosourcepixels=", source%noSourcePixels
!end if




            IsValidExtendedSource = .false.
            return
        end if 
    else 



        if (real(source%noSourcePixels) .lt. real(noHoles * 3)) then    ! was 3             

!if (checkpixel) then
!write(*, *) "Rejecting source at line 4692 nosourcepixels=", source%noSourcePixels
!end if



            IsValidExtendedSource = .false.
            return
        end if  
    end if
    

!if (checkPixel) then
!write(*, *) "Line 4864 noHoles=", noHoles, " nosourcepixels=", source%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2, " nostars=", nostars
!end if



    if (noStars .gt. 0) then
        if ( CountsRatio(pixel) .lt. 2.0) then


!if (checkPixel) then
!write(*, *) "Rejecting source at line 4813"
!endif




            IsValidExtendedSource = .false.
            return



        end if
    end if



    if (source%semiA / source%semiB .gt. 10) then

!if (checkPixel) then
!write(*, *) "Rejecting source at line 4840"
!endif


        IsValidExtendedSource = .false.
  
    else
                       
        IsValidExtendedSource = .true.
!if (checkPixel) then
!write(*, *) "Isvalidextendedsource line 6004 new source no=", source%number, " x=", source%xpos, " y=", source%yPos
!end if
    end if


!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", source%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

  
end Function IsValidExtendedSource

!***************************************************************
! CountsRatio
!***************************************************************
real(single) Function CountsRatio(pixel)
    type(PixelStruct) :: pixel(:) 
    integer :: i, x, y
    real(double) :: f1, f2

    f1 = 0.0_double
    f2 = 0.0_double
    do i = 1, size(pixel)
        x = pixel(i)%x
        y = pixel(i)%y
        if (pLevelImagePtr(x, y) .ne. 0) then
            f1 = f1 + imagePtr(x, y)
        else
            f2 = f2 + imagePtr(x, y)
        end if

    end do

    if ( f1 .ge. 1.0) then
        CountsRatio = real( f2 / f1)
    else
        CountsRatio = 1e6
    end if


end Function CountsRatio





!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended split
! source lie within certain limits
!*****************************************************
logical Function IsValidSplitSource(source, pixel)
     
 
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    
    integer(int16)  :: noHoles, noNeighbours, noStars
    integer(int16) :: width, height 

    integer(int16), parameter :: ratio = 7
    integer(int8) :: dx
    integer(int16) :: i0, j0, il, iu, jl, ju
    integer i, n, k, i1(1)
    integer, parameter :: minSize = 5     



    if (sum(pixel%counts, pixel%counts .ge. 0.0) .lt. minimumCounts) then
        IsValidSplitSource = .false.

        return
         
    end if

    n = size(pixel)

!write(*, *) "Isvalidsplitsource line 6160 n=", n

    k = 0

    do i = 1, n
        k = k + count(plevelImagePtr( max(lbx, pixel(i)%x - 1_int16) : min(nx1, pixel(i)%x + 1_int16), &     
                                      max(lby, pixel(i)%y - 1_int16) : min(ny1, pixel(i)%y + 1_int16)) &
                                      .gt. 0)            
    end do

!write(*, *) "Isvalidsplitsource line 6170 n=", n

    if (k .gt. n / 4) then

        IsValidSplitSource = .false.

        return
         
    end if
    
 
!write(*, *) "Isvalidsplitsource line 6181 n=", n


    !*******************************************
    ! Compute the source parameters
    !*******************************************
   
    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)

   
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)


    width = source%imax - source%imin

    height = source%jmax - source%jmin


    !***************************************************
    ! Locate source-pixel with the most counts
    !**************************************************
    i1 = maxloc(pixel%counts) 
    i0 = pixel(i1(1))%x  ! X coordinate of maximum
    j0 = pixel(i1(1))%y  ! Y coordinare of maximum
    source%xpos = real(i0 + 1)
    source%ypos = real(j0 + 1)




    if (n .lt. minSize) then



!write(*, *) "Isvalidsplitsource line 6201 n=", n
!write(*, *) "Imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax



        source%rawCountRate = -maxval(pixel%counts)

        
        source%extensionFlag = 0
        
        call ComputePointSourceParameters(int(source%number, int16), nx1, ny1, &
                         imagePtr, backgroundImagePtr,                         &
                         i0, j0, source)

!write(*, *) "Line 6216"
        source%noSourcePixels = n
    else

        source%extensionFlag = 1

   
!write(*, *) "Isvalidsplitsource line 6220 n=", n


        call ComputeExtendedSourceParameters(int(source%number, int16), &  ! The source number
                                             pixel,                     &  ! The pixels for this source
                                             source)                       ! The computed moments



        !********************************************************************
        !
        !
        !*******************************************************************
        if (.not. sourceAreaOkay(source, size(pixel), 2.0) ) then

            IsValidSplitsource  = .false.
            return
        end if






    end if


!write(*, *) "Isvalidsplitsource line 6231 n=", n


    if (.not. areaOkay(source, Pixel, 5.0)) then 
        IsValidSplitSource  = .false.
        return
    end if


    dx = max(4 / binx, max(width, height) / 4)

    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidSplitSource = .false.

        return
         
    end if
  
  
!   write(*, *) "Isvalidsplitsource line 6258 n=", n


    if (maxval(imagePtr(il : iu, jl : ju)) .gt. maxval(pixel%counts)) then
        IsValidSplitSource = .false.
!        write(*, *) "Isvalidsplitsource line 6119 rejecting source  n=", n
        return
         
    end if
    

  
    if (count(pixel%x .eq. i0 .and. pixel%y .eq. j0) .eq. 0) then

!  write(*, *) "Isvalidsplitsource line 6131 rejecting source  n=", n

        IsValidSplitSource = .false.

        return
         
    end if
    

!write(*, *) "Line 5031 isvalidsplitsource source number = ", source%number
!write(*, *) "x=", source%xpos, " y=", source%yPos
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax


!write(*, *) "Pixels-"
!do i = 1, size(pixel)
!write(*, *) i, pixel(i)%x, " ",  pixel(i)%y
!end do

    ! Check here to see if might be just mod-8 pattern
   
    width = source%iMax -  source%iMin
    height = source%jMax - source%jMin


    if (width .gt. ratio * height .or. height .gt. ratio * width) then 
        

 !write(*, *) "Isvalidsplitsource line 6163 rejecting source  n=", n



        IsValidSplitSource = .false.

        return
         
    end if
    
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel)

    noNeighbours = sum(pixel%noNeighbours)
    if (noNeighbours  .lt. source%noSourcePixels * 2) then 



!  write(*, *) "Isvalidsplitsource line 6181 rejecting source  n=", n


        IsValidSplitSource  = .false.
        return
    end if


    if (source%noSourcePixels .lt. noHoles * ratio1) then    ! was 200             

!write(*, *) "Rejecting source"
!write(*, *) "Line 3800 noHoles=", noHoles, " nosourcepixels=", source%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

        IsValidSplitSource = .false.
        return
    end if  

           


!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", source%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2



    if (source%noSourcePixels .lt. noStars * ratio2) then              ! was 50

!write(*, *) "Line 6208 rejecting source, noStars=", noStars, " ratio2=", ratio2

        IsValidSplitSource = .false.           
        return
    end if



    if (source%semiA / source%semiB .gt. 10) then
        IsValidSplitSource = .false.
  
    else
                       
        IsValidSplitSource = .true.

        source%rawCountRate = -maxval(pixel%counts)

        call ClassifySource(source, imagePtr, backgroundImagePtr)
   


!write(*, *) "new source no=", source%number, " x=", source%xpos, " y=", source%yPos

    end if

!write(*, *) "Line 1494 noHoles=", noHoles, " nosourcepixels=", source%noSourcePixels
!write(*, *) "ratio1=", ratio1, " ratio2=", ratio2

  
end Function IsValidSplitSource

!*****************************************************
! IsValidExtendedSource
! Checks to see if the parameters of an extended split
! source lie within certain limits
!*****************************************************
logical Function IsValidSplitGrismSource(srcParams, pixel)
    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:) 
    
    integer(int16), parameter :: ratio = 7
    integer(int16), parameter :: dx = 2
    integer(int16) :: i0, j0, il, iu, jl, ju
    integer n
    
    if (sum(pixel%counts, pixel%counts .ge. 0.0) .lt. minimumCounts) then
        IsValidSplitGrismSource = .false.

        return
         
    end if


    if (.not. IsValidFirstOrderSource(srcParams, pixel)) then
        IsValidSplitGrismSource = .false.
        return
    end if


    n = size(pixel)

    if (.not. areaOkay(srcParams, Pixel, 5.0)) then 
        IsValidSplitGrismSource  = .false.
        return
    end if


    i0 = max(lbx, int(srcParams%xpos - 0.5, int16))
    j0 = max(lby, int(srcParams%ypos - 0.5, int16))


    il = max(lbx, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(lby, j0 - dx)
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
        IsValidSplitGrismSource = .false.

        return
         
    end if
  
    
    if (count(pixel%x .eq. i0 .and. pixel%y .eq. j0) .eq. 0) then

        IsValidSplitGrismSource = .false.

    else         
        IsValidSplitGrismSource = .true.

    end if
    

  
end Function IsValidSplitGrismSource

!***********************************************************************
!
!**********************************************************************
logical Function areaOkay(source, pixel, constant)

    type(SourceParameters), intent(in) :: source
    type(PixelStruct) :: pixel(:) 
    real(single), intent(in) :: constant
    real :: area     

    area = real( (source%imax - source%imin + 1) * (source%jmax - source%jmin + 1)) 
 
    if (( size(pixel) * constant) .lt. area) then ! Was 5 
        areaOkay = .false.

    else

        areaOkay = .true.
    end if


!if(checkPixel) then
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "n=", size(pixel), " constant=", constant, " area=", area
!write(*, *) "areaokay=", areaokay
!end if
    
end function areaOkay
!*****************************************************
!
! IsValidSmallExtendedSource
! For "small" extended sources
!
!*****************************************************
logical Function IsValidSmallExtendedSource(source, pixel)
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:) 
    
    real(single), parameter :: delta1 = 1.1, delta2 = 6  ! was 1.2
    integer(int16) :: n, noHoles, i0, j0, il, iu, jl, ju, noStars
    integer(int8) :: dx
    real(single) :: ratio  
    type(PixelStruct) :: maxStarPixel


!    logical :: check
  
!checkPixel = .false.

!if (any( abs(pixel%x - ipixel) < 10) .and. any(abs(pixel%y - jpixel) < 10)) then
!    checkPixel = .true.
!    write(*, *) "is valid smallextendedsoure Line 6224 Checking possible smallextended source at ", &
!       ipixel, ", ", jpixel, " number of pixels=", size(pixel)
!check = .true.
!end if 

    if (maxval(pixel%noNeighbours) .lt. 4) then


!if(checkPixel) then
!write(*, *) "Rejecting source at line 6410"
!end if


        IsValidsmallextendedsource  = .false.
        return
    end if


    il = minval(pixel%x)
    iu = maxval(pixel%x)
    jl = minval(pixel%y)
    ju = maxval(pixel%y)
    n = size(pixel)
   


    !************************************************************************
    !
    ! Check number of neighbours
    !
    !***********************************************************************
    select case (n)
       

        case( 4 : 10) 
            

            if ( count(pixel%noNeighbours > 1) < (n / 2 - 1) .or.  &
                 count(pixel%noNeighbours > 2) < (n / 4 - 1)) then
          
    
!if (checkPixel) then
!write(*, *) "rejecting source at line 6990 n=", n, "  count(pixel%noNeighbours > 1) < n / 2=",  &
!                count(pixel%noNeighbours > 1)
!end if


                IsValidSmallExtendedSource  = .false.
                return
      
            end if

        case (11 :)
   

            if ( count(pixel%noNeighbours > 3) < (n / 4 - 1)) then



!if (checkPixel) then                              
!write(*, *) "rejecting source at line 6653 n=", n, "  count(pixel%noNeighbours > 3)=", & 
!count(pixel%noNeighbours > 3)
!end if


                IsValidSmallExtendedSource  = .false.
                return
      
            end if
        case default
    end select

    dx = 20 / binx


!if(checkPixel) then
!write(*, *) "Line 6509 in isvalidsmallextendedsource"
!end if

 !   if ( CheckForMod8Pattern1( pixel, 2000.0, int(60 / binx, int8) ) ) then
!if(checkPixel) then
!write(*, *) "Rejecting source at line 6494"
!end if
!        IsValidsmallextendedsource  = .false.
!        return
!    end if         
     


    if ( .not. CheckVicinity3(pixel, dx, maxval(pixel%counts * 100.0), 4.0)) then

!if(checkPixel) then
!write(*, *) "Rejecting source at line 6250"
!end if 


        IsValidsmallextendedsource  = .false.
        return
    end if         
     


    !*************************************
    ! Compute the source parameters
    !************************************
    source%rawCountRate = -maxval(pixel%counts)

    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)




    !********************************************************************
    !
    !
    !*******************************************************************
    if (.not. sourceAreaOkay(source, size(pixel), 2.0 )) then

        IsValidsmallextendedsource  = .false.
        return
    end if






    !******************************************************************
    ! Added on January 30th
    !*****************************************************************


    if ( surroundingBrightSource( source ) ) then
 

       
        IsValidSmallextendedsource  = .false.
 

!if(checkPixel) then
!write(*, *) "Rejecting source at line 6415"
!end if

       return
    end if




    if( centreInsideSource(source, pixel) ) then
         

        IsValidSmallextendedSource = .false.

!if(checkPixel) then
!write(*, *) "Rejecting source at line 6429"
!end if



        return
    end if





    i0 = max(0_int16, int(source%xpos - 0.5, int16))
    j0 = max(0_int16, int(source%ypos - 0.5, int16))




    dx = 8 / binx   
    il = max(0_int16, i0 - dx)
    iu = min(nx1, i0 + dx)
    jl = max(0_int16, j0 - dx)
    ju = min(ny1, j0 + dx)


  




    if ( any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then



!if (checkPixel) then
!            write(*, *) "isvalidsmallextebdedsource - Line 6285, i0=", i0, " j0=", j0
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if
 

        IsValidSmallExtendedSource = .false.


!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "levelimage.fits", "testimage")
!call fatal("stopped at line 6301 displaying levelimage.fits")


        return
         
     
  
    end if
 


   

    if (.not. areaOkay(source, Pixel, 4.0)) then

!if (checkPixel) then
!write(*, *) "isvalidsmallextebdedsource - Line 6248rejecting source area not okay"
!end if
 
        IsValidsmallextendedsource  = .false.
        return
    end if


  
   
    ! Check here to see if might be just mod-8 pattern

    noHoles = NumberOfHoles(pixel)

    if ((noHoles * 2) .gt. n) then

!if(checkPixel) then
!write(*, *) "Rejecting source at line 6287"
!end if



        IsValidSmallExtendedSource = .false.           
        return
    end if 





    !***********************************************************
    ! Check if any point-sources within it
    !**********************************************************
    noStars = NumberOfStars( source, maxStarPixel )

    dx = 20 / binx
    if (abs(i0 - maxStarPixel%x) < dx .and. &
        abs(j0 - maxStarPixel%y) < dx) then

        IsValidSmallExtendedSource = .false.

!if(checkPixel) then
!write(*, *) "******** Rejecting source at line 6238 nostars=", nostars

!write(*, *) "i0=", i0, " j0=", j0, " xc=", source%xpos, " y=", source%ypos, " star i0=", &
!maxstarPixel%x, " j0=", maxstarPixel%y, " counts=", maxstarpixel%counts
!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage(plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 6243 displaying e & p levelimage.fits")
!end if
           
        return


    end if
      

!if (checkPixel) then
!write(*, *) "isvalidsmallextendedsourceLine 5220 source no ", source%number, " xpos=", source%xpos, " y=", source%ypos
!write(*, *) "no pixels=", source%noSourcePixels, " no holes=", noHoles, " nostars=", nostars, " ratio1=", ratio1
!write(*, *) "a=", source%semiA, " b=", source%semiB
!end if

    if (noStars .gt. 1) then ! Was 0
        IsValidSmallExtendedSource = .false.

!if(checkPixel) then
!write(*, *) "Rejecting source at line 5667 nostars=", nostars
!end if
           
        return
    end if




!if (checkPixel) write(*, *) "Line 5577, noholes=", noholes, " size pixel=", size(pixel) 
      
 

    if (source%semiA .lt. 1.0) then


!if(checkPixel) then
!write(*, *) "Rejecting source at line 5698"
!end if



        IsValidSmallExtendedSource = .false.           
        return
    end if
 

    ratio = source%semiA / source%semiB 
    
    if (ratio .gt. delta2) then


!if(checkPixel) then
!write(*, *) "Rejecting source at line 5715"
!end if




        IsValidSmallExtendedSource = .false.           

        return
    end if 



    if ( .not. CheckVicinity(pixel, 2.0, int(6 / binx, int8))) then  ! was 1
    
!if(checkPixel) then
!write(*, *) "Line 6317 ********************** AFTER CALLING CHECKVICINITY ************************"
!write(*, *) "Rejecting source at line 6318"
!end if

        IsValidSmallExtendedSource = .false.           

        return
    end if 


!if(checkPixel) then
!write(*, *) "new source"
!end if

    !*************************************************
    ! Reach here then new extended source
    !************************************************
    IsValidSmallExtendedSource = .true.           
    

!if (checkPixel) then
!write(*, *) "New source - IsValidSmallextendedsource- source no=", source%number
!write(*, *) "num pixerls=", source%noSourcePixels
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) " noholes=", noholes, " extensionflag=", source%extensionflag
!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage(plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 5353 displaying e & p levelimage.fits")
!end if 

end Function IsValidSmallExtendedSource


!*****************************************************
!
! IsValidFaintource
! For faint point/extended sources
!
!*****************************************************
logical Function IsValidFaintSource(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
  
    real(single), parameter :: delta = 5.0   ! was 1.5
    real(single) :: sigNoise, fTotal, fBackground, radius, area 
    integer(int16) :: n, i1(1), i0, j0
   
    integer(int8) :: minSize = 5
    
    integer i  
 
    real(single) :: fMin, fMax
  
      
    minSize = 20 / binx
    n = size(pixel)    
 


!     if (any( abs( pixel%x - ipixel) < 10 / binx) .and. any(abs(pixel%y - jpixel) < 10 / binx) &
!              .and. (ipixel >= 0 .and. jPixel >= 0)) then     
!         checkPixel = .true.

!write(*, *) "*************************"
!    write(*, *) "isvalidefaintsource line 6874 checking pixel ", ipixel, ", ", jpixel
!write(*, *) "Pixels"
!do i = 1, n
!write(*, *) i, ") pixel=", pixel(i)
!end do
!write(*, *) "************************************"
!     else
!         checkpixel = .false.      
!     end if



    source%imin = minval(pixel%x)
    source%imax = maxval(pixel%x)
    source%jmin = minval(pixel%y)
    source%jmax = maxval(pixel%y)



    if(n > 100 / binx) then
        if (.not. areaOkay(source, Pixel, 4.0)) then 
            IsValidFaintSource  = .false.
 
!            if(checkPixel) then
!                 write(*, *) "Rejecting source at line 6898"
!             end if  
            return
        end if
    end if

    !*************************************************************************
    !
    ! Check for 1 pixel-wide source
    !
    !*************************************************************************
    if ( n .gt. 3 ) then
        if (source%imin .eq. source%imax .or. source%jmin .eq. source%jmax) then


!if (checkPixel) then
!write(*, *) "rejecting source at line 6919"
!end if

            IsValidFaintSource = .false.
            return 
        end if
   
    end if

    !************************************************************************
    !
    ! Check number of neighbours
    !
    !***********************************************************************
    select case (n)
        case(1 : 3)

            if ( any(pixel%noNeighbours < 1) ) then

               
!if (checkPixel) then
!do i = 1, n
!write(*, *) i, " x=", pixel(i)%x, " y=", pixel(i)%y, " noneighbours=", pixel(i)%noneighbours
!end do                    

!  write(*, *) "rejecting source at line 6961 n=", n, "  count(pixel%noNeighbours > 1) < n / 2=",  &
!  count(pixel%noNeighbours > 1)
!  end if


                IsValidFaintSource  = .false.
                return
      
            end if





        case( 4 : 15)   ! was 10
            

            if ( count(pixel%noNeighbours > 1) < (n / 2 - 1) .or.  &
                 count(pixel%noNeighbours > 2) < (n / 4 - 1)) then
          
    
!if (checkPixel) then
!write(*, *) "rejecting source at line 6990 n=", n, "  count(pixel%noNeighbours > 1) < n / 2=",  &
!count(pixel%noNeighbours > 1)
!end if


                IsValidFaintSource  = .false.
                return
      
            end if

        case (16 :)
   

            if ( count(pixel%noNeighbours == 4) < n / 5 ) then


!if (checkPixel) then                    
!write(*, *) "rejecting source at line 6871"
!write(*, *) "n=", n, "  count(pixel%noNeighbours == 4) ==",  &
!count(pixel%noNeighbours == 4)
!end if


                IsValidFaintSource  = .false.
                return
      
            end if
        case default
   end select







    !***************************************************
    !
    ! Locate source-pixel with the most counts
    !
    !**************************************************
    i1 = maxloc(pixel%counts) 
 
    i0 = pixel(i1(1))%x  ! X coordinate of maximum
    j0 = pixel(i1(1))%y  ! Y coordinare of maximum


    source%xPos = real(i0 + 1) 
    source%yPos = real(j0 + 1)


    !*********************************************************
    !
    ! Check the signal-to-noise
    !
    !*********************************************************
    radius = real(4 / binx)
    call getCounts( imagePtr, source%xpos, source%ypos, radius, fTotal, area)        

!if(checkPixel) then
!  write(*, *) "Line 6978 ftotal=", ftotal, " x=", source%xpos, " y=", source%ypos, " radius=", radius
!end if

    if(fTotal > 0.0) then
        call getCounts( backgroundImagePtr, source%xpos, source%ypos, radius, fBackground)        
   
       sigNoise = (fTotal - fBackground) / sqrt(fTotal)
 
    else
        sigNoise = 0.0
    end if

   
    if (sigNoise < detectParams%minSigNoise ) then
 

!if (checkPixel) then
!write(*, *) "fTotal=", fTotal, " fbackground=", fbackground
!write(*, *) "sum pixel%counts=", sum(pixel%counts), &
!"sum pixel%counts - pixel%backgroundcounts=",  sum(pixel%counts - pixel%backgroundCounts) 
!write(*, *) "rejecting source at line 5561, signoise=", signoise
!end if



        IsValidFaintSource = .false.
        return 
    end if




!    if ( CheckForMod8Pattern1( pixel, 2000.0, int(60 / binx, int8) ) ) then
!if(checkPixel) then 
!write(*, *) "Rejecting pixel at line 6873" 
!end if 
!        IsValidFaintsource  = .false.
!        return
!    end if         
     






  
    fMin = minval(pixel%counts)
    fMax = maxval(pixel%counts)
  
     
    !*************************************************************************
    !
    ! Check surrounding pixels
    !
    !*************************************************************************    
    if ( .not. CheckVicinity(pixel, int( 2 / binx, int8), fmin, 1.0, .true.)) then  ! NB WAS FMIN



!if(checkPixel) then
!write(*, *) "Rejecting pixel at line 7051 signoise=", sigNoise
!call saveimage( elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage( plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 7213  omdetectdetection_mod.f90  displaying elevelimage.fits and plevelimage.fits")
!end if


         IsValidFaintSource = .false.        
         return 
        
    end if






    if ( surroundingBrightSource( source ) ) then
        
        IsValidFaintsource  = .false.
        return
    end if







    if(sigNoise < 100.0) then 
        !************************************************************************
        !
        ! Check surrounding pixels
        ! 
        !************************************************************************    
        if ( .not. CheckVicinity(pixel, int( 6 / binx, int8), fmax * 3.0 , 4.0, .false.)) then



!if(checkPixel) then
!write(*, *) "Rejecting pixel at line 7074 signoise=", sigNoise

!call saveimage( elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage( plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 7213  omdetectdetection_mod.f90  displaying elevelimage.fits and plevelimage.fits")
!end if

            IsValidFaintSource = .false.

        
            return 
        
        end if

        !*************************************************************************
        ! Check surrounding pixels
        !************************************************************************    
!        if ( .not. CheckVicinity(pixel, int( 20 / binx, int8), 50.0 * fmax, 4.0)) then

!if(checkPixel) then
!write(*, *) "Rejecting pixel at line 7095 signoise=", sigNoise

!call saveimage( elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage( plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 7213  omdetectdetection_mod.f90  displaying elevelimage.fits and plevelimage.fits")
!end if

!            IsValidFaintSource = .false.
!            return 
!        endif



 
  



!if(checkPixel) then
!write(*, *) "Line 7109 signoise=", signoise, " fmax=", fmax
!end if

        !*************************************************************************
        !
        ! Check surrounding pixels was 100 * fmax
        !
        !************************************************************************    
 !       if ( .not. CheckVicinity(pixel, int( 30 / binx, int8), 100 * fmax, 4.0)) then

!if(checkPixel) then
!write(*, *) "Rejecting pixel at line 7129 signoise=", sigNoise

!call saveimage( elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage( plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 7213  omdetectdetection_mod.f90  displaying elevelimage.fits and plevelimage.fits")
!endif


!            IsValidFaintSource = .false.
        
!            return 
        
!        endif

    end if ! If signoise




    if (n .lt. minSize) then


        source%rawCountRate = -maxval(pixel%counts)

!if(checkPixel) then
!write(*, *) "calling computepointsourceparameters i0=", i0, " j0=", j0, " image=", imagePtr(i0, j0)
!write(*, *) "rawcountrate=", source%rawCountRate, " n=", n
!end if
        
        source%extensionFlag = 0
        
        call ComputePointSourceParameters(int(source%number, int16), nx1, ny1, &
                         imagePtr, backgroundImagePtr,                         &
                         i0, j0, source)


        source%noSourcePixels = n
    else

        source%extensionFlag = 1

   

        call ComputeExtendedSourceParameters(int(source%number, int16), &  ! The source number
                                             pixel,                     &  ! The pixels for this source
                                             source)                       ! The computed moments






        !********************************************************************
        !
        !
        !*******************************************************************
        if (.not. sourceAreaOkay(source, size(pixel), 2.0) ) then

            IsValidFaintsource  = .false.
            return
        end if






        if (.not. isNullValue(source%semia) ) then
            

    
            if( ( source%semiA > real(50 / binX)) .or. source%semia .gt. ( 10.0 * source%semib ) ) then
                do i = 1, size(pixel)
                    elevelImagePtr( pixel(i)%x, pixel(i)%y ) = 1_int16

                end do
  !              if(checkPixel) then
  !                  write(*, *) "REjecting source at line 7042"
  !              end if  

                isValidFaintSource = .false.
                return
          
            end if 

        end if
    end if


    if( centreInsideSource(source, pixel) ) then
         
!if (checkPixel) then
!    write(*, *) "rejecting source at line 7206 source%xpos=", source%xpos, " y=", source%ypos
!end if


        IsValidFaintSource = .false.
        return
    end if






    if ( abs(source%semia) < 0.1) then

!if (checkPixel) then
!write(*, *) "rejecting source at line 5597 - semia=", source%semia
!end if

        isValidFaintSource = .false.

            
        return
    end if



 

    if( count(plevelImagePtr(source%imin : source%imax, source%jmin : source%jmax) > 0) .gt. n / 10) then
 
!        if(checkPixel) then
!             write(*, *) "Rejecting source at line 7086"
!             write(*, *) "source imin=", source%imin, " imax=", source%imax
!             write(*, *) "source jmin=", source%jmin, " imax=", source%jmax
!             write(*, *) "No of pixels=", n, " no of p pixels within that area=", &
!                   count(plevelImagePtr(source%imin : source%imax, source%jmin : source%jmax) > 0)
!         end if

        IsValidFaintSource = .false.           
        return
    end if


! if(checkPixel) then
! write(*, *) "imin=", source%imin, " imax=", source%imax
! write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
! write(*, *) "count(plevelimagePtr)=",  &
! count(plevelImagePtr(source%imin : source%imax, source%jmin : source%jmax) > 0), "n=", n 
! end if

        
    IsValidFaintSource = .true.           
    source%rawCountRate = -maxval(pixel%counts)

    if( n .ge. 10 / binx) then
        call ClassifySource(source, imagePtr, backgroundImagePtr)
    else
        source%extensionFlag = 0
    end if 



!if(source%xpos < 0.0 .or. source%ypos < 0.0) check = .true.

!if (checkPixel) then
!write(*, *) "Isvalidfaintsource line 7062"
!write(*, *) "new source x=", source%xpos, " y=", source%ypos, " no pixels=", n, " i0=", i0, " j0=", j0
!write(*, *) "ftotal=", ftotal, " fbackground=", fbackground
! write(*, *) "imin=", source%imin, " imax=", source%imax, " jmin=", source%jmin, " imax=", source%jmax
! write(*, *) "signoise=",  (sum(pixel%counts - pixel%backgroundcounts) / sqrt(sum(pixel%counts)))
! write(*, *) "extensionflag=", source%extensionFlag, " numpixels=", source%noSourcePixels, " number=", sourcenumber
! write(*, *) "Semi-a=", source%semia, "+/-", source%semiaerr, " b=", source%semib, "+/-", source%semiberr
!call fatal("stopped")

!call saveimage( elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage( plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("stopped at line 7307  omdetectdetection_mod.f90  displaying elevelimage.fits and plevelimage.fits")
!end if

end function isValidFaintSource

!**************************************************************
!
! Check that the computed centre of the source is
! within the source region, to prevent 2 sources being
! detected twice
!
!*************************************************************
logical function centreInsideSource(source, pixel)


    type(SourceParameters), intent(in) :: source
    type(PixelStruct) :: pixel(:)
    integer(int32) :: i, j, count1, count2 
    integer(int16) :: i0, j0
    real(single) :: dx, dy, minRadius, radius
    integer :: i2(2)

!centreInsideSource = .false.
!return
    count1 = 0
    count2 = 0
    do j = source%jmin, source%jmax

        do i = source%imin, source%imax
            if( pLevelImagePtr( i, j) .ne. 0 ) count1 = count1 + 1
            if( eLevelImagePtr( i, j) .lt. 0 ) count2 = count2 + 1


        end do
    end do 

!if(checkPixel) then
!    write(*, *) "centreinsidesource line 71117 count1=", count1, " count2=", count2
!end if

    if(count2 .gt. 0) then
        centreInsideSource = .true.
        return
    end if



    if(count1 .eq. 0) then
        centreInsideSource = .false.
        return
    end if


    !********************************************************************************
    ! Find the location of the maximum point-source pixel
    !
    !*******************************************************************************

    i2 = maxloc(imagePtr(source%imin : source%imax, source%jMin : source%jmax), &
                mask=plevelImagePtr(source%imin : source%imax, source%jmin : source%jmax) > 0)

    i0 = i2(1) + source%imin - 1
    j0 = i2(2) + source%jmin - 1


!if(checkPixel) then
!    write(*, *) "centreinsidesource line 7145 i0=", i0, " j0=", j0
!end if



    minRadius = 1e6
  
    do i = 1, size(pixel)

        dx = real(pixel(i)%x - i0 )
        dy = real(pixel(i)%y - j0 )
        radius = sqrt(dx * dx + dy * dy)
        if(radius .lt. minRadius) then
            minRadius = radius
        end if 
    end do


!if(checkPixel) then
!    write(*, *) "centreinsidesource line 7165 minradius=", minradius
!end if


    if(minRadius .lt. real(4 / binx) ) then

        centreInsideSource = .true.

    else
        centreInsideSource = .false.

    end if


end function centreInsideSource




!**************************************************************
!
! Check that the computed centre of the source is
! within the source region, to prevent 2 sources being
! detected twice
!
!*************************************************************
logical function surroundingBrightSource(source)


    type(SourceParameters), intent(in) :: source
   
    integer(int32) :: i, j
    integer(int16) :: il, iu, jl, ju, dx
    real(double) :: eCounts, pCounts
    integer :: count1, count2, count3

    dx = 20 / binX
    il = max(0_int16, source%imin - dx)
    iu = min(nx1, source%imax + dx)

    jl = max(0_int16, source%jmin - dx)
    ju = min(ny1, source%jmax + dx)


    !**************************************************************
    ! Find the maximum pixel value inside the source not connected
    ! with a source
    !**************************************************************

    pCounts = 0.0_double
    eCounts = 0.0_double
 
    count1 = 0
    count2 = 0
    count3 = 0 ! Records how many bright pixels there are
    do j = jl, ju

        do i = il, iu
            
            if( eLevelImagePtr( i, j) .eq. number ) then
                count1 = count1 + 1
                if( .not. isNullValue(imagePtr(i, j)) ) then
                    eCounts = eCounts + dble( imagePtr(i, j) )
                end if  
            end if
            if( pLevelImagePtr( i, j) .gt. 0 .or. elevelImagePtr(i, j) .eq. 1) then
     
                count2 = count2 + 1
                if( .not. isNullValue(imagePtr(i, j)) ) then
     
                    if( imagePtr(i, j) .gt. 3000.0) then
                        count3 = count3 + 1
                    else 
                        pCounts = pCounts + dble( imagePtr(i, j) ) 
                    end if
                end if
            endif
        end do
    end do 


!if(checkPixel) then
!write(*, *) "imin=", source%imin, " imax=", source%imax
!write(*, *) "jmin=", source%jmin, " jmax=", source%jmax
!write(*, *) "count1=", count1, " count2=", count2, " number=", number
!write(*, *) "ecounts=", ecounts, " pcounts=", pcounts
!end if


    if(count3 .gt. 6 / binX) then

        surroundingBrightSource = .true.
        return
    end if



    if(count1 .lt. 4 * count2) then

        surroundingBrightSource = .true.
        return
    end if



    if( eCounts .lt. 2.0 * pCounts ) then
        surroundingBrightSource = .true.
    else
        surroundingBrightSource = .false.
    end if 
 

    
end function surroundingBrightSource


!************************************************************************************
!
!
!***********************************************************************************
logical Function CheckVicinity1(source, pixel)
     
    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
    integer(int8) :: dx1, dx2, dx3
    integer(int16) :: i, j, il, iu, jl, ju
    real(single) :: maxSource, maxSource1, sumCounts
    logical :: thin
!    logical :: check


    sumCounts = sum(pixel%counts)

    maxSource = sumCounts * 150.0
    maxSource1 = sumCounts * 250.0





    if( minval(pixel%x) == maxval(pixel%x) .or. &
        minval(pixel%y) == maxval(pixel%y)) then
        thin = .true. ! Just a thin line of pixels
    else
        thin = .false.
    end if


!if (any(pixel%x .eq. ipixel) .and. any(pixel%y .eq. jpixel)) then
!        check = .true.
!write(*, *) "isvalidefaintsopurce line 5380 checking pixel ", ipixel, ", ", jpixel
!else
!check = .false.
!end if



    if (source%semia / source%semib > 4.0 .or. thin .or. size(pixel) < 4 / binx) then

        dx1 = 6 / binx
        dx2 = 40 / binx

    else if (source%semia / source%semib > 2.0 ) then

        dx1 = 4 / binx
        dx2 = 30 / binx



    else
        dx1 = 2 / binx
        dx2 = 12 / binx
    end if
   
    dx3 = 30 / binx

    !***************************************************************
    !
    ! Check pixels to the left or right of the source for bright feature
    !
    !**************************************************************
    do j = minval(pixel%y), maxval(pixel%y)


        !*********************************************************
        ! Check for point sources
        !******************************************************** 

        il = max(0_int16, minval(pixel%x, pixel%y .eq. j) - dx1)
        iu = min(nx1, maxval(pixel%x, pixel%y .eq. j) + dx1)
       

        do i = il, iu
        
            if (plevelImagePtr(i, j) .ne. 0) then
      
                checkVicinity1 = .false.

!if (check) then         
!write(*, *) "rejecting source at line 5427 i=", i, " j=", j, " plevelimageptr=", plevelImagePtr(i, j)

!call saveimage(elevelimagePtr, nx1+1_int16, ny1+1_int16, "elevelimage.fits", "testimage")
!call saveimage(plevelimagePtr, nx1+1_int16, ny1+1_int16, "plevelimage.fits", "testimage")
!call fatal("displaying elevelimage.fits and plevelimage.fits")

 !end if                 
                return


           


           
            end if 
        end do


        !****************************************************************
        ! Check to the left for a moderately bright source
        !***************************************************************
        il = max(0_int16, minval(pixel%x, pixel%y .eq. j) - dx2)
        iu = max(0_int16, minval(pixel%x, pixel%y .eq. j) - 1_int16) 
         
        do i = il, iu
        
            if ( imagePtr(i, j) .ge. maxSource) then
                checkVicinity1 = .false.


!if (check) then
!write(*, *) "i=", i, " j=", j, " image=", imagePtr(i, j), " maxsources=", maxSource
!write(*, *) "rejecting source at line 5449, dx2=", dx2, " il=", il, " iu=", iu
!end if


                return
            end if 
        end do


        !****************************************************************
        ! Check to the left for a very bright source
        !***************************************************************
        il = max(0_int16, minval(pixel%x, pixel%y .eq. j) - dx3)
  
     

        do i = il, iu
        
            if ( imagePtr(i, j) .ge. maxSource1) then
                checkVicinity1 = .false.


!if (check) then
!write(*, *) "i=", i, " j=", j, " image=", imagePtr(i, j), " maxsources=", maxSource
!write(*, *) "rejecting source at line 5449"
!end if







                return
            end if 
        end do






        !****************************************************************
        ! Check to the right for a moderately bright source
        !***************************************************************
        il = min(nx1, minval(pixel%x, pixel%y .eq. j) + 1_int16)
        iu = min(nx1, minval(pixel%x, pixel%y .eq. j) + dx2) 




        do i = il, iu
        
            if ( imagePtr(i, j) .ge. maxSource) then
                checkVicinity1 = .false.

!if (check) then
!write(*, *) "rejecting source at line 5471"
!end if


                return
            end if 
        end do

  

        !****************************************************************
        ! Check to the right for a very bright source
        !***************************************************************
        iu = min(nx1, minval(pixel%x, pixel%y .eq. j) + dx3) 




        do i = il, iu
        
            if ( imagePtr(i, j) .ge. maxSource1) then
                checkVicinity1 = .false.

!if (check) then
!write(*, *) "rejecting source at line 5471"
!end if



                return
            end if 
        end do

    end do 






    !*****************************************************
    ! Check below or above for a very bright source
    !****************************************************

    do i = minval(pixel%x), maxval(pixel%x)
        jl = max(0_int16, minval(pixel%y, pixel%x .eq. i) - dx1)
        ju = min(ny1, maxval(pixel%y, pixel%x .eq. i) + dx1) 


        do j = jl, ju
        
            if (plevelImagePtr(i, j) .ne. 0) then

               
                checkVicinity1 = .false.

!if (check) then
!write(*, *) "rejecting source at line 5494"
!end if





                return
           
            end if 
        end do

        jl = max(0_int16, minval(pixel%y, pixel%x .eq. i) - dx2)
        ju = max(0_int16, minval(pixel%y, pixel%x .eq. i) - 1_int16) 
        


        do j = jl, ju
        
            if ( imagePtr(i, j) .ge. maxSource) then

!if (check) then
!write(*, *) "rejecting source at line 5513"
!end if



                checkVicinity1 = .false.
                return
            end if 
        end do


        jl = max(0_int16, minval(pixel%y, pixel%x .eq. i) - dx3)
   
        

        do j = jl, ju
        
            if ( imagePtr(i, j) .ge. maxSource1) then

!if (check) then
!write(*, *) "rejecting source at line 5513"
!end if



                checkVicinity1 = .false.
                return
            end if 
        end do


  

        jl = min(ny1, maxval(pixel%y, pixel%x .eq. i) + 1_int16)
        ju = min(ny1, maxval(pixel%y, pixel%x .eq. i) + dx2) 
        
!     write(*, *) "Line 5623 jl=", jl, " ju=", ju, " nx1=", nx1, " ny1=", ny1


        do j = jl, ju
        
            if ( imagePtr(i, j) .ge. maxSource) then


!if (check) then
!write(*, *) "rejecting source at line 5539"
!end if




                checkVicinity1 = .false.
                return
            end if 
        end do


        ju = min(ny1, maxval(pixel%y, pixel%x .eq. i) + dx3) 
  
        do j = jl, ju
        
            if ( imagePtr(i, j) .ge. maxSource1) then

!if (check) then
!write(*, *) "rejecting source at line 5539"
!end if




                checkVicinity1 = .false.
                return
            end if 
        end do



    end do

    checkVicinity1 = .true.


end Function CheckVicinity1

!**************************************************************
! Function neighbourDistance
!*************************************************************
logical function neighbourDistance(pixel, sourceNumber, levelImage)
    type(PixelStruct) :: pixel(:)
    integer(int16), intent(in) :: sourceNumber
    integer(int16), pointer :: levelImage(:, :)
    integer(int16) :: i1(1), i2(2), i0, j0
    integer(int16) :: il, iu, jl, ju

    real(single) :: dx, dy, height, gaussian, radius
    integer(int16), parameter :: dx1 = 100

    

    i1 = maxloc(pixel%counts) 
    i0 = pixel(i1(1))%x  ! X coordinate of maximum
    j0 = pixel(i1(1))%y  ! Y coordinare of maximum


    il = max(0_int16, i0 - dx1)
    jl = max(0_int16, j0 - dx1)
    iu = min( int(ubound(imagePtr, 1), int16), i0 + dx1)
    ju = min( int(ubound(imagePtr, 2), int16), j0 + dx1)


    i2 = maxloc(imagePtr(il : iu, jl : ju), mask=levelImage(il : iu, jl : ju) == sourceNumber)
    i2(1) = i2(1) + il - 1
    i2(2) = i2(2) + jl - 1


!if(checkPixel) then    
!    write(*, *) "Neighbourdistance line 7337 i2(1)=", i2(1), " i2(2)=", i2(2), &
!        " il=", il, " jl=", jl, " iu=", iu, " ju=", ju
!    write(*, *) "max pixel counts=", maxval(pixel%counts), " p sourcenumber=", sourceNumber 
!end if


    height = 0.75 * imagePtr(i2(1), i2(2))
    i2 = minloc(imagePtr(il : iu, jl : ju), mask=levelImage(il : iu, jl : ju) == sourceNumber &
              .and. imagePtr(il : iu, jl : ju) .ge. height)



!if(checkPixel) then
    
!    write(*, *) "Neighbourdistance line 7351 i2(1)=", i2(1), " i2(2)=", i2(2), " il=", il, " jl=", jl
!    write(*, *) "max pixel counts=", maxval(pixel%counts), " p sourcenumber=", sourceNumber 
!    write(*, *) "height=", height
!end if





    i2(1) = i2(1) + il - 1
    i2(2) = i2(2) + jl - 1

   


    dx = real(i0 - i2(1))
    dy = real(j0 - i2(2))




 
    radius = sqrt(dx * dx + dy * dy)

    if(radius < real(16 / binx)) then
        neighbourDistance = .false.

!write(*, *) "Rejecting at line 7358 distance=", radius, " height=", height
!write(*, *) "i2(1)=", i2(1), " i2(2)=", i2(2), " plevelimage=", PlevelImagePtr(i2(1), i2(2))
!write(*, *) "image(i2(1), i2(2))=", imagePtr(i2(1), i2(2))
!write(*, *) "max pixel counts=", maxval(pixel%counts) 

        return
    end if 

    height = imagePtr(i2(1), i2(2)) - backgroundImagePtr(i2(1), i2(2))

    gaussian = height * exp( -radius * radius / real(8  / binx) )
    gaussian = gaussian + max(2.0, 2.0 * sqrt(gaussian))



    i1 = minloc(pixel%counts) 
    i0 = pixel(i1(1))%x  ! X coordinate of maximum
    j0 = pixel(i1(1))%y  ! Y coordinare of maximum

    if( (imagePtr(i0, j0) - backgroundImagePtr(i0, j0))  > gaussian) then

        neighbourDistance = .true.
    else

!write(*, *) "Rejecting at line 7381 distance=", radius
!write(*, *) "i0=", i0, " j0=", j0, " image=", imagePtr(i0, j0)


        neighbourDistance = .false.
    end if    
    

end function neighbourDistance






!************************************************************************************
!
!
!***********************************************************************************
logical Function CheckVicinity2(pixel, ratio, dx)
      
    type(PixelStruct) :: pixel(:)
    real(single), intent(in) :: ratio
    integer(int8), intent(in) :: dx
    integer(int8) :: dx1
    integer(int16) :: i, j, il, iu, jl, ju, number
    real(double) :: sCounts, ecounts, counts
   
    dx1 = 2 / binx

    il = max(0_int16, minval(pixel%x) - dx1)
    iu = min(nx1, maxval(pixel%x) + dx1)

    jl = max(0_int16, minval(pixel%y) - dx1)
    ju = min(ny1, maxval(pixel%y) + dx1)
   


    do j = jl, ju
        do i = il, iu
            

            !********************************************************
            ! Check for close extended sources
            !********************************************************
            if (elevelImagePtr(i, j) .lt. 0 ) then
   
!if (checkPixel) then
!write(*, *) "checkvicinity2 - rejecting source at line 8015"
!end if

                CheckVicinity2 = .false. 

                return

            end if    

        end do
    end do



   
    !*************************************************************
    ! Find the actual source counts (ie exclude pixels assigned
    ! to a point-source
    !*************************************************************

      
    sCounts = ratio * sum( dble(pixel%counts) )! The counts in the pixels assigned to the source


    ecounts = sCounts 

!if (checkPixel) then
!write(*, *) "Checkvicinity2 line 7540 sourcecounts=", scounts, " ratio=", ratio
!write(*, *) "Pixel counts=", sum(pixel%counts)
!end if
   
  


  
    il = max(0_int16, minval(pixel%x) - dx)
    iu = min(nx1, maxval(pixel%x) + dx)

    jl = max(0_int16, minval(pixel%y) - dx)
    ju = min(ny1, maxval(pixel%y) + dx)
   
   
!if (checkPixel) then
!write(*, *) "Checkvicinity2 line 7556 dx=", dx, " il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if



    do j = jl, ju
        do i = il, iu
            

            !********************************************************
            ! Check for close extended sources
            !********************************************************
            if (elevelImagePtr(i, j) .lt. 0) then
                number = elevelImagePtr(i, j)
                counts = sum( imagePtr, mask=elevelImagePtr == number )




!if (checkPixel) then
!write(*, *) "checkvicinitt2 line 8078 counts=", counts, " ecounts=", ecounts
!end if





                if (counts > eCounts) then
                    CheckVicinity2 = .false. 


!if (checkPixel) then
!write(*, *) "rejecting source at line 8091 counts=", counts, " ecounts=", ecounts
!end if

                    return

                end if    

            end if


            !********************************************************
            ! Check for close point sources
            !********************************************************
            if (plevelImagePtr(i, j) .gt. 0) then

             
                number = plevelImagePtr(i, j)
            

                counts = sum(imagePtr, mask=plevelImagePtr == number)


                if (counts > sCounts) then


!if (checkPixel) then
!write(*, *) "rejecting source at line 8119 point-source counts=", counts, " sourcecountsxratio=", scounts
!write(*, *) "plevelimage number=", number, " no of pixels=", count(plevelImagePtr == number)
!write(*, *) "maxvalue=", maxval(imagePtr, plevelimagePtr==number), " max(pixel)=", maxval(pixel%counts)
!write(*, *) "sum pixel counts=", sum(pixel%counts), " sum p counts=", sum(imagePtr, plevelimagePtr==number)
!write(*, *) "i=", i, " j=", j, " number=", number
!end if


!                if (.not. neighbourDistance(pixel, number, plevelImagePtr)) then
 
!                    if(checkPixel) then
!                        write(*, *) "Rejected source at line 7537"
!                    end if


                    CheckVicinity2 = .false. 
                    return

                end if    

    
            end if


        end do
    end do 

    CheckVicinity2 = .true. 
    
    

end Function CheckVicinity2












!************************************************************************************
!
! Function CheckVicinity3
!
!***********************************************************************************
logical Function CheckVicinity3(pixel, dx, height, nsigma, echeck)
     
    type(PixelStruct) :: pixel(:)
    integer(int8), intent(in) :: dx
    real(single), intent(in) :: height, nsigma
    logical, optional, intent(in) :: echeck
    integer(int32) :: i
    integer(int16) :: j, k, il, iu, jl, ju


!if (checkPixel) then
!write(*, *) "Checkvicinity3 line 8181"
!end if


    do i = 1, size(pixel)

        il = max( 0_int16, pixel(i)%x - dx)
        iu = min( nx1, pixel(i)%x + dx)
        jl = max( 0_int16, pixel(i)%y - dx)
        ju = min( ny1, pixel(i)%y + dx)

        do j = jl, ju
            do k = il, iu
               
                if(elevelImagePtr(k, j) == number) cycle
                if (present(echeck)) then
                    if(echeck) then 
                        if(elevelImagePtr(k, j) < 0 ) then

!if (checkPixel) then
!write(*, *) "k=", k, " j=", j, " elevelimagePtr=", elevelImagePtr(k, j)
!write(*, *) "Checkvicinity3 line 8198 elevelimagept < 0"
!end if

                            CheckVicinity3 = .false.
                            return
                        end if
                    end if



                end if 


                if(imagePtr(k, j) > height .and. plevelImagePtr(k, j) > 0) then

!if (checkPixel) then
!write(*, *) "Checkvicinity3 line 8214 rejecting pixel, height=", height, &
!" imagePtr(", k, ", ", j, ")=", imagePtr(k, j)
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if

                    CheckVicinity3 = .false.
                    return
                end if

                if( imagePtr(k, j) < ( backgroundImagePtr(k, j) - nsigma * &
                      max(1.0, sqrt(backgroundImagePtr(k, j) ) ) ) ) then

!if (checkPixel) then
! write(*, *) "Checkvicinity3 8227 rejecting pixel, image=", imagePtr(k, j),  &
! " backgroundimagePtr(", k, ", ", j, ")=", backgroundimagePtr(k, j)
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju, "k=", k, " j=", j
!end if

                    CheckVicinity3 = .false.
                    return
                end if

            end do
        end do
    end do
       
    CheckVicinity3 = .true.
                
end function CheckVicinity3





!**************************************************************
!
! Checks the pixels around a source for a mod-8 pattern
!
!
!************************************************************
logical Function CheckForMod8Pattern1(pixel, height, dx)
     
    type(PixelStruct) :: pixel(:)
    real(single), optional, intent(in) :: height
    integer(int8), intent(in), optional :: dx
    integer(int8) :: dx1, nPix, nPix1
    integer(int16) :: il, iu, jl, ju, i, j, i1, j1    
    integer :: count1, count2
    real(single) :: fMax

    if(present(dx)) then
        dx1 = dx
    else
        dx1 = 4 / binx  ! was 20 / binx
    end if

    il = max( 0_int16, minval(Pixel%x) - dx1)
    iu = min( nx1, maxval(pixel%x) + dx1)
    jl = max( 0_int16, minval(pixel%y) - dx1)
    ju = min( ny1, maxval(pixel%y) + dx1)
    
    if(present(height)) then
        fMax = height
    else
        fMax = maxval(pixel%counts) * 50.0
    end if

    
    nPix = 10 / binx
    nPix1 = 10 / binx


!if(checkPixel) then
!write(*, *) "Checkformod8pattern line 8282 fmax=", fMax
!end if

    !**************************************************************
    !
    ! Check for a mod-8 pattern if any pixel within the box exceeds
    ! fMax
    !
    !**************************************************************

    if(maxval(imagePtr(il : iu, jl : ju)) < fMax) then
        CheckForMod8Pattern1 = .false.
        return
    end if 



      
    il = max(0, minval(pixel%x) - dx1 - 2 )
    iu = min(nx1, int(maxval(pixel%x) + dx1 - 1, int16))
    jl = max(0_int16, int(minval(pixel%y) - dx1, int16))
    ju = min(ny1, int(maxval(pixel%y) + dx1, int16))



!if(checkPixel) then
!write(*, *) "Checkformod8pattern1 line 8304 fMax=", fMax
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if



    do i = il, iu
        count1 = 0
        count2 = 0
        i1 = i + 1
        do j = jl, ju         
                
            if(imagePtr(i1, j) > imagePtr(i, j) ) then
                count1 = count1 + 1
            else
                count1 = 0
            end if

            if(imagePtr(i1, j) < imagePtr(i, j) ) then
                count2 = count2 + 1
            else
                count2 = 0
            end if



        end do 



!if(checkPixel) then
!write(*, *) "Checkformod8pattern1 line 8339 i=", i, " count1=", count1, " count2=", count2
!end if



        if( count1 >= nPix1 .or. count2 >= nPix1) then ! Possible mod-8 pattern

!if(checkPixel) then
!    write(*, *) "Checkformod8pattern line 8343 settotrue"
!end if



            checkForMod8Pattern1 = .true.
            return
        end if
    end do


    jl = max(0, minval(pixel%y) - dx1 - 2 )
    ju = min(ny1, int(maxval(pixel%y) + dx1 - 1, int16))
    il = max(0_int16, minval(pixel%x) - dx1)
    iu = min(nx1, maxval(pixel%x) + dx1)


!if(checkPixel) then
!write(*, *) "Checkformod8pattern1 line 8354 fMax=", fMax
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if





    do j = jl, ju
        count1 = 0
        count2 = 0
        j1 = j + 1
        do i = il, iu
      
            if(imagePtr(i, j1) > imagePtr(i, j) ) then
                count1 = count1 + 1
            else
                count1 = 0
            end if
            if(imagePtr(i, j1) < imagePtr(i, j) ) then
                count2 = count2 + 1
            else
                count2 = 0
            end if

        end do 

!if(checkPixel) then
!write(*, *) "Checkformod8pattern1 line 8393 j=", j, " count1=", count1, " count2=", count2
!end if




        if( count1 >= nPix1 .or. count2 >= nPix1) then ! Possible mod-8 pattern



!if(checkPixel) then
!    write(*, *) "Checkformod8pattern1 line 8405 settotrue"
!end if

            CheckForMod8Pattern1 = .true.
            return
        end if
    end do




    CheckForMod8Pattern1 = .false.
        

end function CheckForMod8pattern1



!**************************************************************
!
! Checks the pixels around a source for a mod-8 pattern
!
!
!************************************************************
logical Function CheckForMod8Pattern(pixel, height, dx)
     
    type(PixelStruct) :: pixel(:)
    real(single), optional, intent(in) :: height
    integer(int8), intent(in), optional :: dx
    integer(int8) :: dx1, nPix, nPix1
    integer(int16) :: il, iu, jl, ju, i, j    
    integer :: count1, count2
    real(single) :: fMax

    if(present(dx)) then
        dx1 = dx
    else
        dx1 = 4 / binx  ! was 20 / binx
    end if

    il = max( 0_int16, minval(Pixel%x) - dx1)
    iu = min( nx1, maxval(pixel%x) + dx1)
    jl = max( 0_int16, minval(pixel%y) - dx1)
    ju = min( ny1, maxval(pixel%y) + dx1)
    
    if(present(height)) then
        fMax = height
    else
        fMax = maxval(pixel%counts) * 50.0
    end if

    
    nPix = 10 / binx
    nPix1 = 8 / binx

!if(checkPixel) then
!write(*, *) "Checkformod8pattern line 8460 fmax=", fMax, &
!   " il=", il, " iu=", iu, " jl=", jl, " ju=", ju 
!end if

    !**************************************************************
    !
    ! Check for a mod-8 pattern if any pixel within the box exceeds
    ! fMax
    !
    !**************************************************************

    if(maxval(imagePtr(il : iu, jl : ju)) > fMax) then


     
 
     
 
        do i = minval(pixel%x), maxval(pixel%x)
            if(i == 0 .or. i .eq. nx1) cycle

            count1 = 0
            count2 = 0
            do j = minval(pixel%y), maxval(pixel%y)


                if( elevelImagePtr(i, j) .ne. number) cycle
                if( elevelImagePtr(i - 1, j) .eq. 0 .or.   &
                    pLevelImagePtr(i - 1, j) .ne. 0 ) then

                    count1 = count1 + 1
                else
                    count1 = 0
                end if  

            
                if( elevelImagePtr(i + 1, j) .eq. 0 .or. &
                    pLevelImagePtr(i + 1, j) .ne. 0 ) then

                    count2 = count2 + 1
                else
                    count2 = 0
                end if  

                if(count1 .ge. nPix.or. count2 .ge. nPix) then
                     

!if(checkPixel) then
!   write(*, *) "Checkformod8pattern set to true at line 8319 - count1=", count1, " count2=", count2, " npix=", npix
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if
                    


                    CheckForMod8Pattern = .true.
                    return
               
                end if  
            end do

        end do
 
        do i = minval(pixel%y), maxval(pixel%y)

            if(i == 0 .or. i .eq. ny1) cycle

            count1 = 0
            count2 = 0
            do j = minval(pixel%x), maxval(pixel%x)
                if(elevelImagePtr(j, i) .ne. number) cycle
      
            
            
                if ( elevelImagePtr(j, i - 1) .eq. 0 .or.         &
                     pLevelImagePtr(j, i - 1) .ne. 0 ) then

                    count1 = count1 + 1
                else
                    count1 = 0
                end if  

            
                if (elevelImagePtr(j, i + 1) .eq. 0 .or. &
                    pLevelImagePtr(j, i + 1) .ne. 0 ) then


                    count2 = count2 + 1
                else
                    count2 = 0
                end if

!if(checkPixel) then
!write(*, *) "Line 8375 count1=", count1, " count2=", count2, " npix=", npix
!end if   
                if(count1 .ge. nPix .or. count2 .ge. nPix) then
!if (checkPixel) then
!write(*, *) "Checkformod8pattern set to true at line 8368 - count1=", count1, " count2=", count2, " npix=", npix
!write(*, *) "il=", il, " iu=", iu, " jl=", jl, " ju=", ju
!end if
                    CheckForMod8Pattern = .true.
                    return
               
                end if
            end do 
        end do
    end if


   checkForMod8Pattern = .false.






!if(checkPixel) then
!write(*, *) "Checkformod8pattern line 8387 set to false"
!end if

    
end function checkForMod8Pattern







!**************************************************
! Check that no surrounding pixel is greater than the
! source pixel
!
!**************************************************
logical Function PixelBackgroundOkay(pixel)
    type(PixelStruct), intent(in) :: pixel(:)

    integer :: il, iu, jl, ju, i, j, k
    integer(int16) ::  dx = 2
    integer :: n
    real(single) :: maxSignalToNoise
    logical :: check            

   
    n = size(pixel)

    if (n .lt. max(4, 6 / binx)) then
        maxSignalToNoise = maxval((pixel%counts - pixel%backgroundCounts) / sqrt(pixel%counts))
        if (maxSignalToNoise .gt. 4.0) then
            check = .false.
        else
            check = .true.
        end if
    else
        check = .false.
    end if
 
    dx = 2 / binx  ! was 4 / binx


    do i = 1, n
        il = max(lbx, pixel(i)%x - dx)
        iu = min(pixel(i)%x + dx, nx1)
        jl = max(lby, pixel(i)%y - dx)
        ju = min(pixel(i)%y + dx, ny1)
       
        do k = jl, ju
            do j = il, iu
                if (j .eq. pixel(i)%x .and. k .eq. pixel(i)%y) cycle
                if (elevelImagePtr(j, k) .lt. 0) then
                    PixelBackgroundOkay = .false.

                    return
                end if
                if (plevelImagePtr(j, k) .gt. 0) then
                    PixelBackgroundOkay = .false.

                    return
                end if
            
                if (.not. check) cycle
                if (elevelImagePtr(j, k) .eq. 0 .or. elevelImagePtr(j, k) .eq. 1) then ! Was == 1
                 
                   
                    if (pixel(i)%counts .lt. imagePtr(j, k) + max(1.0, 0.25 * sqrt(imagePtr(j, k)))) then ! was 1.0
                   
 
                    
                        PixelBackgroundOkay = .false.

                        return
                    end if
                end if
            end do
        end do
    end do 
    PixelBackgroundOkay = .true.

   

end Function PixelBackgroundOkay
   
!**************************************************
!
!
!**************************************************
logical Function PixelBackgroundOkay1(pixel)
    type(PixelStruct), intent(in) :: pixel

    integer :: il, iu, jl, ju, i, j
    integer :: n
    real(double) :: counts
    integer(int16), parameter :: dx = 2

    il = max(lbx, pixel%x - dx)
    iu = min(pixel%x + dx, nx1)
    jl = max(lby, pixel%y - dx)
    ju = min(pixel%y + dx, ny1)
   
    
    counts = 0.0_double
    n = 0
    do j = jl, ju
        do i = il, iu
            if (i .eq. pixel%x .and. j .eq. pixel%y) cycle
            if (elevelImagePtr(i, j) .ne. 0) cycle
            
            n = n + 1
            counts = counts + dble(smoothedimagePtr(i, j))
            
        end do
    end do 
    if (n .eq. 0) then
        PixelBackgroundOkay1 = .true.
        return
    end if

    counts = counts / dble(n)

    if (pixel%counts .ge. (max(5.0_double, counts + 2.0 * sqrt(counts)))) then
        PixelBackgroundOkay1 = .true.
    else
        PixelBackgroundOkay1 = .false.
    end if

   

end Function PixelBackgroundOkay1
   
!**************************************************
!
!
!**************************************************
logical Function SignalToNoiseOkay(pixel, nSigma)
    type(PixelStruct) :: pixel(:)
    real(single), intent(in) :: nSigma

    integer :: il1, iu1, jl1, ju1, il, iu, jl, ju, i, j
    integer :: n, n1
   

    integer(int16) :: dx = 3
    integer, parameter :: maxCounts = 10000
    
    real(single) :: median, sigma, sigNoise, minCount
    real(single) :: sourceCounts, backgroundCounts, totalCounts, pcount(maxCounts)
    
    dx = 12 / binx
    
    il = max(lbx, minval(pixel%x) - dx)
    iu = min(maxval(pixel%x) + dx, nx1)
    jl = max(lby, minval(pixel%y) - dx)
    ju = min(maxval(pixel%y) + dx, ny1)
   
    dx = 6 / binx

    il1 = max(lbx, minval(pixel%x) - dx)
    iu1 = min(maxval(pixel%x) + dx, nx1)
    jl1 = max(lby, minval(pixel%y) - dx)
    ju1 = min(maxval(pixel%y) + dx, ny1)
  

!write(*, *) "Signal to noise check line 4204"    
    pcount = 0.0
    n = 0

 iLoop :   do j = jl, ju
        do i = il, iu
            if (i .ge. il1 .and. i .le. iu1 .and. &
                j .ge. jl1 .and. j .le. ju1) cycle 
            if (elevelImagePtr(i, j) .ne. 0) cycle
                   
            n = n + 1
     
            pcount(n) = smoothedimagePtr(i, j)
            if (n .eq. maxCounts) exit iLoop
        end do
    end do iLoop




    if (n .lt. 3) then

        SignalToNoiseOkay = .false.
        return
    end if

     
!write(*, *) "n=", n, " minval(pcount)=", minval(pcount(1 : n)), " max=", maxval(pcount(1 : n))


    call computeStatistics3(n, pcount(1 : n), median, sigma)
    sigma = sqrt(sigma)



!write(*, *) "median from background mode=", median, " sigma=", sigma

    
    minCount = (maxval(pixel%counts) - median) * 0.5 ! was 0.5

    n = size(pixel)

    n1 = count(pixel%counts .lt. max(5.0, (median + nSigma * sqrt(median))))



    if (n .lt. 4 .and. n1 .gt. 0) then
        SignalToNoiseOkay = .false.
        return
    else  if (n1 .gt. n / 3 ) then

        SignalToNoiseOkay = .false.
        return
    end if
        

    if (sigma .gt. max(1.0, 2.0 * sqrt(median))) then
        SignalToNoiseOkay = .false.
        return
    end if

    

    totalCounts = sum(pixel%counts)
    backgroundCounts = real(n) * median

    sourceCounts = totalCounts - backgroundCounts

    if (sourceCounts .le. 0.0) then
        SignalToNoiseOkay = .false.
        return
    end if

    sigNoise = sourceCounts / sqrt(totalCounts)

    if (sigNoise .ge. 2.0) then
        SignalToNoiseOkay = .true.
    else

        SignalToNoiseOkay = .false.


    end if

end Function SignalToNoiseOkay
   
!*****************************************************
! IsValidPointSource2
! For small extended sources
!*****************************************************
logical Function IsValidPointSource2(srcParams, pixel)
     

    type(SourceParameters), intent(inout) :: srcParams
    type(PixelStruct) :: pixel(:)
    
    real(single), parameter :: delta1 = 1.5, delta2 = 5.0  ! was 1.2
    real(single) :: ratio
    integer :: noHoles, noStars

    ! Check here to see if might be just mod-8 pattern


 
    noHoles = NumberOfHoles(pixel)
    noStars = NumberOfStars(pixel)
           
    if (noStars .gt. 0) then              ! was 50
        IsValidPointSource2 = .false.           
        return
    end if

    if (srcParams%semiA .lt. 1.0) then
        IsValidPointSource2 = .false.           
        return
    end if

    if (srcParams%semiB .lt. 1.0) then
        IsValidPointSource2 = .false.           
        return
    end if


    if (srcParams%semiA .gt. 2.0) then
        IsValidPointSource2 = .false.           
        return
    end if


    ratio = srcParams%semiA / srcParams%semiB 
    
    if (ratio .ge. delta1) then
 
        IsValidPointSource2 = .true.           
        return 
    end if
    srcParams%extensionFlag = 0

    
    IsValidPointSource2 = .true.           
  

end Function IsValidPointSource2

!*****************************************************
! NoChecks
! For fast-chain
!*****************************************************
logical Function NoChecks(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
   
 
    integer(int16) :: i(1), i0, j0 

     i = maxloc(pixel%counts)
     i0 = min(nx1, pixel(i(1))%x)
     j0 = min(ny1, pixel(i(1))%y)

     source%xPos = real(i0 + 1) 
     source%yPos = real(j0 + 1)
        
     source%extensionFlag = 0

     call ComputePointSourceParameters(int(source%number, int16), nx1, ny1, &
                         smoothedimagePtr, backgroundImagePtr, &
                         i0, j0, source)
    
    
    NoChecks = .true.

end Function NoChecks

!*****************************************************
! IsValidGrismSource
! Checks to see if the parameters of a point
! source lie within certain limits
!*****************************************************
logical Function IsValidGrismSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
 
     
    ! Compute the source parameters

    source%rawCountRate = -maxval(pixel%counts)

    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)
   
     
  
    ! Check here to see if might be just mod-8 pattern
    if (source%semiA .lt. 3.0) then
        
        IsValidGrismSource = .false.
    else
        IsValidGrismSource = .true.
    endif

end Function IsValidGrismSource


!*****************************************************
! IsValidZeroOrderSource
! For zeroth-order spectrum
!*****************************************************
logical Function IsValidZeroOrderSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
    integer(int16) :: i, il, iu, jl, ju, i0(1), j0, minY, maxY
    integer(int16) :: dx = 10, dx1      
   

    real(single), parameter :: delta = 10.0



    source%rawCountRate = -maxval(pixel%counts)


    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)          ! The computed moments
   
 
    if (source%semia < 2.0) then
        IsValidZeroOrderSource = .false.
        return
    end if    
  

    if (source%semiA / source%semiB > delta) then
        IsValidZeroOrderSource = .false.
        return
    end if

  

    if (abs(abs(source%theta) - 90.0) > 30.0) then
        IsValidZeroOrderSource = .false.
        return
    end if


    if (size(pixel) .gt. 400 / binx) then
        IsValidZeroOrderSource = .true.
        return
    end if



   
    ! Check no source immediately below

    dx = 40 / binx
    dx1 = 4 / binx

    j0 = minval(pixel%y)
    i0 = maxloc(pixel%counts, mask=pixel%y .eq. j0)
    i0(1) = pixel(i0(1))%x    
    il = max(0_int16, i0(1) - dx1)
    iu = min(nx1, i0(1) + dx1)
    ju = j0
    jl = max(0_int16, j0 - dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
       IsValidZeroOrderSource = .false.
       return
    end if

    ! Check no source immediately above


    j0 = maxval(pixel%y)
    i0 = maxloc(pixel%counts, pixel%y .eq. j0)
    i0(1) = pixel(i0(1))%x    

    il = max(0_int16, i0(1) - dx1)
    iu = min(nx1, i0(1) + dx1)
    jl = j0
    ju = min(ny1, j0 + dx)


    if (any(elevelImagePtr(il : iu, jl : ju) .lt. 0)) then
       IsValidZeroOrderSource = .false.
       return
   
    end if
!  IsValidZeroOrderSource = .true.
!return




    minY = minval(pixel%y)
   
    maxY = maxval(pixel%y)
    jl = minY + 2
    ju = maxY - 2

    if (jl .gt. ju) then
       IsValidZeroOrderSource = .true.
       return
    end if

       
    do i = jl, ju

        il = minval(pixel%x, pixel%y .eq. i) - 1
        if (il .gt. 0) then
            if (imagePtr(il, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
        end if
        il = il - 1


        if (il .gt. 0) then
            if (imagePtr(il, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
        end if



        iu = maxval(pixel%x, pixel%y .eq. i) + 1
        if (iu .le. nx1) then
            if (imagePtr(iu, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
    
        end if




        iu = iu + 1



        if (iu .le. nx1) then
            if (imagePtr(iu, i) .lt. 0.0) then
                IsValidZeroOrderSource = .false.
                return
            end if
   

        end if

    end do

    IsValidZeroOrderSource = .true.


end Function IsValidZeroOrderSource

!*****************************************************
! IsValidFirstOrderSource -
! For first-order spectra
!*****************************************************
logical Function IsValidFirstOrderSource(source, pixel)
     

    type(SourceParameters), intent(inout) :: source
    type(PixelStruct) :: pixel(:)
    integer(int16) :: minxc, maxxc
    integer(int16) :: i0, j0
    integer(int16) :: dx = 5

    source%rawCountRate = -maxval(pixel%counts)


    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)
 
    if (abs(abs(source%theta) - 90.0) > 2.0) then
        IsValidFirstOrderSource = .false.

        return
    end if

 
    j0 = nint(source%ypos - 0.5)
    if (j0 .lt. 0) then
        j0 = 0
    else 
        if (j0 .gt. ny1) j0 = ny1
    end if
    
    i0 = nint(source%xpos - 0.5)
    if (i0 .lt. 0) then
        i0 = 0
    else 
        if (i0 .gt. nx1) i0 = nx1
    end if
  
    dx = 6 / binx
    minxc = minval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    minxc = min(minxc, i0 - dx)
    maxxc = maxval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    maxxc = max(maxxc, i0 + dx)


    where(pixel%x .lt. minxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where

    where(pixel%x .gt. maxxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where

    call ComputeExtendedSourceParameters(sourceNumber, &  ! The source number
                                         pixel,        &  ! The pixels for this source
                                         source)


    i0 = nint(source%xpos - 0.5)
    if (i0 .lt. 0) then
        i0 = 0
    else 
        if (i0 .gt. nx1) i0 = nx1
    end if
  

    j0 = nint(source%ypos - 0.5)
    if (j0 .lt. 0) then
        j0 = 0
    else 
        if (j0 .gt. ny1) j0 = ny1
    end if


    minxc = minval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    minxc = min(minxc, i0 - dx)
    maxxc = maxval(pixel%x, pixel%y .eq. j0 .and. pixel%number .eq. sourceNumber)
    maxxc = max(maxxc, i0 + dx)

    where(pixel%x .lt. minxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where


    where(pixel%x .gt. maxxc) 
        pixel%counts = 0.0
        pixel%number = 0
    end where


    if (abs(abs(source%theta) - 90.0) > 2.0) then
        IsValidFirstOrderSource = .false.

        return
    end if


    if (source%semia < 10.0 .or. source%semiA / source%semiB < 10) then
        IsValidFirstOrderSource = .false.

    else
       
       IsValidFirstOrderSource = .true.
    end if    



  
end Function IsValidFirstOrderSource

!************************************************************
!
!
!************************************************************
subroutine GetPSfParameters

    real(single) :: xCentre, yCentre, xPos, yPos
 
    integer :: i, j, k
    real(single), pointer :: backgroundImage(:, :) => null()
    real(single), pointer :: tempImage(:, :) => null()


    type(SourceParameters) :: source
   

    fCircle = 0.0
    xCentre = size(psfParams%psfImage, 1) / 2.0
    yCentre = size(psfParams%psfImage, 2) / 2.0

    do i = 1, detectParams%nCircles
        k = 0
        do j = 1, detectParams%nSize
            xPos = xCentre + rCosTheta(j, i)
            if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
            yPos = yCentre + rSinTheta(j, i)
                
            if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle  
            k = k + 1
            fCircle(k) = Interpolate(psfParams%psfImage, xPos, yPos)
        end do    
        call computeStatistics3(k, fCircle(1 : detectParams%nSize), mean(i), sigma(i))

        if (i .gt. 1) then
            psfParams%psfRatio(i - 1) = mean(i - 1) / mean(i)
            
        end if
    end do

 !   fwhm = ComputeFWHM(psfParams%psfimage, xcentre, ycentre)

 

    psfParams%psfRatio(1) = mean(1) / mean(2)
  
    psfParams%psfRatio(2) = mean(1) / mean(3)
 
    psfParams%psfRatio = psfParams%psfRatio * 0.5



    !******************************************************************************
    !
    !******************************************************************************


    backgroundimage => allocate2dsingle(0, ubound(psfparams%psfimage, 1), 0, ubound(psfparams%psfimage, 2), "backgroundimage")
    backgroundimage = 10.0


    tempimage => allocate2dsingle(0, ubound(psfparams%psfimage, 1), 0, ubound(psfparams%psfimage, 2), "backgroundimage")
    tempimage = backgroundimage + psfparams%psfimage * 100.0
    



    source%xpos = size(psfparams%psfImage, 1) / 2.0 + 0.5
  
    source%ypos = size(psfparams%psfImage, 2) / 2.0 + 0.5
  


    deallocate(tempimage, backgroundImage)
    nullify(tempimage, backgroundImage)


end subroutine GetPSFParameters


!*****************************************************************************
! Compute the FWHM of the source at the given position 
!
!*****************************************************************************
real(single) function ComputeFWHM(image, xpos, ypos)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: xPos, yPos
    real(single) :: radius
    real(single) :: f1, f2
    real(double) :: rMin, rMax
    integer :: i    

    FWHMImagePtr => image

    fcentre = Interpolate(FWHMimagePtr, xPos, yPos)
    psfXcentre = xpos
    psfycentre = ypos

    fcentre = fcentre / 2.0
  
    f1 = GridInterpolate(dble(xPos + 0.1), dble(yPos), psfGridFit)
    f2 = GridInterpolate(dble(xPos + 2.0), dble(yPos), psfGridFit)

    rmin = 0.25
    f1 = Interpolate(FWHMimagePtr, real(xpos + rmin), real(yPos))


    do i = 1, 50
        rmax = rmin + real(i) * 0.25
        f2 = Interpolate(FWHMimagePtr, real(xPos + rmax), real(yPos))
        if ((f1 - fcentre) / (f2 - fcentre) .lt. 0.0) exit
    end do 

    rmax = rmax + 0.25

    rmin = 0.5 / real(binx)
    rmax = real(4 / binx)
    radius = ZeroIn(FWHMFunction, rmin, rmax, 0.0_double, 0.0_double)
   



!    write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre + radius), real(psfyCentre))
!  write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre - radius), real(psfyCentre))
!  write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimagePtr, real(psfxCentre), real(psfyCentre + radius))
!write(*, *) "Computed radius=", radius, " counts=", Interpolate(FWHMimageyPtr, real(psfxCentre), real(psfyCentre - radius))

 
    ComputeFWHM = 2.0 * radius

    write(errstr,'(a, f0.2, a, f0.8)') "Computed FWHM=",  (2.0 * radius), " Function value=", FWHMFunction(dble(radius))
    call message(AppMsg, Verbose)
 

    nullify(FWHMImagePtr)

end function ComputeFWHM

!*****************************************************************************
! Function FWHMFunction
!
!*****************************************************************************
real(double) Function FWHMFunction(radius)
    real(double), intent(in) :: radius
    real(double) :: xPos, yPos, theta, dtheta
    real(single) :: xlOwer, xUpper, yLower, yUpper
    integer :: i, k
    real(double) :: mean

    dtheta = TWO_PI / 36.0

    k = 0

    xLower = 1.0
    yLower = 1.0
    xUpper = size(FWHMimagePtr, 1) - 1.0
    yUpper = size(FWHMimagePtr, 2) - 1.0

    mean = 0.0_double
    do i = 1, 36
        theta = dtheta * dble(i)
        xPos = psfxCentre + radius * cos(theta)
        if (xPos .lt. xLower .or. xPos .gt. xUpper) cycle 
        yPos = psfyCentre + radius * sin(theta)
                
        if (yPos .lt. yLower .or. yPos .gt. yUpper) cycle  
        k = k + 1
        mean = mean + dble(Interpolate(FWHMimagePtr, real(xPos), real(yPos)))

        
    end do

    if (k .gt. 10) then

        mean = mean / dble(k)

        FWHMFunction = fCentre - mean
    else
        FWHMFunction = -1.0_double

    end if

!write(*, *) "k=", k, " radius=", radius, " mean=", mean, " FWHMFunction=", FWHMFunction  

end Function FWHMFunction

!***********************************************
!
!**********************************************
subroutine MinimumList(n, data, nsize)
    integer, intent(inout) :: n
    real(single) :: data(1: n)

    integer, intent(in) :: nsize
    real(single) :: tempdata(1: n)

    real(double) :: minimum
    integer :: i, j, k, kmin
    real(double) :: fmin
   
    fmin = huge(1.0)

    kmin = 1
    do i = 1, n
        k = i
        minimum = 0.0_double
        do j = 1, nsize
            minimum = minimum + dble(data(k))
            k = k + 1
            if ( k .gt. n) k = 1
        end do
   
        if (minimum .lt. fmin) then
            fmin = minimum
            kmin = i
        end if
 
    end do


    do j = 1, nsize

        tempdata(j) = data(kmin)
        kmin = kmin + 1
        if ( kmin .gt. n) kmin = 1

 
    end do
    data = tempdata    
    n = nsize

end subroutine MinimumList    

!****************************************************************************
! For removing extended light round bright sources (mod-8 patterns, etc
!
!***************************************************************************  
subroutine BrightSourceMask(npSources, iStart, pSourceList, image, backgroundimage, check)
    integer(int16), intent(inout) :: npSources
    integer(int16), intent(in) :: iStart
    type(SourceLinkedList) :: pSourceList  ! The linked-list point source-list
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    logical, intent(in) :: check
 

    integer(int16) :: il1, iu1, jl1 , ju1, di1, di2
    integer(int16) :: il2, iu2, jl2, ju2      
    type(sourceParameters), pointer :: pSourceData(:) => null()
    integer(int16) :: i, j, k 
    real(single) :: height
     

    if (npSources .eq. 0) return

!write(*, *) "*************** Brightsourcemask line 9076 *********************************************"

    height = 1.5

    ! Load the point-source data
    psourceData => SourceDataArray(psourceList, npSources)

 
    di1 = 2 / binx
    di2 = 4 / binx   ! was 4
   
    do i = iStart, npSources

      
        if (.not. btest(pSourceData(i)%qualityFlag, 4)) cycle

!write(*, *) "source ", i, " x=", psourcedata(i)%xpos, " y=", psourcedata(i)%ypos
!write(*, *) "imin=", psourcedata(i)%imin, " imax=", psourcedata(i)%imax
!write(*, *) "jmin=", psourcedata(i)%jmin, " jmax=", psourcedata(i)%jmax


        !*****************************************************************
        ! Check each bright source
        !**************************************************************** 
        il1 = max(lbx, pSourceData(i)%imin - di1)
        iu1 = min(nx1, pSourceData(i)%imax + di1)
        jl1 = max(lby, pSourceData(i)%jmin - di1)
        ju1 = min(ny1, pSourceData(i)%jmax + di1)
     

        il2 = max(lbx, pSourceData(i)%imin - di2)
        iu2 = min(nx1, pSourceData(i)%imax + di2)
        jl2 = max(lby, pSourceData(i)%jmin - di2)
        ju2 = min(ny1, pSourceData(i)%jmax + di2)
     

        do k = jl2, ju2
         
            do j = il2, iu2
               
       

                !**********************************************************************
                !  FRIDAY HAVE BLANKED THIS OUT
                !********************************************************************** 
                if (check) then
                    if (image(j, k) .ge. 100.0) cycle
                end if

                emaskimage(j, k) = .true.
       
                if (check) then            
                    if( image(j, k) .lt. HeightFunction5(j, k, height, backgroundImage)) cycle
                end if
 
       
                
                if ( k .lt. jl1 .or. k .gt. ju1 .or. &
                     j .lt. il1 .or. j .gt. iu1) cycle
   

      
                maskimage(j, k) = .true.
       

       
            end do
        end do





      
    end do
  
    deallocate(pSourceData)
    nullify(pSourceData)


!if (emaskimage(ipixel, jpixel)) then
!write(*, *) "maskimage(", ipixel, ", ", jpixel, ")=", maskimage(ipixel, jpixel)
!call fatal("bug at line 8861 in brightsourcemask")
!end if






end subroutine BrightSourceMask


!****************************************************************************
! For removing extended light round bright sources (mod-8 patterns, etc
!
!***************************************************************************  
subroutine BrightSourceMask1(npSources, iStart, pSourceList, image, backgroundimage)
    integer(int16), intent(inout) :: npSources, iStart
    type(SourceLinkedList) :: pSourceList  ! The linked-list point source-list
    real(single), pointer :: image(:, :), backgroundImage(:, :)
    integer(int16) :: il, iu, jl , ju, di
      
    type(sourceParameters), pointer :: pSourceData(:) => null()
    integer(int16) :: i, j, k 

     
    real(single) :: height, radius, maxRadius1, maxradius2, dx, dy  


    if (npSources .eq. 0) return


    height = 1.5

    ! Load the point-source data
    psourceData => SourceDataArray(psourceList, npSources)


    di = 24 / binx ! was 24   
    maxRadius1 = real(di)
    
    di = 40 / binx   

    maxRadius2 = real(di)


    do i = iStart, npSources

      
        if (.not. btest(pSourceData(i)%qualityFlag, 4)) cycle



        !*****************************************************************
        ! Check each bright source
        !**************************************************************** 
        il = max(lbx, pSourceData(i)%imin - di)
        iu = min(nx1, pSourceData(i)%imax + di)
        jl = max(lby, pSourceData(i)%jmin - di)
        ju = min(ny1, pSourceData(i)%jmax + di)
     
        do k = jl, ju
            dy = pSourceData(i)%ypos - real(k + 1)
       

     

            do j = il, iu

                if( image(j, k) .lt. HeightFunction5(j, k, height, backgroundImage)) cycle
          
               
                dx = pSourceData(i)%xpos - real(j + 1)
      
                radius = sqrt(dx * dx + dy * dy)
                if (radius .gt. maxRadius2) cycle
           
       
                emaskimage(j, k) = .true.
           
                if (radius .le. maxradius1) maskImage(j, k) = .true. ! Don't use these pixels in source detection
       
            end do
        end do
      
    end do
  
    deallocate(pSourceData)
    nullify(pSourceData)


end subroutine BrightSourceMask1

!******************************************************************
!
! Looks for a mod-8 pattern around a source
!
!******************************************************************
logical Function IsMod8Source(source, image, imin, imax, jmin, jmax)
    type(sourceParameters), intent(inout) :: source
    real(single), pointer, dimension(:, :) :: image   ! , backgroundImage 
    integer(int16), intent(out) :: imin, imax, jmin, jmax


    ! Local variables
    integer(int16) :: i1, i2, j1, j2, nx, ny, i0, j0, il, iu, jl, ju

    integer(int16) :: il1, iu1, jl1, ju1, n, nmin
    integer(int16) :: maxWidth

    integer(int16) :: iPlus1, jPlus1, iMinus1, jMinus1
    integer(int16) :: iPlus2, jPlus2, iMinus2, jMinus2
    integer(int16) :: iPlus3, jPlus3, iMinus3, jMinus3



    integer(int16) :: dx, n1, n3
    integer(int16) :: k1, k2, k3, k4, k5, kmin 
    real(single) :: fmax, fMin
!    real(single) ::  fTotal, fBackground, radius, rate, area
    real(single), parameter :: nsigma = 2.0 ! was 2
!integer, parameter :: number = 1



!if (source%number == number .or. number < 0) then
!write(*, *) "Line 7082 x=", source%xpos, " y=", source%ypos, &  
!            " imin=", source%imin, " imax=", source%imax, &
!            " jmin=", source%jmin, " jmax=", source%jmax
!end if

!IsMod8Source = .false.
!return

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)
    


    dx = 4 / binx  ! Was 16
    il = max( lbx, i0 - dx )
    iu = min( nx, i0 + dx )
    jl = max(lby, j0 - dx)
    ju = min(ny, j0 + dx)


    fmax = maxval( image(il : iu, jl : ju) )
    fmin = 0.1 * fmax 
   

    source%rawCountRate = -fmax


    !***************************************************************************
    ! April 23rd
    !
    !***************************************************************************

!    radius = real(12 / binx)
!    call getCounts( image, source%xpos, source%ypos, radius, fTotal, area)        
!    call getCounts( backgroundImage, source%xpos, source%ypos, radius, fBackground)        


!    if(ftotal .gt. 0.0) then
!        rate =  (fTotal - fBackground) * ( PI * radius * radius / area ) / detectParams%exposure

!        if(rate .ge. 60.0) then
!            isMod8Source = .true.
!            source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag
!        elseif (rate .le. 40.0) then

!            isMod8Source = .false.
!            return
!        else
!            isMod8Source = .false.
!        end if
!    else
!        isMod8Source = .false.
!        return
!    end if  
        
 

    dx = 30 / binx


    maxwidth = 8 / binx


    if (fmax .lt. 1000.0 ) then
        isMod8Source = .false.      
        return
    elseif(fmax .ge. 10000.0) then ! WAS 10000
        isMod8Source = .true. 
        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag


!write(*, *) source%number, " ) mod8source Line 9777  x=", source%xpos, &
!" y=", source%ypos, " ftotal=", ftotal, &
!" fbackground=", fbackground, " rate=", rate, " area=", area, " exposure=", detectParams%exposure, &
!" fmax=", fmax

    else     
        isMod8Source = .false.
    end if

    
    imin = max(0_int16, i0 - maxWidth)
    imax = min(nx, i0 + maxWidth)
       
    jmin = max(0_int16, j0 - maxWidth)
    jmax = min(ny, j0 + maxWidth)




!if (source%number == number .or. number < 0) then
!write(*, *) "Line 7153 imin=", imin, " imax=", imax, &
!            " jmin=", jmin, " jmax=", jmax, " fmin=", fmin
!end if




    il1 = max( lbx, i0 - dx )
    iu1 = min( nx,  i0 + dx )
    jl1 = max( lby, j0 - dx )
    ju1 = min( ny,  j0 + dx )

  

    dx = 10 / binx ! was 6 / binx  


    il = max( lbx, i0 - dx )
    iu = min( nx,  i0 + dx )
    jl = max( lby, j0 - dx )
    ju = min( ny,  j0 + dx )


    !****************************************************************
    ! Check both sides of the source
    !****************************************************************
   

    k1 = 0
    k2 = 0
    k3 = 0
    k4 = 0
  
    nmin = 2 * dx 

    !****************************************************************
    ! Checking left-hand side of the source for a mod-8 pattern
    !***************************************************************
    n = ju - jl + 1
   
    if (n .ge. 10) then
        nmin = n - 3
        


!if (source%number == number .or. number < 0) then
!write(*, *) "Checking left hand side of source, n=", n, " nmin=", nmin
!end if


 
 iLoop1  : do i1 = i0 - 10 / binx, il1, -1
               
    
            iMinus1 = i1 - 1
            iMinus2 = i1 - 2
            iminus3 = i1 - 3
            if (iminus3 .lt. 0) exit               
 
                 
            if ( any(image( i1 : i1, jl : ju) > fmax) ) exit

            !***********************************************************************************
            ! Check that the column of pixels centred on iminus1 is less than the column
            ! centred on i1
            !**********************************************************************************  
   
            n1 = count( image( i1 : i1, jl : ju) .gt. ( image( iminus1 : iminus1, jl : ju) + &
                 max(2.0, nsigma * sqrt( image( iminus1 : iminus1, jl : ju) ))) )
       

            if(n1 < nmin) cycle



            !*********************************************************************
            ! Check that the column of pixels to the left of iminus1 is greater than
            ! the column centred on iminus1 
            !********************************************************************
            
       

            n3 = max( count( image(iMinus2 : iMinus2, jl : ju) .gt. image( iminus1 : iminus1, jl : ju) ), &
                      count( image(iMinus3 : iMinus3, jl : ju) .gt. image( iminus1 : iminus1, jl : ju) ))
       


!if (source%number == number .or. number < 0) then
!write(*, *) "x=", source%xpos, " y=", source%ypos, " i0=", i0, " j0=", j0, " i1=", i1, " fmin=", fmin
!write(*, *) "Line 8066 n1=", n1, " il=", il, " jl=", jl, " ju=", ju, " nmin=", nmin
!end if

           
            if ( n3 .ge. nmin) then



                !***********************************************************************
                ! Maybe a mod-8 pattern to the left of the source
                !***********************************************************************           
   
             
                imin = min(imin, iminus3)



!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8064 n1=", n1, " il=", il, " jl=", jl, " ju=", ju, " nmin=", nmin, &
!     "  maxwidth=", maxwidth
!end if


                k1 = 1

            end if
           
        end do iLoop1


     
        !****************************************************************
        ! Checking right-hand side of the source for a mod-8 pattern
        !***************************************************************

!if (source%number == number .or. number < 0) then
!write(*, *) "cHECKING RIGHT-HAND SIDE OF SOURCE i0=", i0, " iu1=", iu1, " maxwidth=", maxwidth
!end if
    
 
iLoop2 : do i2 = i0 + 10 / binx, iu1   
        
            iPlus1 = i2 + 1
            iPlus2 = i2 + 2
            iplus3 = i2 + 3

            if (iplus3 .gt. nx) exit       


            if ( any(image( i2 : i2, jl : ju) > fmax) ) exit
     
            !*****************************************************************************
            ! Check that the column of pixels centred on column i2 is greater than that
            ! centered on column iplus1
            !*****************************************************************************
                    
            n1 = count( image(i2 : i2, jl : ju) .gt. ( image( iplus1 : iplus1, jl : ju) + &
                        max(2.0, nsigma * sqrt( image( iplus1 : iplus1, jl : ju) ))) )
           



            if(n1 .lt. nmin) cycle



            !*********************************************************************
            ! Check that the column of pixels to the right of iplus11 is greater than
            ! the column centred on iplus11 
            !********************************************************************
      

            n3 = max( count( image(iplus2 : iplus2, jl : ju) .gt. image( iplus1 : iplus1, jl : ju)), &
                      count( image(iplus3 : iplus3, jl : ju) .gt. image( iplus1 : iplus1, jl : ju)) )
           






!if (source%number == number .or. number < 0) then
!write(*, *) "i2=", i2, " n1=", n1, " nmin=", nmin, " jl=", jl, " ju=", ju, " iu1=", iu1
!write(*, *) "pixels"
!do i1 = jl, ju
!write(*, *) "image(", iplus1, ",", i1, ")=", image(iplus1, i1), " image(", i2, ",", i1, ")=", image(i2, i1)
!end do
!end if
 
                   
     
            if ( n3 .ge. nmin ) then



                !***********************************************************************
                ! Maybe a mod-8 pattern to the right of the source
                !***********************************************************************           
    
              
 
               imax = max(imax, iplus3)

!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8104 n1=", n1, " il=", il, " jl=", jl, " ju=", ju, " nmin=", nmin, &
!" maxwidth=", maxwidth
!end if


                k2 = 1 

           
           end if

       
        end do iLoop2



    end if

    !****************************************************************
    ! Check below and above the source
    !****************************************************************
    

    n = iu - il + 1

    
    if (n .ge. 10) then
        nmin = n - 3
      

!if (source%number == number .or. number < 0) then
!write(*, *) "Checking above the source, j0=", j0, " ju1=", ju1
!end if

 
jLoop1 : do j1 = j0 + 10 / binx, ju1

            !****************************************************************
            ! Checking above the source for a mod-8 pattern
            !***************************************************************
    
            k5 = k5 + 1 
            jPlus1 = j1 + 1
            jPlus2 = j1 + 2
            jPlus3 = j1 + 3
       
            if (jplus3 .gt. ny) exit      
               
 
                     
            if ( any(image( il : iu, j1 : j1) > fmax) ) exit
     


            !*****************************************************************************
            ! Check that the row of pixels centred on row j1 is greater than that
            ! centered on column jplus1
            !*****************************************************************************
            n1 = count( image(il : iu, j1 : j1) .gt. ( image( il : iu, jplus1 : jplus1) + &
                 max(2.0, nsigma * sqrt( image( il : iu, jplus1 : jplus1) ))) )
           

            if(n1 < nmin) cycle
          

            n3 = max( count( image(il : iu, jPlus2 : jPlus2) .gt. image( il : iu, jplus1 : jplus1)), &
                      count( image(il : iu, jPlus3 : jPlus3) .gt. image( il : iu, jplus1 : jplus1)) )
           


!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8155 n1=", n1, " il=", il, " iu=", iu, " j1=", j1, "  nmin=", nmin
!end if

          
            if (n3 .ge. nmin) then



                !***********************************************************************
                ! Maybe a mod-8 pattern above the source
                !***********************************************************************           
                
                jmax = max(jmax, jplus3)


!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8446 n1=", n1, " il=", il, " iu=", iu, " j1=", j1, "  nmin=", nmin, &
!" maxwidth=", maxwidth
!end if

                k3 = 1
            end if   
        end do jLoop1
   
   

        !****************************************************************
        ! Checking below the source for a mod-8 pattern
        !***************************************************************



!if (source%number == number .or. number < 0) then
!write(*, *) "Checking below the source j0=", j0, " jl1=", jl1
!end if

  
jLoop2 : do j2 = j0 - 10 / binx, jl1, -1
         
            jMinus1 = j2 - 1
            jMinus2 = j2 - 2
            jMinus3 = j2 - 3


            if (jMinus3 .lt. 0) exit     
           
   
     
         
            if ( any(image( il : iu, j2 : j2) > fmax) ) exit


            !*****************************************************************************
            ! Check that the row of pixels centred on row j2 is greater than that
            ! centered on column jminus1
            !*****************************************************************************
  

     

            n1 = count( image(il : iu, j2 : j2) .gt. ( image( il : iu, jminus1 : jminus1) + &
                 max(2.0, nsigma * sqrt( image( il : iu, jminus1 : jminus1) ))) )


          
            if(n1 < nmin) cycle


            !*****************************************************************************
            ! Check that the row of pixels centred on row j1 is greater than that
            ! centered on column jplus1
            !*****************************************************************************
      

            n3 = max( count( image(il : iu, jMinus2 : jMinus2) .gt. image( il : iu, jminus1 : jminus1) ), &
                      count( image(il : iu, jMinus3 : jMinus3) .gt. image( il : iu, jminus1 : jminus1) ) )


       

!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8189 n1=", n1, " il=", il, " iu=", iu, " j2=", j2, " nmin=", nmin
!end if

            if ( n3 .ge. nmin) then


                !***********************************************************************
                ! Maybe a mod-8 pattern below the source
                !***********************************************************************           
         

                jmin = min(jmin, jminus3)

                             
              
             
!if (source%number == number .or. number < 0) then
!write(*, *) "Line 8189 n1=", n1, " il=", il, " iu=", iu, " j2=", j2, " nmin=", nmin, &
!" maxwidth=", maxwidth
!end if


            end if
            
          
        end do jLoop2
    end if



    kmin = 3
    if (source%xpos <= real(12 / binx) .or. source%ypos <= real(12 / binx)) kmin = kmin  - 1
    if (source%xpos >= real(nx1 - 12 / binx) .or. source%ypos >= real(ny1 - 12 / binx)) kmin = kmin - 1
              


    if ( (k1 + k2 + k3 + k4) .ge. kmin ) then ! Accept it is as having a mod-8 pattern 


!write(*, *) source%number, " ) mod8source Line 10185  x=", source%xpos, &
!" y=", source%ypos, " ftotal=", ftotal, &
!" fbackground=", fbackground, " rate=", rate, " area=", area, " exposure=", detectParams%exposure, &
!" fmax=", fmax

  
        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag
        IsMod8Source = .true.
       
          
!if (source%number == number .or. number < 0) then
!write(*, *) "source ", sourceNumber, " satisfies criterion for a mod-8 source"
!write(*, *) "ismod8sources=", ismod8source, " x=", source%xpos, " y=", source%ypos
!end if
    
    end if
  


!if (source%number == number .or. number < 0) then
!write(*, *) "K1=", K1, " K2=", K2, " K3=", K3, " K4=", K4
!write(*, *) "Line 7284 ismod8source=", ismod8source, " maxwidth=", maxwidth
!write(*, *) "sourceno=", source%number, " x=", source%xpos, " y=", source%ypos, &
!   " qualityflag=", source%qualityFlag
!write(*, *) "imin=", imin, " imax=", imax, &
!            " jmin=", jmin, " jmax=", jmax
!write(*, *) "ismod8sources=", ismod8source
!ismod8source = .false.
!end if

end Function IsMod8Source
!************************************************************************
!
!
!************************************************************************
recursive subroutine goodPixel(i, j, minHeight, count, checked, sourceCounts, totalCounts)
    integer(int16), intent(in) :: i, j
    integer, intent(inout) :: count
    real(single), intent(in) :: minHeight
    logical, pointer :: checked(:, :)
    real(double), intent(out) :: sourceCounts, totalCounts
    real(double) :: diff
    integer(int16) i0, j0

    if (elevelImagePtr(i, j) .lt. 0 .or. elevelImagePtr(i, j) .eq. 1) return
 
    if (i < lbound(checked, 1) .or. i > ubound(checked, 1)) return
    if (j < lbound(checked, 2) .or. j > ubound(checked, 2)) return

    if (checked(i, j)) then
        return
    else
        checked(i, j) = .true.
    end if

    !*******************************************************
    ! Check to see if pixel is at least the minimum height
    ! above the background
    !*******************************************************
    
    diff = imagePtr(i, j) - backgroundImagePtr(i, j)


    if (diff < minHeight) return
    sourceCounts = sourceCounts + diff
    totalCounts = totalCounts + dble(imagePtr(i, j))

    count = count + 1

    if (count > 100) return ! Safety check


    



!write(*, *) "goodpixel i=", i, " j=", j, " minheight=", minheight, &
!   " diff=", diff, " count=", count


    i0 = i - 1
    if (i0 >= 0) then
        call goodPixel(i0, j, minHeight, count, checked, sourceCounts, totalCounts)
    end if

    i0 = i0 + 1
    if (i0 <= nx1) then
        call goodPixel(i0, j, minHeight, count, checked, sourceCounts, totalCounts)
    end if


    j0 = j - 1
    if (j0 >= 0) then
        call goodPixel(i, j0, minHeight, count, checked, sourceCounts, totalCounts)
    end if

    j0 = j0 + 1
    if (j0 <= ny1) then
        call goodPixel(i, j0, minHeight, count, checked, sourceCounts, totalCounts)
    end if



end subroutine goodPixel


!**************************************************
!
!
!*************************************************
logical function checkforbadPixels(source, i0, j0)
    type(sourceParameters), intent(inout) :: source

    integer(int16), intent(in) :: i0, j0
    integer(int16) :: dx, il, iu, jl, ju
    integer :: i, j, count1, count2
    

    dx = 20 / binx

    il = i0 - dx
    jl = j0 - dx


    il = max(0_int16, il)
    jl = max(0_int16, jl)


    il = min(nx1, il)
    jl = min(ny1, jl)



    iu = i0 + dx
    ju = j0 + dx


    iu = max(0_int16, iu)
    ju = max(0_int16, ju)


    iu = min(nx1, iu)
    ju = min(ny1, ju)
    

    count1 = 0 
    count2 = 0   


    do i = il, iu
        do j = jl, ju
	    if ( imagePtr(i, j) > 65000.0) then
	  
	        count1 = count1 + 1
             

            elseif ( imagePtr(i, j) > 5000.0) then
	  
	        count2 = count2 + 1
             

           end if




        end do
    end do


  

    if ( count1 > 0 .and. count1 < 3 .and. count2 == 0 ) then


        source%qualityFlag = ibset(source%qualityFlag, 9) 
        checkForBadPixels = .true. 

    else

        checkForBadPixels = .false. 
    end if
end function checkforbadPixels

!*****************************************************************
!
!****************************************************************
subroutine sourceCircle( source, imagePtr, backgroundImagePtr )

    type(sourceParameters), intent(inout) :: source
    real(single), pointer, dimension(:, :) :: imagePtr, backgroundImagePtr
    real(single) :: xcentre, ycentre, theta
 
    integer :: i, j, k
    integer :: n, nrad, nangle
    real(single) :: rad, dr, dtheta, xPos, yPos, startRadius  
    real(single), pointer, dimension(:) :: fcircle => null()
    real(single), pointer, dimension(:) :: radius => null(), fradius=> null()
    real(single), pointer, dimension(:) :: dy => null()

    real(single) :: f, maxRadius, std
    integer(int16) :: i0, j0, il, iu, jl, ju, nx, ny
    integer(int8) :: dx1
    real(double) :: xmin, xmax, tol

    nrad = 80 / binx
    

    nangle = 36
    fcircle => allocateSingle(1, nangle, "fcircle")
    radius => allocateSingle(1, nrad, "radius")
    dy => allocateSingle(1, nrad, "dy")

    fradius => allocateSingle(1, nrad, "fradius")



    if (btest(source%qualityFlag, 4)) then
        startRadius = real( 30 / binx )
    else
        startRadius = real( 12 / binx )
    end if
  




    dr = real(2 / binx)

    dtheta = real(360 / nangle) * degreesToRadians

    
    xcentre = source%xpos - 0.5
    ycentre = source%ypos - 0.5




    n = 0
    rad = real(120 / binx)
    k = 0
    do j = 1, nangle
        theta = real(j) * dtheta 
        xpos = xcentre + rad * cos( theta )
            

         
        if (xPos .lt. xLower .or. xPos .gt. xUpper) then
            cycle
        end if 
        ypos = ycentre + rad * sin( theta )
        if (yPos .lt. yLower .or. yPos .gt. yUpper) then
            cycle
        end if 
        k = k + 1 
                 
          
        fCircle(k) = Interpolate( backgroundimagePtr, xPos, yPos )
        if( isNullValue(fCircle(k) )) then
            k = k - 1
        end if
           
    end do






    if (k > 10) then
 !       write(*, *) "omdetectdetection line calling medianvalue 10466 k=", k
        fBackground = medianValue(fcircle, k) 
        
    else
        deallocate(fcircle, radius, fradius)
        return 
    end if






    do i = 1, nrad
        rad = startRadius + real(i - 1) * dr
        k = 0
        do j = 1, nangle
            theta = real(j) * dtheta 
            xpos = xcentre + rad * cos( theta )
            

         
            if (xPos .lt. xLower .or. xPos .gt. xUpper) then
                cycle
            end if 
            ypos = ycentre + rad * sin( theta )
            if (yPos .lt. yLower .or. yPos .gt. yUpper) then
                cycle
            end if 
            k = k + 1 
                 
          
            fCircle(k) = Interpolate( imagePtr, xPos, yPos )
            if( isNullValue(fCircle(k) ) )then
                k = k - 1
            end if
   
        
        end do
         
 



        if (k > 10) then
            
!      write(*, *) "omdetectdetection line calling medianvalue 10512 k=", k
 

            f = medianValue(fcircle, k) 
 
 

            if (n > 0) then
                if (f < fradius(n)) then



                    n = n + 1
                    radius(n) = rad
                    fradius(n) = f        



                end if      
            else


                n = n + 1
                radius(n) = rad
                fradius(n) = f        
 
   
              
            end if 
    
   
        end if




    end do 



    if (n > 4) then          
        !*************************************************
        !
        !*************************************************
    

        sourceCircleradii => radius
        sourceCirclef => fradius

        nSourceRadii = n
        dy = 0.0   ! No errors
        order = 2  ! Second-order polynomial fit
   
        !********************************************************************************
        ! Fit a least-squares polynoimal y=a + b * x + c * x * x to the profile
        !******************************************************************************** 
        call leastsquaresPolyFit(n, radius, fradius, dy, order, 1_int8, xcoeff, dxcoeff, std) 
 
!       write(*, *) "Polynomail fit- std=", std 

        xmin = dble( radius(1) ) - 20.0
        xmax = dble( radius(n) ) + 20.0
        tol = 0.000001_double

        !*********************************************************************************
        ! Locate the radius where the profile reaches a height of fbackground
        !*********************************************************************************

        f = BRENT( xmin, xmax, tol, xmin, radiusFunction ) 

        nullify(sourceCircleRadii, sourceCirclef)

        maxRadius = min(real(80 / binx), real(xmin))




!        write(*, *) "Sourcecircle no=", source%number, " x=", source%xpos, " y=", source%ypos
!        write(*, *) "fbackground=", fbackground
!        write(*, *) "maxradius=", xmin, " f=", f

        if ( f < 1.0 ) then ! Reasonable function minimum
            source%axialRatio = maxRadius * real( binx )





            nx = ubound(imagePtr, 1)
            ny = ubound(imagePtr, 2)


            i0 = int(xcentre)
            j0 = int(ycentre)
           
            maxRadius = min(maxRadius, real(40 / binx))        

            dx1 = int(maxRadius + 2)
            il = max( 0_int16, i0 - dx1)
            iu = min( nx, i0 + dx1)
            jl = max( 0_int16, j0 - dx1)
            ju = min( ny, j0 + dx1)

                

        end if


    end if    



    deallocate(fcircle, radius, fradius, dy)
!call fatal("stopped")

end subroutine sourceCircle
!**********************************************************************
!
!
!*********************************************************************
real(double) function radiusFunction( x )
    real(double), intent(in) :: x
    real(double) :: f


    f = dble( EvaluatePolyNomial( real(x), order, xcoeff ) )

    radiusFunction = dabs(f - fBackground)

!write(*, *) "x=", x, " f=", f, " radiusfunctions=", radiusFunction

end function radiusFunction





!******************************************************************
!
!*****************************************************************
logical Function IsMod8Source1(source, image, backgroundimage, imin, imax, jmin, jmax)

    type(sourceParameters), intent(inout) :: source
    real(single), pointer, dimension(:, :) :: image, backgroundimage
    integer(int16), intent(out) :: imin, imax, jmin, jmax


    integer(int16) :: nx, ny, i, j, mincount, count, i0, j0
    real(single) :: nsigma = 2.0
    integer(int16) :: il, iu, jl, ju, l, maxwidth
  
    integer(int8) :: num

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    imin = source%imin
    imax = source%imax
    jmin = source%jmin
    jmax = source%jmax
    
    nsigma = 2.0

  
   

    i0 = int(source%xpos - 0.4999)
    j0 = int(source%ypos - 0.4999)
    
    num = 0

    !******************************************************************
    ! Check to the left of the centre
    !*****************************************************************

    iu = max(0_int16, i0 - int(8 / binx, int16))
    il = max(0_int16, i0 - int(30 / binx, int16))
   
    jl = max(0_int16, j0 - int(10 / binx, int16))
    ju = min(ny, j0 + int(10 / binx, int16))
    mincount = ju - jl
   
    
    maxwidth = 0


    l = 0
    do i = iu, il, -1
        count = 0
        do j = jl, ju
            if(image(i, j) < backgroundimage(i, j) - nsigma * max(1.0, &
                           sqrt(backgroundImage(i, j) ))) then
                count = count + 1
            end if
        end do
        if(count >= mincount) then
            l = 1
            maxwidth = max(maxwidth, i0 - i)
        end if     
    end do          
    num = num + l 

    !******************************************************************
    ! Check to the right of the centre
    !*****************************************************************


    il = min(nx, i0 + int(8 / binx, int16))
    iu = min(nx, i0 + int(30 / binx, int16))
    
  
    
    l = 0
    do i = il, iu
        count = 0
        do j = jl, ju
            if(image(i, j) < backgroundimage(i, j) - nsigma * max(1.0, &
                           sqrt(backgroundImage(i, j) ))) then
                count = count + 1
            end if
        end do
        if(count >= mincount) then
            l = 1
            maxwidth = max(maxwidth, i - i0)

        end if     
    end do          
    num = num + l

    !******************************************************************
    ! Check to the top of the centre
    !*****************************************************************

    ju = min(ny, j0 + int(8 / binx, int16))
    jl = min(ny, j0 + int(30 / binx, int16))
   
    il = max(0_int16, i0 - int(10 / binx, int16))
    iu = min(nx, i0 + int(10 / binx, int16))

    mincount = iu - il
   
    l = 0

    do j = jl, ju
        count = 0
        do i = il, iu
            if(image(i, j) < backgroundimage(i, j) - nsigma * max(1.0, &
                           sqrt(backgroundImage(i, j) ))) then
                count = count + 1
            end if
        end do

        if(count >= mincount) then
            l = 1
            maxwidth = max(maxwidth, j - j0)

        end if     
    end do          
    num = num + l

    !******************************************************************
    ! Check to the bottom of the centre
    !*****************************************************************

    ju = max(0_int16, j0 - int(8 / binx, int16))
    jl = max(0_int16, j0 - int(30 / binx, int16))
   
 
    
    l = 0
    do j = ju, jl, -1
        count = 0
        do i = il, iu
            if(image(i, j) < backgroundimage(i, j) - nsigma * max(1.0, &
                           sqrt(backgroundImage(i, j) ))) then
                count = count + 1
            end if
        end do
        if(count >= mincount) then
            l = 1
            maxwidth = max(maxwidth, j0 - j)

        end if     
    end do          

    num = num + l
    if(num > 1) then
        IsMod8Source1 = .true.

        imin = max(0_int16, i0 - maxWidth)
        imax = min(nx, i0 + maxWidth)
       
        jmin = max(0_int16, j0 - maxWidth)
        jmax = min(ny, j0 + maxWidth)

!        write(*, *) "ismod8source1- mod-8 pattern around source x=", source%xpos, " y=", source%ypos
!        write(*, *) "imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax 


    else
        IsMod8Source1 = .false.
    endif




end function ismod8source1


!*************************************************************************
!
!*************************************************************************
!subroutine ComputeSourceSeprations(nsource, source)
!    integer(int16), intent(in) :: nsource
!    type(sourceParameters), pointer :: source(:)
!    real(single) :: distance   
!    integer(int16) :: i, j

!    do i = 1, nsource
!        if( isnullvalue(source(i)%semia) .or. source(i)%semia > 100.0) cycle
!        do j = i + 1, nsource
!            if( isnullvalue(source(j)%semia) .or. source(j)%semia > 100.0) cycle
     
!            distance = sourceSeparation(source(i), source(j), 3.0)
!            if(abs( distance ) < 20.0) then

!                write(*, *) "Source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos
!                write(*, *)               " semia=", source(i)%semia, " semib=", source(i)%semib, &
!                        " theta=", source(i)%theta

!                write(*, *) "Source ", j, " x=", source(j)%xpos, " y=", source(j)%ypos

!                write(*, *)               " semia=", source(j)%semia, " semib=", source(j)%semib, &
!                        " theta=", source(j)%theta

!                write(*, *) "Separation=", distance


!            end if
!        end do
!    end do

!end subroutine ComputeSourceSeparations
!****************************************************************
!
!
!***************************************************************
real(single) function sourceSeparation1(source1, source2, nsigma)
    
    type(sourceParameters), intent(in) :: source1, source2


    real(single), intent(in) :: nsigma
    real(single) :: dx, dy, angle, theta, radius
    real(single) :: x, y, r1, r2


    !*********************************************************
    ! Compute the distance between the 2 sources
    !**********************************************************
    dx = source1%xpos - source2%xpos
    dy = source1%ypos - source2%ypos
  
   
    radius = sqrt(dx * dx + dy * dy)




    !********************************************************************
    ! Compute the distance of source 1 from source 2 in terms
    ! of source 1's semi-major axis,  semi-minor axis and position angle
    !*********************************************************************


    if (abs(dx) .ne. 0.0 .or. abs(dy) .ne. 0.0) then
        angle = atan2(dy, dx)
    else
        angle = 0.0
    end if
    if (angle .lt. 0.0) angle = angle + PI

    theta = angle + source1%theta * degreesToRadians
    if(theta .lt. 0.0) theta = theta + PI

!write(*, *) "angle=", angle * radiansToDegrees
!write(*, *) " pa1=", source1%theta, " pa2=", source2%theta 

    x = abs(source1%semia * cos( theta ))
    y = abs(source1%semib * sin( theta ))
       
!   write(*, *) "x=", x, " y=", y
    r1 = sqrt( x * x + y * y)
 
    !********************************************************************
    ! Compute the distance of source 1 from source 2 in terms
    ! of source 2's semi-major axis,  semi-minor axis and position angle
    !*********************************************************************

    theta = angle + source2%theta * degreesToRadians
    if(theta .lt. 0.0) theta = theta + PI


    x = abs(source2%semia * cos( theta ))
    y = abs(source2%semib * sin( theta ))
!   write(*, *) "x=", x, " y=", y
 

    r2 = sqrt( x * x + y * y)
 
  
    !**********************************************************
    ! If separation is <= 0.0 then assume to be overlapping
    !**********************************************************

    sourceSeparation1 = max(0.0, radius - nsigma * (r1 + r2))



!write(*, *) "Source1- x=", source1%xpos, " y=",source1%ypos, " a=", &
!  source1%semia, " b=", source1%semia, " b=", source1%theta, " r1=", r1



!write(*, *) "Source2- x=", source2%xpos, " y=",source2%ypos, " a=", &
!  source2%semia, " b=", source2%semia, " b=", source2%theta, " r1=", r1
!write(*, *) "radius=", radius

end function sourceSeparation1




!****************************************************************
!
! Computes distance between source and x, y, in terms
! of source radius
!
!***************************************************************
real(single) function sourceSeparation2(source, xPos, yPos, radius1)
    
    type(sourceParameters), intent(in) :: source
    real(single), intent(in) :: xPos, yPos
    real(single), optional, intent(out) :: radius1

    real(single) :: dx, dy, angle, theta
    real(single) :: x, y, distance


    !*********************************************************
    ! Compute the distance between the 2 sources
    !**********************************************************
    dx = source%xpos - xpos
    dy = source%ypos - ypos
  
    distance = sqrt(dx * dx + dy * dy) 

    if (present(radius1) )  radius1 = distance

    !********************************************************************
    ! Compute the distance of source 1 from source 2 in terms
    ! of source 1's semi-major axis,  semi-minor axis and position angle
    !*********************************************************************


    if ( (abs(dx) + abs(dy)) .ne. 0.0) then
        angle = atan2(dy, dx)
    else
        angle = 0.0
    end if
    if (angle .lt. 0.0) angle = angle + TWO_PI

    theta = angle - source%theta * degreesToRadians
    if(theta .lt. 0.0) theta = theta + TWO_PI

    !*******************************************************
    !
    ! source separation as a ratio of true distance / sigma
    !
    !******************************************************
  

    x = abs(source%semia * cos( theta )) 
    y = abs(source%semib * sin( theta )) 
!    write(*, *) "Sourceseparation2"

!    write(*, *) "dx=", dx, " dy=", dy, " angle=", angle * radiansToDegrees, " theta=", theta * radiansToDegrees
!    write(*, *) "x=", x, " y=", y, " distance=", distance, &
!       " distance in terms of axes=", distance / sqrt( x * x + y * y)    
 
    sourceSeparation2 = distance / sqrt( x * x + y * y)
 
end function sourceSeparation2


!*****************************************************************
!
!
!
!****************************************************************
logical function checkSourceCirclePixels( source, imagePtr, radius)

    type(sourceParameters), intent(in) :: source
    real(single), pointer, dimension(:, :) :: imagePtr
    real(single), intent(in) :: radius
    real(single) :: xcentre, ycentre, theta
 
    integer :: j, k, k1
    integer :: nangle
    real(single) :: dtheta, xPos, yPos, median, f, fMin, fMax
    real(single), pointer, dimension(:) :: fcircle => null()
   
  
    nangle = 36

    fcircle => allocateSingle(1, nangle, "fcircle")
   
    dtheta = real(360 / nangle) * degreesToRadians

    xcentre = source%xpos - 0.5
    ycentre = source%ypos - 0.5



    k = 0

    do j = 1, nangle
        theta = real(j) * dtheta 
        xpos = xcentre + radius * cos( theta )
            

         
        if (xPos .lt. xLower .or. xPos .gt. xUpper) then
            cycle
        end if 
        ypos = ycentre + radius * sin( theta )
        if (yPos .lt. yLower .or. yPos .gt. yUpper) then
            cycle
        end if 
     
                 
          
        f = Interpolate( imagePtr, xPos, yPos )
        if( .not. isNullValue( f )) then
            k = k + 1
            fcircle(k) = f
        end if
           
    end do
    checkSourceCirclePixels = .true.

    if(k .ge. nAngle / 2) then
        k1 = k / 2 
        call MinimumList(k, fCircle(1 : nangle), k1 ) 




        median = max(1.0, medianValue(fcircle, k1)) 
        fMin = max(1.0, minVal(fCircle(1 : k1)))
        fMax = maxVal(fCircle(1 : k1))

        if( fMax .gt. median + 10.0 * sqrt(median) ) then
            checkSourceCirclePixels = .false.
             !write(*, *) "Line 11098 Going to delete source ", source%number, " x=", xcentre, " y=", ycentre
                !write(*, *) "Median=", median, " fmin=", fmin, " fmax=", fmax


        else
            if( fMax / fMin .gt. 10.0) then

                !write(*, *) "line 11105 Going to delete source ", source%number, " x=", xcentre, " y=", ycentre
                !write(*, *) "Median=", median, " fmin=", fmin, " fmax=", fmax

                checkSourceCirclePixels = .false.
   
            end if 

        end if  
        

    end if
    deallocate(fcircle)

    nullify(fcircle)

end function checkSourceCirclePixels







!**************************************************************
!
! Computes the area of the source using the semi-major 
! and minor-axes and compares it with the number of pixels
! that have been assigned to the source
! 
!
!*************************************************************
logical function sourceAreaOkay(source, numPixels, maxRatio1)
    type(SourceParameters), intent(in) :: source
    integer, intent(in) :: numPixels      

 
    real(single), intent(in) :: maxRatio1
    real(double) :: area, maxRatio
    real(single) :: theta, cs, sn, x, y, x0, y0, a1, b1
    integer(int16) :: xMin, xMax, yMin, yMax
    integer(int16) :: i0, j0, i, j, i1, i2, j1, j2, r
    integer :: count1, count2, count3   

!         sourceAreaOkay = .true.
!return



    if( numPixels < 10 / binX) then
        sourceAreaOkay = .true.
        return
    end if


    if (isNullValue( source%semia ) .or. isNullValue(source%semib) ) then
         sourceAreaOkay = .true.
         return
    else

           !**********************************************************
        !
        ! The following should enclose most of the source
        !
        !**********************************************************         

        area = 4.0 * PI * source%semia * source%semib

        if( (source%semia / source%semib) > 3.0) then
            maxRatio = 0.5
        else     
            maxRatio = maxRatio1
        end if 

! if(checkPixel) then
!    write(*, *) "Line 10887, area (4xPIxaxb)=", &
!                  area, " maxratio=", maxRatio, " numpixels=", numPixels
!                write(*, *) "x=", source%xpos, " y=", source%ypos, &
!                          " a=", source%semia, " b=", source%semib
!end if
        if( area  > maxRatio * dble(numPixels) ) then

!            if(checkPixel) then
!                write(*, *) "Line rejecting source at line 11176, area=", &
!                  area, " maxratio=", maxRatio, " numpixels=", numPixels
!                write(*, *) "x=", source%xpos, " y=", source%ypos, &
!                    " a=", source%semia, " b=", source%semib
!            end if              
            sourceAreaOkay = .false.
        else

            if( source%theta < 0.0) then
                theta = (source%theta + 180.0) * degreesToRadians
            else
                theta = source%theta * degreesToRadians
            end if
            cs = cos(theta)
            sn = sin(theta) 
            xMin = int(source%xpos - source%semia * cs - 0.5, int16)
            xMax = int(source%xpos + source%semia * cs - 0.5, int16)
            yMin = int(source%ypos - source%semib * sn - 0.5, int16)
            yMax = int(source%ypos + source%semib * sn - 0.5, int16)
            if(   (xMin < source%iMin .and. xMax > source%iMax ) .or. &
                  (yMin < source%jMin .and. yMax > source%jMax ) ) then
                sourceAreaOkay = .false.
                return
        
            end if    
        end if
    end if
  

    !***************************************************************************
    !
    !
    !
    !**************************************************************************
    i0 = int(source%xpos - 0.5)
    j0 = int(source%yPos - 0.5)
   
        
    i0 = int(x0)
    j0 = int(y0)
    
    a1 = source%semiA * 2.0
    b1 = source%semiB * 2.0
    
    r = ceiling( a1 )
   
    ! Determine upper and lower array indices for
    ! summing counts

    i1 = i0 - r - 1
    i1 = max(0_int16, i1)
  
    i2 = i0 + r + 1
    i2 = min(nx1, i2)  

    j1 = j0 - r - 1
    j1 = max(0_int16, j1)

    j2 = j0 + r + 1
    j2 = min(ny1, j2)
 

    x0 = source%xPos - 1.0
    y0 = source%yPos - 1.0


    if( source%theta < 0.0) then
        theta = (source%theta + 180.0) * degreesToRadians
    else
        theta = source%theta * degreesToRadians
    end if
 
    count1 = 0    
    count2 = 0
    count3 = 0

    do i = i1, i2
        x = real(i)
        do j = j1, j2  
            y = real(j)
       
              
            !********************************************************************
            ! Point has to be within the inner and outer ellipse
            !********************************************************************
            if ( PointInEllipse(x, y, x0, y0, a1, b1, theta) ) then
                count1 = count1 + 1
           
                if( elevelImagePtr(i, j) == number) then
                  
                    count2 = count2 + 1                   
                else
            
                    count3 = count3 + 1
                end if 
            end if
                    
        end do           
    end do 

    if( real(count1) > (2.0 * real(count2)) .or. real(count3) .gt. real(count2) ) then

        sourceAreaOkay = .false.
    else

        sourceAreaOkay = .true.
    end if    
 

end function sourceAreaOkay


end module OmdetectDetection_mod


